{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Path Data and Higher-Order De Bruijn Graphs\n",
    "\n",
    "## Prerequisites\n",
    "\n",
    "First, we need to set up our Python environment that has PyTorch, PyTorch Geometric and PathpyG installed. Depending on where you are executing this notebook, this might already be (partially) done. E.g. Google Colab has PyTorch installed by default so we only need to install the remaining dependencies. The DevContainer that is part of our GitHub Repository on the other hand already has all of the necessary dependencies installed. \n",
    "\n",
    "In the following, we install the packages for usage in Google Colab using Jupyter magic commands. For other environments comment in or out the commands as necessary. For more details on how to install `pathpyG` especially if you want to install it with GPU-support, we refer to our [documentation](https://www.pathpy.net/dev/getting_started/). Note that `%%capture` discards the full output of the cell to not clutter this tutorial with unnecessary installation details. If you want to print the output, you can comment `%%capture` out."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%capture\n",
    "# !pip install torch\n",
    "!pip install torch_geometric\n",
    "!pip install git+https://github.com/pathpy/pathpyG.git"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Motivation and Learning Objective\n",
    "\n",
    "While `pathpyG` is useful to handle and visualize static graphs - as the name suggests - its main advantage is that it facilitates the analysis of time series data that can be used to calculate **paths** in a graph. As we shall see in the following tutorial, there are various situations in which naturally have access to data on paths, including data on (random) walks or trajectories, traces of dynamical processes giving rise to node sequences or directed acyclic graphs, or time-respecting paths in temporal graphs. ``pathpyG` can be used to model patterns in such data based on higher-order De Bruijn graph models.\n",
    "\n",
    "In this first unit, we will show how `pathpyG` supports to represent data on paths in graphs. Lile graphs, such data are internally stored as tensors, which facilitates GPU-based operations to create higher-order De Bruijn graphs.\n",
    "\n",
    "We first import the modules `torch` and `pathpyG`. By setting the device used by `torch`, we can specify whether we want to run our code on the CPU or on the GPU."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "import pathpyG as pp\n",
    "from torch_geometric.data import Data\n",
    "from torch_geometric import EdgeIndex\n",
    "\n",
    "pp.config['torch']['device'] = 'cpu'"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "For the following examples, we consider a simple directed graph with five nodes `a`, `b`, `c`, `d`, `e` and four edges:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<style>\n",
       "svg circle.node {\n",
       "  fill: #3b5998;\n",
       "  stroke: #1b3978;\n",
       "  stroke-width: 2.5px;\n",
       "  r: 15px;\n",
       "  opacity: 1;\n",
       "}\n",
       "\n",
       ".link {\n",
       "  stroke: #969595;\n",
       "  stroke-opacity: .75;\n",
       "  stroke-width: 2.5px;\n",
       "}\n",
       "\n",
       ".arrow {\n",
       "  fill: #969595;\n",
       "}\n",
       "\n",
       "\n",
       ".label-text {\n",
       "    fill: #969595;\n",
       "    font-size: 16px;\n",
       "    font-family: sans-serif;\n",
       "}\n",
       "\n",
       "</style>\n",
       "\n",
       "<div id = \"x5150ef7aa1e44915b94230bef19db660\"> </div>\n",
       "<script charset=\"utf-8\" src=\"https://d3js.org/d3.v5.min.js\"></script>\n",
       "<script charset=\"utf-8\">\n",
       "// Load via requireJS if available (jupyter notebook environment)\n",
       "try {\n",
       "    // Problem: require.config will raise an exception when called for the second time \n",
       "    require.config({\n",
       "        paths: {\n",
       "            d3: \"https://d3js.org/d3.v5.min.js\".replace(\".js\", \"\")\n",
       "        }\n",
       "    });\n",
       "    console.log(\"OKAY: requireJS was detected.\");\n",
       "}\n",
       "catch(err){\n",
       "    // a reference error indicates that requireJS does not exist. \n",
       "    // other errors may occur due to multiple calls to config\n",
       "    if (err instanceof ReferenceError){\n",
       "        console.log(\"WARNING: NO requireJS was detected!\");\n",
       "\n",
       "        // Helper function that waits for d3js to be loaded\n",
       "        require = function require(symbols, callback) {\n",
       "            var ms = 10;\n",
       "            window.setTimeout(function(t) {\n",
       "                if (window[symbols[0]])\n",
       "                    callback(window[symbols[0]]);\n",
       "                else \n",
       "                    window.setTimeout(arguments.callee, ms);\n",
       "            }, ms);\n",
       "        }\n",
       "    }\n",
       "};\n",
       "require(['d3'], function(d3){ //START\n",
       "const data = {\"edges\": [{\"uid\": \"a-c\", \"source\": \"a\", \"target\": \"c\", \"weight\": 1, \"color\": \"gray\"}, {\"uid\": \"b-c\", \"source\": \"b\", \"target\": \"c\", \"weight\": 1, \"color\": \"gray\"}, {\"uid\": \"c-d\", \"source\": \"c\", \"target\": \"d\", \"weight\": 1, \"color\": \"gray\"}, {\"uid\": \"c-e\", \"source\": \"c\", \"target\": \"e\", \"weight\": 1, \"color\": \"gray\"}], \"nodes\": [{\"uid\": \"a\", \"label\": \"a\"}, {\"uid\": \"b\", \"label\": \"b\"}, {\"uid\": \"c\", \"label\": \"c\"}, {\"uid\": \"d\", \"label\": \"d\"}, {\"uid\": \"e\", \"label\": \"e\"}]}\n",
       "const config = {\"node_label\": [\"a\", \"b\", \"c\", \"d\", \"e\"], \"edge_color\": \"gray\", \"directed\": true, \"curved\": true, \"selector\": \"#x5150ef7aa1e44915b94230bef19db660\"}\n",
       "console.log(\"Static Network Template\");\n",
       "/* Resources\n",
       "   https://bl.ocks.org/mapio/53fed7d84cd1812d6a6639ed7aa83868\n",
       "   https://codepen.io/smlo/pen/JdMOej\n",
       "*/\n",
       "\n",
       "// variables from the config file\n",
       "const selector = config.selector;\n",
       "const width = config.width || 800;\n",
       "const height = config.height || 600;\n",
       "const charge_distance = config.charge_distance || 400;\n",
       "const charge_force = config.charge_force || -3000;\n",
       "const curved = config.curved || false;\n",
       "const directed = config.directed || false;\n",
       "// const weight = false;\n",
       "\n",
       "/* Create a svg element to display the network */\n",
       "var svg = d3.select(selector)\n",
       "    .append('svg')\n",
       "    .attr('width', width)\n",
       "    .attr('height', height)\n",
       "\n",
       "// add container to store the elements\n",
       "var container = svg.append(\"g\");\n",
       "\n",
       "/*Add zoom function to the container */\n",
       "svg.call(\n",
       "    d3.zoom()\n",
       "        .scaleExtent([.1, 4])\n",
       "        .on(\"zoom\", function() { container.attr(\"transform\", d3.event.transform); })\n",
       ");\n",
       "\n",
       "\n",
       "/*Load nodes and links from the data */\n",
       "var nodes = data.nodes\n",
       "var links = data.edges\n",
       "\n",
       "/*Create arrow head with same color as the edge */\n",
       "function marker (color) {\n",
       "       var reference;\n",
       "       svg.append(\"svg:defs\").selectAll(\"marker\")\n",
       "          .data([reference])\n",
       "          .enter().append(\"svg:marker\")\n",
       "          .attr(\"id\", \"arrow\"+color)\n",
       "          .attr(\"viewBox\", \"0 -5 10 10\")\n",
       "          .attr(\"refX\", 10)\n",
       "          .attr(\"refY\", -0)\n",
       "          .attr(\"markerWidth\", 6)\n",
       "          .attr(\"markerHeight\", 6)\n",
       "          .attr(\"orient\", \"auto\")\n",
       "          .append(\"svg:path\")\n",
       "          .attr('class','.arrow')\n",
       "          .attr(\"d\", \"M0,-5L10,0L0,5\")\n",
       "          .style('opacity',1)\n",
       "          .style(\"fill\", color);\n",
       "       return \"url(#\" + \"arrow\"+color + \")\";\n",
       "     };\n",
       "\n",
       "/*Link creation template */\n",
       "var link = container.append(\"g\").attr(\"class\", \"links\")\n",
       "    .selectAll(\".link\")\n",
       "    .data(links)\n",
       "    .enter()\n",
       "    .append(\"path\")\n",
       "    .attr(\"class\", \"link\")\n",
       "    .style(\"stroke\", function(d) { return d.color; })\n",
       "    .style(\"stroke-opacity\", function(d) { return d.opacity; })\n",
       "    .style(\"stroke-width\", function(d){  return d.size })\n",
       "    .style(\"fill\",\"none\")\n",
       "    .attr(\"marker-end\", function (d) {if(directed){return marker(d.color)}else{return null}; });\n",
       "\n",
       "    //.attr(\"marker-end\", function (d) { return marker(d.color); });\n",
       "    //.attr(\"marker-end\", \"url(#arrow)\");\n",
       "\n",
       "/*Node creation template */\n",
       "var node = container.append(\"g\").attr(\"class\", \"nodes\")\n",
       "    .selectAll(\"circle.node\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"circle\")\n",
       "    .attr(\"class\", \"node\")\n",
       "    .attr(\"x\", function(d) { return d.x; })\n",
       "    .attr(\"y\", function(d) { return d.y; })\n",
       "    .style(\"r\", function(d){  return d.size+\"px\"; })\n",
       "    .style(\"fill\", function(d) { return d.color; })\n",
       "    .style(\"opacity\", function(d) { return d.opacity; });\n",
       "\n",
       "/*Label creation template */\n",
       "var text = container.append(\"g\").attr(\"class\",\"labels\")\n",
       "    .selectAll(\"g\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"g\")\n",
       "\n",
       "text.append(\"text\")\n",
       "    .attr(\"class\", \"label-text\")\n",
       "    .attr(\"x\", function(d) {\n",
       "        var r = (d.size === undefined) ? 15 : d.size;\n",
       "        return 5 + r; })\n",
       "    .attr(\"dy\", \".31em\")\n",
       "    .text(function(d) { return d.label; });\n",
       "\n",
       "/*Scale weight for d3js */\n",
       "var weightScale = d3.scaleLinear()\n",
       "    .domain(d3.extent(links, function (d) { return d.weight }))\n",
       "    .range([.1, 1]);\n",
       "\n",
       "/*Simulation of the forces*/\n",
       "var simulation = d3.forceSimulation(nodes)\n",
       "    .force(\"links\", d3.forceLink(links)\n",
       "           .id(function(d) {return d.uid; })\n",
       "           .distance(50)\n",
       "           .strength(function(d){return weightScale(d.weight);})\n",
       "          )\n",
       "    .force(\"charge\", d3.forceManyBody()\n",
       "           .strength(charge_force)\n",
       "           .distanceMax(charge_distance)\n",
       "          )\n",
       "    .force(\"center\", d3.forceCenter(width / 2, height / 2))\n",
       "    .on(\"tick\", ticked);\n",
       "\n",
       "/*Update of the node and edge objects*/\n",
       "function ticked() {\n",
       "    node.call(updateNode);\n",
       "    link.call(updateLink);\n",
       "    text.call(updateText);\n",
       "};\n",
       "\n",
       "/*Update link positions */\n",
       "function updateLink(link) {\n",
       "    // link\n",
       "    //     .attr(\"x1\", function(d) { return d.source.x; })\n",
       "    //     .attr(\"y1\", function(d) { return d.source.y; })\n",
       "    //     .attr(\"x2\", function(d) { return d.target.x; })\n",
       "    //     .attr(\"y2\", function(d) { return d.target.y; });\n",
       "\n",
       "\n",
       "    link.attr(\"d\", function(d) {\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        return \"M\" +\n",
       "            d.source.x + \",\" +\n",
       "            d.source.y + \"A\" +\n",
       "            dr + \",\" + dr + \" 0 0,1 \" +\n",
       "            d.target.x + \",\" +\n",
       "            d.target.y;\n",
       "    });\n",
       "\n",
       "    // recalculate and back off the distance\n",
       "    link.attr(\"d\", function (d, i) {\n",
       "        var pl = this.getTotalLength();\n",
       "        var r = (d.target.size === undefined) ? 15 : d.target.size;\n",
       "        var m = this.getPointAtLength(pl - r);\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        var result = \"M\" + d.source.x + \",\" + d.source.y + \"A\" + dr + \",\" + dr + \" 0 0,1 \" + m.x + \",\" + m.y;\n",
       "        return result;\n",
       "    });\n",
       "};\n",
       "\n",
       "\n",
       "/*Update node positions */\n",
       "function updateNode(node) {\n",
       "    node.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "    // node\n",
       "    //     .attr(\"cx\", function(d) { return d.x; })\n",
       "    //     .attr(\"cy\", function(d) { return d.y; });\n",
       "};\n",
       "\n",
       "/*Update text positions */\n",
       "function updateText(text) {\n",
       "    text.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "};\n",
       "\n",
       "/*Add drag functionality to the node objects*/\n",
       "node.call(\n",
       "    d3.drag()\n",
       "        .on(\"start\", dragstarted)\n",
       "        .on(\"drag\", dragged)\n",
       "        .on(\"end\", dragended)\n",
       ");\n",
       "\n",
       "function dragstarted(d) {\n",
       "    d3.event.sourceEvent.stopPropagation();\n",
       "    if (!d3.event.active) simulation.alphaTarget(0.3).restart();\n",
       "    d.fx = d.x;\n",
       "    d.fy = d.y;\n",
       "};\n",
       "\n",
       "function dragged(d) {\n",
       "    d.fx = d3.event.x;\n",
       "    d.fy = d3.event.y;\n",
       "};\n",
       "\n",
       "function dragended(d) {\n",
       "    if (!d3.event.active) simulation.alphaTarget(0);\n",
       "    d.fx = null;\n",
       "    d.fy = null;\n",
       "};\n",
       "\n",
       "}); //END\n",
       "\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "g = pp.Graph.from_edge_list([('a', 'c'),\n",
    "                             ('b', 'c'),\n",
    "                             ('c', 'd'),\n",
    "                             ('c', 'e')])\n",
    "pp.plot(g, node_label=g.nodes, edge_color='gray');"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Using `PathData` to store walks or paths in a graph\n",
    "\n",
    "Assume that we have time series data that captures observations of trajectories (i.e. walks or paths) in the graph above. For example, we could observe four walks of length two, four each of the following:\n",
    "\n",
    "- 4 x `a` -> `c` -> `d`  \n",
    "- 4 x `b` -> `c` -> `e` \n",
    "\n",
    "Note that we define the length of a walk or path as the number of edges that are traversed, i.e. a sequence that consists of a single node, e.g. `a`, is considered a walk of length zero, while every edge in a graph is a walk of length one.\n",
    "\n",
    "`pp.PathData` supports to store and model such sequential data. We first create an instance of the `PathData` class. To consistently map node IDs to indices across `Graph` and `PathData` objects, we can pass the `IndexMap` object from the `Graph` above in the constructor. We then use the `append_walk` function to add observations of our two walks, where the `weight` argument is used to indicate the number of times each path or walk has been observed."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "PathData with 2 paths with total weight 8.0\n"
     ]
    }
   ],
   "source": [
    "paths = pp.PathData(g.mapping)\n",
    "\n",
    "paths.append_walk(('a', 'c', 'd'), weight=4.0)\n",
    "paths.append_walk(('b', 'c', 'e'), weight=4.0)\n",
    "print(paths)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Let us inspect how those walks are internally stored in the `PathData` object. We find that the class internally stores a `pyG.Data` object, which contains the properties `edge_index`, `node_sequence`, `dag_weight`, `dag_num_edges` and `dag_num_nodes` that can be used to access all of the individual paths."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "paths.data"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The `edge_index` tensor represents an ordered sequence of edges traversed by the walk, where the indices of nodes map to the `node_sequence` tensor. This additional mapping is neccessary since walks can traverse the same edge multiple times. Moreover, it allows to internally concatenate multiple walks into a single `Data` object, which is needed for fast GPU-based operations on path data. We can access the first path as follows:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "tensor([[0, 1],\n",
       "        [1, 2]])"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "paths.data.edge_index[:, :paths.data.dag_num_edges[0]]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The `node_sequence` tensor tells us that the node with index `1` in the `edge_index` maps to the node in the graph with index `2`, which is node `c`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "tensor([[0],\n",
       "        [2],\n",
       "        [3]])"
      ]
     },
     "execution_count": 5,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "paths.data.node_sequence[:paths.data.dag_num_nodes[0]]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can index the second edge index accordingly and can see that the `node_sequence` tensor maps to the sequence `b -> c -> e`:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "tensor([[[1],\n",
       "         [2]],\n",
       "\n",
       "        [[2],\n",
       "         [4]]])"
      ]
     },
     "execution_count": 6,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "paths.data.node_sequence[paths.data.edge_index[:, paths.data.dag_num_edges[0]:]]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can actually see a collection of walks as a higher-order generalization of the usual way to define graphs as a collection of dyadic edges (which are simply walks of length one). From this point of view, a standard static (weighted) graph is simply a first-order model of node sequences, which only considers the frequency at which edges are traversed. \n",
    "\n",
    "To generate such a first-order model, we can use the class `MultiOderModel` and use the first-layer of the model, which is simply a weighted static graph where edge weights count the number of times each edge is traversed by a path. We will explain the class `MultiOrderModel`, which generalizes this concept to higher-order graph models for any order $k$ in a moment. For now, we can just use it to generate a first-order weighted graph as follows.\n",
    "\n",
    "The generated graph is again based on a `pyG.Data` object that contains an edge_index and edge weights. As we can see, for the example above the edge_index is just a concatenation of the edge indices of individual walks, where the node indices have been mapped to the correct nodes."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "EdgeIndex([[0, 1, 2, 2],\n",
      "           [2, 2, 3, 4]], sparse_size=(5, 5), nnz=4, sort_order=row)\n",
      "tensor([4., 4., 4., 4.])\n"
     ]
    },
    {
     "data": {
      "text/html": [
       "<style>\n",
       "svg circle.node {\n",
       "  fill: #3b5998;\n",
       "  stroke: #1b3978;\n",
       "  stroke-width: 2.5px;\n",
       "  r: 15px;\n",
       "  opacity: 1;\n",
       "}\n",
       "\n",
       ".link {\n",
       "  stroke: #969595;\n",
       "  stroke-opacity: .75;\n",
       "  stroke-width: 2.5px;\n",
       "}\n",
       "\n",
       ".arrow {\n",
       "  fill: #969595;\n",
       "}\n",
       "\n",
       "\n",
       ".label-text {\n",
       "    fill: #969595;\n",
       "    font-size: 16px;\n",
       "    font-family: sans-serif;\n",
       "}\n",
       "\n",
       "</style>\n",
       "\n",
       "<div id = \"x64bf2f21e77d4683bb4db86d19adb16a\"> </div>\n",
       "<script charset=\"utf-8\" src=\"https://d3js.org/d3.v5.min.js\"></script>\n",
       "<script charset=\"utf-8\">\n",
       "// Load via requireJS if available (jupyter notebook environment)\n",
       "try {\n",
       "    // Problem: require.config will raise an exception when called for the second time \n",
       "    require.config({\n",
       "        paths: {\n",
       "            d3: \"https://d3js.org/d3.v5.min.js\".replace(\".js\", \"\")\n",
       "        }\n",
       "    });\n",
       "    console.log(\"OKAY: requireJS was detected.\");\n",
       "}\n",
       "catch(err){\n",
       "    // a reference error indicates that requireJS does not exist. \n",
       "    // other errors may occur due to multiple calls to config\n",
       "    if (err instanceof ReferenceError){\n",
       "        console.log(\"WARNING: NO requireJS was detected!\");\n",
       "\n",
       "        // Helper function that waits for d3js to be loaded\n",
       "        require = function require(symbols, callback) {\n",
       "            var ms = 10;\n",
       "            window.setTimeout(function(t) {\n",
       "                if (window[symbols[0]])\n",
       "                    callback(window[symbols[0]]);\n",
       "                else \n",
       "                    window.setTimeout(arguments.callee, ms);\n",
       "            }, ms);\n",
       "        }\n",
       "    }\n",
       "};\n",
       "require(['d3'], function(d3){ //START\n",
       "const data = {\"edges\": [{\"uid\": \"a-c\", \"source\": \"a\", \"target\": \"c\", \"weight\": 4.0}, {\"uid\": \"b-c\", \"source\": \"b\", \"target\": \"c\", \"weight\": 4.0}, {\"uid\": \"c-d\", \"source\": \"c\", \"target\": \"d\", \"weight\": 4.0}, {\"uid\": \"c-e\", \"source\": \"c\", \"target\": \"e\", \"weight\": 4.0}], \"nodes\": [{\"uid\": \"a\", \"label\": \"a\"}, {\"uid\": \"b\", \"label\": \"b\"}, {\"uid\": \"c\", \"label\": \"c\"}, {\"uid\": \"d\", \"label\": \"d\"}, {\"uid\": \"e\", \"label\": \"e\"}]}\n",
       "const config = {\"node_label\": [\"a\", \"b\", \"c\", \"d\", \"e\"], \"directed\": true, \"curved\": true, \"selector\": \"#x64bf2f21e77d4683bb4db86d19adb16a\"}\n",
       "console.log(\"Static Network Template\");\n",
       "/* Resources\n",
       "   https://bl.ocks.org/mapio/53fed7d84cd1812d6a6639ed7aa83868\n",
       "   https://codepen.io/smlo/pen/JdMOej\n",
       "*/\n",
       "\n",
       "// variables from the config file\n",
       "const selector = config.selector;\n",
       "const width = config.width || 800;\n",
       "const height = config.height || 600;\n",
       "const charge_distance = config.charge_distance || 400;\n",
       "const charge_force = config.charge_force || -3000;\n",
       "const curved = config.curved || false;\n",
       "const directed = config.directed || false;\n",
       "// const weight = false;\n",
       "\n",
       "/* Create a svg element to display the network */\n",
       "var svg = d3.select(selector)\n",
       "    .append('svg')\n",
       "    .attr('width', width)\n",
       "    .attr('height', height)\n",
       "\n",
       "// add container to store the elements\n",
       "var container = svg.append(\"g\");\n",
       "\n",
       "/*Add zoom function to the container */\n",
       "svg.call(\n",
       "    d3.zoom()\n",
       "        .scaleExtent([.1, 4])\n",
       "        .on(\"zoom\", function() { container.attr(\"transform\", d3.event.transform); })\n",
       ");\n",
       "\n",
       "\n",
       "/*Load nodes and links from the data */\n",
       "var nodes = data.nodes\n",
       "var links = data.edges\n",
       "\n",
       "/*Create arrow head with same color as the edge */\n",
       "function marker (color) {\n",
       "       var reference;\n",
       "       svg.append(\"svg:defs\").selectAll(\"marker\")\n",
       "          .data([reference])\n",
       "          .enter().append(\"svg:marker\")\n",
       "          .attr(\"id\", \"arrow\"+color)\n",
       "          .attr(\"viewBox\", \"0 -5 10 10\")\n",
       "          .attr(\"refX\", 10)\n",
       "          .attr(\"refY\", -0)\n",
       "          .attr(\"markerWidth\", 6)\n",
       "          .attr(\"markerHeight\", 6)\n",
       "          .attr(\"orient\", \"auto\")\n",
       "          .append(\"svg:path\")\n",
       "          .attr('class','.arrow')\n",
       "          .attr(\"d\", \"M0,-5L10,0L0,5\")\n",
       "          .style('opacity',1)\n",
       "          .style(\"fill\", color);\n",
       "       return \"url(#\" + \"arrow\"+color + \")\";\n",
       "     };\n",
       "\n",
       "/*Link creation template */\n",
       "var link = container.append(\"g\").attr(\"class\", \"links\")\n",
       "    .selectAll(\".link\")\n",
       "    .data(links)\n",
       "    .enter()\n",
       "    .append(\"path\")\n",
       "    .attr(\"class\", \"link\")\n",
       "    .style(\"stroke\", function(d) { return d.color; })\n",
       "    .style(\"stroke-opacity\", function(d) { return d.opacity; })\n",
       "    .style(\"stroke-width\", function(d){  return d.size })\n",
       "    .style(\"fill\",\"none\")\n",
       "    .attr(\"marker-end\", function (d) {if(directed){return marker(d.color)}else{return null}; });\n",
       "\n",
       "    //.attr(\"marker-end\", function (d) { return marker(d.color); });\n",
       "    //.attr(\"marker-end\", \"url(#arrow)\");\n",
       "\n",
       "/*Node creation template */\n",
       "var node = container.append(\"g\").attr(\"class\", \"nodes\")\n",
       "    .selectAll(\"circle.node\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"circle\")\n",
       "    .attr(\"class\", \"node\")\n",
       "    .attr(\"x\", function(d) { return d.x; })\n",
       "    .attr(\"y\", function(d) { return d.y; })\n",
       "    .style(\"r\", function(d){  return d.size+\"px\"; })\n",
       "    .style(\"fill\", function(d) { return d.color; })\n",
       "    .style(\"opacity\", function(d) { return d.opacity; });\n",
       "\n",
       "/*Label creation template */\n",
       "var text = container.append(\"g\").attr(\"class\",\"labels\")\n",
       "    .selectAll(\"g\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"g\")\n",
       "\n",
       "text.append(\"text\")\n",
       "    .attr(\"class\", \"label-text\")\n",
       "    .attr(\"x\", function(d) {\n",
       "        var r = (d.size === undefined) ? 15 : d.size;\n",
       "        return 5 + r; })\n",
       "    .attr(\"dy\", \".31em\")\n",
       "    .text(function(d) { return d.label; });\n",
       "\n",
       "/*Scale weight for d3js */\n",
       "var weightScale = d3.scaleLinear()\n",
       "    .domain(d3.extent(links, function (d) { return d.weight }))\n",
       "    .range([.1, 1]);\n",
       "\n",
       "/*Simulation of the forces*/\n",
       "var simulation = d3.forceSimulation(nodes)\n",
       "    .force(\"links\", d3.forceLink(links)\n",
       "           .id(function(d) {return d.uid; })\n",
       "           .distance(50)\n",
       "           .strength(function(d){return weightScale(d.weight);})\n",
       "          )\n",
       "    .force(\"charge\", d3.forceManyBody()\n",
       "           .strength(charge_force)\n",
       "           .distanceMax(charge_distance)\n",
       "          )\n",
       "    .force(\"center\", d3.forceCenter(width / 2, height / 2))\n",
       "    .on(\"tick\", ticked);\n",
       "\n",
       "/*Update of the node and edge objects*/\n",
       "function ticked() {\n",
       "    node.call(updateNode);\n",
       "    link.call(updateLink);\n",
       "    text.call(updateText);\n",
       "};\n",
       "\n",
       "/*Update link positions */\n",
       "function updateLink(link) {\n",
       "    // link\n",
       "    //     .attr(\"x1\", function(d) { return d.source.x; })\n",
       "    //     .attr(\"y1\", function(d) { return d.source.y; })\n",
       "    //     .attr(\"x2\", function(d) { return d.target.x; })\n",
       "    //     .attr(\"y2\", function(d) { return d.target.y; });\n",
       "\n",
       "\n",
       "    link.attr(\"d\", function(d) {\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        return \"M\" +\n",
       "            d.source.x + \",\" +\n",
       "            d.source.y + \"A\" +\n",
       "            dr + \",\" + dr + \" 0 0,1 \" +\n",
       "            d.target.x + \",\" +\n",
       "            d.target.y;\n",
       "    });\n",
       "\n",
       "    // recalculate and back off the distance\n",
       "    link.attr(\"d\", function (d, i) {\n",
       "        var pl = this.getTotalLength();\n",
       "        var r = (d.target.size === undefined) ? 15 : d.target.size;\n",
       "        var m = this.getPointAtLength(pl - r);\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        var result = \"M\" + d.source.x + \",\" + d.source.y + \"A\" + dr + \",\" + dr + \" 0 0,1 \" + m.x + \",\" + m.y;\n",
       "        return result;\n",
       "    });\n",
       "};\n",
       "\n",
       "\n",
       "/*Update node positions */\n",
       "function updateNode(node) {\n",
       "    node.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "    // node\n",
       "    //     .attr(\"cx\", function(d) { return d.x; })\n",
       "    //     .attr(\"cy\", function(d) { return d.y; });\n",
       "};\n",
       "\n",
       "/*Update text positions */\n",
       "function updateText(text) {\n",
       "    text.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "};\n",
       "\n",
       "/*Add drag functionality to the node objects*/\n",
       "node.call(\n",
       "    d3.drag()\n",
       "        .on(\"start\", dragstarted)\n",
       "        .on(\"drag\", dragged)\n",
       "        .on(\"end\", dragended)\n",
       ");\n",
       "\n",
       "function dragstarted(d) {\n",
       "    d3.event.sourceEvent.stopPropagation();\n",
       "    if (!d3.event.active) simulation.alphaTarget(0.3).restart();\n",
       "    d.fx = d.x;\n",
       "    d.fy = d.y;\n",
       "};\n",
       "\n",
       "function dragged(d) {\n",
       "    d.fx = d3.event.x;\n",
       "    d.fy = d3.event.y;\n",
       "};\n",
       "\n",
       "function dragended(d) {\n",
       "    if (!d3.event.active) simulation.alphaTarget(0);\n",
       "    d.fx = null;\n",
       "    d.fy = null;\n",
       "};\n",
       "\n",
       "}); //END\n",
       "\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "m = pp.MultiOrderModel.from_PathData(paths, max_order=1)\n",
    "g = m.layers[1]\n",
    "print(g.data.edge_index)\n",
    "print(g.data.edge_weight)\n",
    "pp.plot(g, node_label=paths.mapping.node_ids.tolist());"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Why are data on paths and walks interesting in the first place. The answer is that they provide information on the **causal topology of complex systems**, i.e. which nodes can possibly causally influence each other via paths that follow the arrow of time. This information is lost if we were to split paths into an (unordered) collection of dyadyic interactions between pairs of nodes, i.e. if we were to only onsider links.\n",
    "\n",
    "To illustrate this, let us assume that the four walks above tell us which paths information (or whatever you may be interested in) can take in the simple graph above. That is, we observe something moving from `a` via `c` to `d` and from `b` via `c` to `e`, and each of those events occur four times. However, we never observed that something moving from `a` to `c` ended up in `d`. And neither did we observe that something moving from `b` to `c` ended up in `e`. This means that - assuming that we completely observed all walks or paths - there is no way that `a` can causally influence `e` or that `b` could causally influence `d` via the center node `c`. Note that this is not what we would assume if we consider possible paths in the topology of the underlying graph, where paths of length two exist between all four pairs of nodes (`a`, `d`), (`a`, `e`), (`b`, `d`), (`b`, `e`).\n",
    "\n",
    "Hence, we can use data capturing actually observed paths or walks ion a network in contrast to which paths or walks would theoretically be possible based on the topology. \n",
    "\n",
    "As a contrast, consider the following observations of walks in the same graph. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "PathData with 4 paths with total weight 8.0\n"
     ]
    }
   ],
   "source": [
    "paths_2 = pp.PathData(g.mapping)\n",
    "\n",
    "paths_2.append_walk(('a', 'c', 'd'), weight=2)\n",
    "paths_2.append_walk(('a', 'c', 'e'), weight=2)\n",
    "paths_2.append_walk(('b', 'c', 'd'), weight=2)\n",
    "paths_2.append_walk(('b', 'c', 'e'), weight=2)\n",
    "print(paths_2)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Here we have observed walks along all four possible paths of length two, each walk occurring only two times. Like in the example before, each edge was traversed exactly four times and thus the weighted edge index of a first-order graph model is identical to the one before:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "EdgeIndex([[0, 1, 2, 2],\n",
      "           [2, 2, 3, 4]], sparse_size=(5, 5), nnz=4, sort_order=row)\n",
      "tensor([4., 4., 4., 4.])\n"
     ]
    },
    {
     "data": {
      "text/html": [
       "<style>\n",
       "svg circle.node {\n",
       "  fill: #3b5998;\n",
       "  stroke: #1b3978;\n",
       "  stroke-width: 2.5px;\n",
       "  r: 15px;\n",
       "  opacity: 1;\n",
       "}\n",
       "\n",
       ".link {\n",
       "  stroke: #969595;\n",
       "  stroke-opacity: .75;\n",
       "  stroke-width: 2.5px;\n",
       "}\n",
       "\n",
       ".arrow {\n",
       "  fill: #969595;\n",
       "}\n",
       "\n",
       "\n",
       ".label-text {\n",
       "    fill: #969595;\n",
       "    font-size: 16px;\n",
       "    font-family: sans-serif;\n",
       "}\n",
       "\n",
       "</style>\n",
       "\n",
       "<div id = \"x0591da7236f5410294b6c1a688576b61\"> </div>\n",
       "<script charset=\"utf-8\" src=\"https://d3js.org/d3.v5.min.js\"></script>\n",
       "<script charset=\"utf-8\">\n",
       "// Load via requireJS if available (jupyter notebook environment)\n",
       "try {\n",
       "    // Problem: require.config will raise an exception when called for the second time \n",
       "    require.config({\n",
       "        paths: {\n",
       "            d3: \"https://d3js.org/d3.v5.min.js\".replace(\".js\", \"\")\n",
       "        }\n",
       "    });\n",
       "    console.log(\"OKAY: requireJS was detected.\");\n",
       "}\n",
       "catch(err){\n",
       "    // a reference error indicates that requireJS does not exist. \n",
       "    // other errors may occur due to multiple calls to config\n",
       "    if (err instanceof ReferenceError){\n",
       "        console.log(\"WARNING: NO requireJS was detected!\");\n",
       "\n",
       "        // Helper function that waits for d3js to be loaded\n",
       "        require = function require(symbols, callback) {\n",
       "            var ms = 10;\n",
       "            window.setTimeout(function(t) {\n",
       "                if (window[symbols[0]])\n",
       "                    callback(window[symbols[0]]);\n",
       "                else \n",
       "                    window.setTimeout(arguments.callee, ms);\n",
       "            }, ms);\n",
       "        }\n",
       "    }\n",
       "};\n",
       "require(['d3'], function(d3){ //START\n",
       "const data = {\"edges\": [{\"uid\": \"a-c\", \"source\": \"a\", \"target\": \"c\", \"weight\": 4.0}, {\"uid\": \"b-c\", \"source\": \"b\", \"target\": \"c\", \"weight\": 4.0}, {\"uid\": \"c-d\", \"source\": \"c\", \"target\": \"d\", \"weight\": 4.0}, {\"uid\": \"c-e\", \"source\": \"c\", \"target\": \"e\", \"weight\": 4.0}], \"nodes\": [{\"uid\": \"a\", \"label\": \"a\"}, {\"uid\": \"b\", \"label\": \"b\"}, {\"uid\": \"c\", \"label\": \"c\"}, {\"uid\": \"d\", \"label\": \"d\"}, {\"uid\": \"e\", \"label\": \"e\"}]}\n",
       "const config = {\"node_label\": [\"a\", \"b\", \"c\", \"d\", \"e\"], \"directed\": true, \"curved\": true, \"selector\": \"#x0591da7236f5410294b6c1a688576b61\"}\n",
       "console.log(\"Static Network Template\");\n",
       "/* Resources\n",
       "   https://bl.ocks.org/mapio/53fed7d84cd1812d6a6639ed7aa83868\n",
       "   https://codepen.io/smlo/pen/JdMOej\n",
       "*/\n",
       "\n",
       "// variables from the config file\n",
       "const selector = config.selector;\n",
       "const width = config.width || 800;\n",
       "const height = config.height || 600;\n",
       "const charge_distance = config.charge_distance || 400;\n",
       "const charge_force = config.charge_force || -3000;\n",
       "const curved = config.curved || false;\n",
       "const directed = config.directed || false;\n",
       "// const weight = false;\n",
       "\n",
       "/* Create a svg element to display the network */\n",
       "var svg = d3.select(selector)\n",
       "    .append('svg')\n",
       "    .attr('width', width)\n",
       "    .attr('height', height)\n",
       "\n",
       "// add container to store the elements\n",
       "var container = svg.append(\"g\");\n",
       "\n",
       "/*Add zoom function to the container */\n",
       "svg.call(\n",
       "    d3.zoom()\n",
       "        .scaleExtent([.1, 4])\n",
       "        .on(\"zoom\", function() { container.attr(\"transform\", d3.event.transform); })\n",
       ");\n",
       "\n",
       "\n",
       "/*Load nodes and links from the data */\n",
       "var nodes = data.nodes\n",
       "var links = data.edges\n",
       "\n",
       "/*Create arrow head with same color as the edge */\n",
       "function marker (color) {\n",
       "       var reference;\n",
       "       svg.append(\"svg:defs\").selectAll(\"marker\")\n",
       "          .data([reference])\n",
       "          .enter().append(\"svg:marker\")\n",
       "          .attr(\"id\", \"arrow\"+color)\n",
       "          .attr(\"viewBox\", \"0 -5 10 10\")\n",
       "          .attr(\"refX\", 10)\n",
       "          .attr(\"refY\", -0)\n",
       "          .attr(\"markerWidth\", 6)\n",
       "          .attr(\"markerHeight\", 6)\n",
       "          .attr(\"orient\", \"auto\")\n",
       "          .append(\"svg:path\")\n",
       "          .attr('class','.arrow')\n",
       "          .attr(\"d\", \"M0,-5L10,0L0,5\")\n",
       "          .style('opacity',1)\n",
       "          .style(\"fill\", color);\n",
       "       return \"url(#\" + \"arrow\"+color + \")\";\n",
       "     };\n",
       "\n",
       "/*Link creation template */\n",
       "var link = container.append(\"g\").attr(\"class\", \"links\")\n",
       "    .selectAll(\".link\")\n",
       "    .data(links)\n",
       "    .enter()\n",
       "    .append(\"path\")\n",
       "    .attr(\"class\", \"link\")\n",
       "    .style(\"stroke\", function(d) { return d.color; })\n",
       "    .style(\"stroke-opacity\", function(d) { return d.opacity; })\n",
       "    .style(\"stroke-width\", function(d){  return d.size })\n",
       "    .style(\"fill\",\"none\")\n",
       "    .attr(\"marker-end\", function (d) {if(directed){return marker(d.color)}else{return null}; });\n",
       "\n",
       "    //.attr(\"marker-end\", function (d) { return marker(d.color); });\n",
       "    //.attr(\"marker-end\", \"url(#arrow)\");\n",
       "\n",
       "/*Node creation template */\n",
       "var node = container.append(\"g\").attr(\"class\", \"nodes\")\n",
       "    .selectAll(\"circle.node\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"circle\")\n",
       "    .attr(\"class\", \"node\")\n",
       "    .attr(\"x\", function(d) { return d.x; })\n",
       "    .attr(\"y\", function(d) { return d.y; })\n",
       "    .style(\"r\", function(d){  return d.size+\"px\"; })\n",
       "    .style(\"fill\", function(d) { return d.color; })\n",
       "    .style(\"opacity\", function(d) { return d.opacity; });\n",
       "\n",
       "/*Label creation template */\n",
       "var text = container.append(\"g\").attr(\"class\",\"labels\")\n",
       "    .selectAll(\"g\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"g\")\n",
       "\n",
       "text.append(\"text\")\n",
       "    .attr(\"class\", \"label-text\")\n",
       "    .attr(\"x\", function(d) {\n",
       "        var r = (d.size === undefined) ? 15 : d.size;\n",
       "        return 5 + r; })\n",
       "    .attr(\"dy\", \".31em\")\n",
       "    .text(function(d) { return d.label; });\n",
       "\n",
       "/*Scale weight for d3js */\n",
       "var weightScale = d3.scaleLinear()\n",
       "    .domain(d3.extent(links, function (d) { return d.weight }))\n",
       "    .range([.1, 1]);\n",
       "\n",
       "/*Simulation of the forces*/\n",
       "var simulation = d3.forceSimulation(nodes)\n",
       "    .force(\"links\", d3.forceLink(links)\n",
       "           .id(function(d) {return d.uid; })\n",
       "           .distance(50)\n",
       "           .strength(function(d){return weightScale(d.weight);})\n",
       "          )\n",
       "    .force(\"charge\", d3.forceManyBody()\n",
       "           .strength(charge_force)\n",
       "           .distanceMax(charge_distance)\n",
       "          )\n",
       "    .force(\"center\", d3.forceCenter(width / 2, height / 2))\n",
       "    .on(\"tick\", ticked);\n",
       "\n",
       "/*Update of the node and edge objects*/\n",
       "function ticked() {\n",
       "    node.call(updateNode);\n",
       "    link.call(updateLink);\n",
       "    text.call(updateText);\n",
       "};\n",
       "\n",
       "/*Update link positions */\n",
       "function updateLink(link) {\n",
       "    // link\n",
       "    //     .attr(\"x1\", function(d) { return d.source.x; })\n",
       "    //     .attr(\"y1\", function(d) { return d.source.y; })\n",
       "    //     .attr(\"x2\", function(d) { return d.target.x; })\n",
       "    //     .attr(\"y2\", function(d) { return d.target.y; });\n",
       "\n",
       "\n",
       "    link.attr(\"d\", function(d) {\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        return \"M\" +\n",
       "            d.source.x + \",\" +\n",
       "            d.source.y + \"A\" +\n",
       "            dr + \",\" + dr + \" 0 0,1 \" +\n",
       "            d.target.x + \",\" +\n",
       "            d.target.y;\n",
       "    });\n",
       "\n",
       "    // recalculate and back off the distance\n",
       "    link.attr(\"d\", function (d, i) {\n",
       "        var pl = this.getTotalLength();\n",
       "        var r = (d.target.size === undefined) ? 15 : d.target.size;\n",
       "        var m = this.getPointAtLength(pl - r);\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        var result = \"M\" + d.source.x + \",\" + d.source.y + \"A\" + dr + \",\" + dr + \" 0 0,1 \" + m.x + \",\" + m.y;\n",
       "        return result;\n",
       "    });\n",
       "};\n",
       "\n",
       "\n",
       "/*Update node positions */\n",
       "function updateNode(node) {\n",
       "    node.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "    // node\n",
       "    //     .attr(\"cx\", function(d) { return d.x; })\n",
       "    //     .attr(\"cy\", function(d) { return d.y; });\n",
       "};\n",
       "\n",
       "/*Update text positions */\n",
       "function updateText(text) {\n",
       "    text.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "};\n",
       "\n",
       "/*Add drag functionality to the node objects*/\n",
       "node.call(\n",
       "    d3.drag()\n",
       "        .on(\"start\", dragstarted)\n",
       "        .on(\"drag\", dragged)\n",
       "        .on(\"end\", dragended)\n",
       ");\n",
       "\n",
       "function dragstarted(d) {\n",
       "    d3.event.sourceEvent.stopPropagation();\n",
       "    if (!d3.event.active) simulation.alphaTarget(0.3).restart();\n",
       "    d.fx = d.x;\n",
       "    d.fy = d.y;\n",
       "};\n",
       "\n",
       "function dragged(d) {\n",
       "    d.fx = d3.event.x;\n",
       "    d.fy = d3.event.y;\n",
       "};\n",
       "\n",
       "function dragended(d) {\n",
       "    if (!d3.event.active) simulation.alphaTarget(0);\n",
       "    d.fx = null;\n",
       "    d.fy = null;\n",
       "};\n",
       "\n",
       "}); //END\n",
       "\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "m = pp.MultiOrderModel.from_PathData(paths_2, max_order=1)\n",
    "g = m.layers[1]\n",
    "print(g.data.edge_index)\n",
    "print(g.data.edge_weight)\n",
    "pp.plot(g, node_label=g.mapping.node_ids.tolist());"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This is a first-order graph representation, as it only captures the (weighted) edges in the underlying path data, i.e. we could say that we only count the frequency of paths (or walks) of length one. This naturally gives rise to an `edge_index` tensor with shape $(2,m)$, where $m$ is the number of unique edges in the graph that are traversed by the paths.\n",
    "\n",
    "## From Graphs to Higher-Order De Bruijn Graph Models\n",
    "\n",
    "As we have seen above, the use of a first-order graph model discards information in path data, which capture which nodes can possibly causally influence each other via paths. A key feature of `pathpyG` is it allows to generalize this first-order modelling perspective to $k$-th order De Bruijn graph models for paths, where the nodes in a $k$-th order De Bruijn graph model are sequences of $k$ nodes. Edges connect pairs of nodes that overlap in $k-1$ nodes and capture paths of length $k$.\n",
    "\n",
    "A De Bruijn graph of order $k=1$ is simply a normal (weighted) static graph consisting of nodes and edges. Pairs of nodes connected by edges overlap in $k-1=0$ nodes and capture paths of length $k=1$, i.e. simple dyadic edges in the underlying path data.\n",
    "\n",
    "For a De Bruijn graph with order $k=2$, in our example above, an edge connects a pair of nodes $(a,b)$ and $(b,c)$ that overlaps in the $k-1=1$ node $b$. Such an edge represents the path $a -> b -> c$ of length two. We can use the `MultiOderModel` class to generate a second-order De Bruijn graph representation of the path data above. We just have to set the `max_order` parameter to two and use the second layer of the resulting `MultiOrderModel` instance."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<style>\n",
       "svg circle.node {\n",
       "  fill: #3b5998;\n",
       "  stroke: #1b3978;\n",
       "  stroke-width: 2.5px;\n",
       "  r: 15px;\n",
       "  opacity: 1;\n",
       "}\n",
       "\n",
       ".link {\n",
       "  stroke: #969595;\n",
       "  stroke-opacity: .75;\n",
       "  stroke-width: 2.5px;\n",
       "}\n",
       "\n",
       ".arrow {\n",
       "  fill: #969595;\n",
       "}\n",
       "\n",
       "\n",
       ".label-text {\n",
       "    fill: #969595;\n",
       "    font-size: 16px;\n",
       "    font-family: sans-serif;\n",
       "}\n",
       "\n",
       "</style>\n",
       "\n",
       "<div id = \"x9a153f4640c147a89234443f01d2bf62\"> </div>\n",
       "<script charset=\"utf-8\" src=\"https://d3js.org/d3.v5.min.js\"></script>\n",
       "<script charset=\"utf-8\">\n",
       "// Load via requireJS if available (jupyter notebook environment)\n",
       "try {\n",
       "    // Problem: require.config will raise an exception when called for the second time \n",
       "    require.config({\n",
       "        paths: {\n",
       "            d3: \"https://d3js.org/d3.v5.min.js\".replace(\".js\", \"\")\n",
       "        }\n",
       "    });\n",
       "    console.log(\"OKAY: requireJS was detected.\");\n",
       "}\n",
       "catch(err){\n",
       "    // a reference error indicates that requireJS does not exist. \n",
       "    // other errors may occur due to multiple calls to config\n",
       "    if (err instanceof ReferenceError){\n",
       "        console.log(\"WARNING: NO requireJS was detected!\");\n",
       "\n",
       "        // Helper function that waits for d3js to be loaded\n",
       "        require = function require(symbols, callback) {\n",
       "            var ms = 10;\n",
       "            window.setTimeout(function(t) {\n",
       "                if (window[symbols[0]])\n",
       "                    callback(window[symbols[0]]);\n",
       "                else \n",
       "                    window.setTimeout(arguments.callee, ms);\n",
       "            }, ms);\n",
       "        }\n",
       "    }\n",
       "};\n",
       "require(['d3'], function(d3){ //START\n",
       "const data = {\"edges\": [{\"uid\": \"('a', 'c')-('c', 'd')\", \"source\": \"('a', 'c')\", \"target\": \"('c', 'd')\", \"weight\": 4.0, \"size\": 5}, {\"uid\": \"('b', 'c')-('c', 'e')\", \"source\": \"('b', 'c')\", \"target\": \"('c', 'e')\", \"weight\": 4.0, \"size\": 5}], \"nodes\": [{\"uid\": \"('a', 'c')\", \"label\": [\"a\", \"c\"]}, {\"uid\": \"('b', 'c')\", \"label\": [\"b\", \"c\"]}, {\"uid\": \"('c', 'd')\", \"label\": [\"c\", \"d\"]}, {\"uid\": \"('c', 'e')\", \"label\": [\"c\", \"e\"]}]}\n",
       "const config = {\"node_label\": [[\"a\", \"c\"], [\"b\", \"c\"], [\"c\", \"d\"], [\"c\", \"e\"]], \"edge_size\": 5, \"directed\": true, \"curved\": true, \"selector\": \"#x9a153f4640c147a89234443f01d2bf62\"}\n",
       "console.log(\"Static Network Template\");\n",
       "/* Resources\n",
       "   https://bl.ocks.org/mapio/53fed7d84cd1812d6a6639ed7aa83868\n",
       "   https://codepen.io/smlo/pen/JdMOej\n",
       "*/\n",
       "\n",
       "// variables from the config file\n",
       "const selector = config.selector;\n",
       "const width = config.width || 800;\n",
       "const height = config.height || 600;\n",
       "const charge_distance = config.charge_distance || 400;\n",
       "const charge_force = config.charge_force || -3000;\n",
       "const curved = config.curved || false;\n",
       "const directed = config.directed || false;\n",
       "// const weight = false;\n",
       "\n",
       "/* Create a svg element to display the network */\n",
       "var svg = d3.select(selector)\n",
       "    .append('svg')\n",
       "    .attr('width', width)\n",
       "    .attr('height', height)\n",
       "\n",
       "// add container to store the elements\n",
       "var container = svg.append(\"g\");\n",
       "\n",
       "/*Add zoom function to the container */\n",
       "svg.call(\n",
       "    d3.zoom()\n",
       "        .scaleExtent([.1, 4])\n",
       "        .on(\"zoom\", function() { container.attr(\"transform\", d3.event.transform); })\n",
       ");\n",
       "\n",
       "\n",
       "/*Load nodes and links from the data */\n",
       "var nodes = data.nodes\n",
       "var links = data.edges\n",
       "\n",
       "/*Create arrow head with same color as the edge */\n",
       "function marker (color) {\n",
       "       var reference;\n",
       "       svg.append(\"svg:defs\").selectAll(\"marker\")\n",
       "          .data([reference])\n",
       "          .enter().append(\"svg:marker\")\n",
       "          .attr(\"id\", \"arrow\"+color)\n",
       "          .attr(\"viewBox\", \"0 -5 10 10\")\n",
       "          .attr(\"refX\", 10)\n",
       "          .attr(\"refY\", -0)\n",
       "          .attr(\"markerWidth\", 6)\n",
       "          .attr(\"markerHeight\", 6)\n",
       "          .attr(\"orient\", \"auto\")\n",
       "          .append(\"svg:path\")\n",
       "          .attr('class','.arrow')\n",
       "          .attr(\"d\", \"M0,-5L10,0L0,5\")\n",
       "          .style('opacity',1)\n",
       "          .style(\"fill\", color);\n",
       "       return \"url(#\" + \"arrow\"+color + \")\";\n",
       "     };\n",
       "\n",
       "/*Link creation template */\n",
       "var link = container.append(\"g\").attr(\"class\", \"links\")\n",
       "    .selectAll(\".link\")\n",
       "    .data(links)\n",
       "    .enter()\n",
       "    .append(\"path\")\n",
       "    .attr(\"class\", \"link\")\n",
       "    .style(\"stroke\", function(d) { return d.color; })\n",
       "    .style(\"stroke-opacity\", function(d) { return d.opacity; })\n",
       "    .style(\"stroke-width\", function(d){  return d.size })\n",
       "    .style(\"fill\",\"none\")\n",
       "    .attr(\"marker-end\", function (d) {if(directed){return marker(d.color)}else{return null}; });\n",
       "\n",
       "    //.attr(\"marker-end\", function (d) { return marker(d.color); });\n",
       "    //.attr(\"marker-end\", \"url(#arrow)\");\n",
       "\n",
       "/*Node creation template */\n",
       "var node = container.append(\"g\").attr(\"class\", \"nodes\")\n",
       "    .selectAll(\"circle.node\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"circle\")\n",
       "    .attr(\"class\", \"node\")\n",
       "    .attr(\"x\", function(d) { return d.x; })\n",
       "    .attr(\"y\", function(d) { return d.y; })\n",
       "    .style(\"r\", function(d){  return d.size+\"px\"; })\n",
       "    .style(\"fill\", function(d) { return d.color; })\n",
       "    .style(\"opacity\", function(d) { return d.opacity; });\n",
       "\n",
       "/*Label creation template */\n",
       "var text = container.append(\"g\").attr(\"class\",\"labels\")\n",
       "    .selectAll(\"g\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"g\")\n",
       "\n",
       "text.append(\"text\")\n",
       "    .attr(\"class\", \"label-text\")\n",
       "    .attr(\"x\", function(d) {\n",
       "        var r = (d.size === undefined) ? 15 : d.size;\n",
       "        return 5 + r; })\n",
       "    .attr(\"dy\", \".31em\")\n",
       "    .text(function(d) { return d.label; });\n",
       "\n",
       "/*Scale weight for d3js */\n",
       "var weightScale = d3.scaleLinear()\n",
       "    .domain(d3.extent(links, function (d) { return d.weight }))\n",
       "    .range([.1, 1]);\n",
       "\n",
       "/*Simulation of the forces*/\n",
       "var simulation = d3.forceSimulation(nodes)\n",
       "    .force(\"links\", d3.forceLink(links)\n",
       "           .id(function(d) {return d.uid; })\n",
       "           .distance(50)\n",
       "           .strength(function(d){return weightScale(d.weight);})\n",
       "          )\n",
       "    .force(\"charge\", d3.forceManyBody()\n",
       "           .strength(charge_force)\n",
       "           .distanceMax(charge_distance)\n",
       "          )\n",
       "    .force(\"center\", d3.forceCenter(width / 2, height / 2))\n",
       "    .on(\"tick\", ticked);\n",
       "\n",
       "/*Update of the node and edge objects*/\n",
       "function ticked() {\n",
       "    node.call(updateNode);\n",
       "    link.call(updateLink);\n",
       "    text.call(updateText);\n",
       "};\n",
       "\n",
       "/*Update link positions */\n",
       "function updateLink(link) {\n",
       "    // link\n",
       "    //     .attr(\"x1\", function(d) { return d.source.x; })\n",
       "    //     .attr(\"y1\", function(d) { return d.source.y; })\n",
       "    //     .attr(\"x2\", function(d) { return d.target.x; })\n",
       "    //     .attr(\"y2\", function(d) { return d.target.y; });\n",
       "\n",
       "\n",
       "    link.attr(\"d\", function(d) {\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        return \"M\" +\n",
       "            d.source.x + \",\" +\n",
       "            d.source.y + \"A\" +\n",
       "            dr + \",\" + dr + \" 0 0,1 \" +\n",
       "            d.target.x + \",\" +\n",
       "            d.target.y;\n",
       "    });\n",
       "\n",
       "    // recalculate and back off the distance\n",
       "    link.attr(\"d\", function (d, i) {\n",
       "        var pl = this.getTotalLength();\n",
       "        var r = (d.target.size === undefined) ? 15 : d.target.size;\n",
       "        var m = this.getPointAtLength(pl - r);\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        var result = \"M\" + d.source.x + \",\" + d.source.y + \"A\" + dr + \",\" + dr + \" 0 0,1 \" + m.x + \",\" + m.y;\n",
       "        return result;\n",
       "    });\n",
       "};\n",
       "\n",
       "\n",
       "/*Update node positions */\n",
       "function updateNode(node) {\n",
       "    node.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "    // node\n",
       "    //     .attr(\"cx\", function(d) { return d.x; })\n",
       "    //     .attr(\"cy\", function(d) { return d.y; });\n",
       "};\n",
       "\n",
       "/*Update text positions */\n",
       "function updateText(text) {\n",
       "    text.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "};\n",
       "\n",
       "/*Add drag functionality to the node objects*/\n",
       "node.call(\n",
       "    d3.drag()\n",
       "        .on(\"start\", dragstarted)\n",
       "        .on(\"drag\", dragged)\n",
       "        .on(\"end\", dragended)\n",
       ");\n",
       "\n",
       "function dragstarted(d) {\n",
       "    d3.event.sourceEvent.stopPropagation();\n",
       "    if (!d3.event.active) simulation.alphaTarget(0.3).restart();\n",
       "    d.fx = d.x;\n",
       "    d.fy = d.y;\n",
       "};\n",
       "\n",
       "function dragged(d) {\n",
       "    d.fx = d3.event.x;\n",
       "    d.fy = d3.event.y;\n",
       "};\n",
       "\n",
       "function dragended(d) {\n",
       "    if (!d3.event.active) simulation.alphaTarget(0);\n",
       "    d.fx = null;\n",
       "    d.fy = null;\n",
       "};\n",
       "\n",
       "}); //END\n",
       "\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "m = pp.MultiOrderModel.from_PathData(paths, max_order=2)\n",
    "g = m.layers[2]\n",
    "pp.plot(g, node_label=g.nodes, edge_size=5);"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "For $k=2$, we obtain a second-order De Bruijn graph where second-order nodes are first-order edges and second-order edges represent walks of length two in the original graph. Edge weights capture observation frequencies of those walks. In our example, we have two different walks of length two ($a$ -> $c$ -> $d$ and $b$ -> $c$ -> $e$), represented by two edges $(a-c, c-d)$ and $(b-c, c-e)$. Each of those walks appears four times so the weights of both edges are four."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "('a', 'c') -> 0\n",
      "('b', 'c') -> 1\n",
      "('c', 'd') -> 2\n",
      "('c', 'e') -> 3\n",
      "\n",
      "EdgeIndex([[0, 1],\n",
      "           [2, 3]], sparse_size=(4, 4), nnz=2, sort_order=row)\n",
      "tensor([4., 4.])\n"
     ]
    }
   ],
   "source": [
    "print(g.mapping)\n",
    "print(g.data.edge_index)\n",
    "print(g.data.edge_weight)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "While this goes beyond the scope of this tutorial, thanks to the tensor-based representation of paths, the construction ofhigher-order De Bruijn graphs can be done based on efficient GPU operations, i.e. we can scale it up to large graphs.\n",
    "\n",
    "Let us have a closer look at our examples above. While the first-order edge indices of the two path objects `paths` and `paths_2` are the same, we find that the second-order edge indices are actually different. For `paths_2` we have four different paths of length two, each occurring twice. Hence, our second-order De Bruijn graph has four edges, each with weight two. These edges correspond to all possible paths of length two in the underlying graph."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<style>\n",
       "svg circle.node {\n",
       "  fill: #3b5998;\n",
       "  stroke: #1b3978;\n",
       "  stroke-width: 2.5px;\n",
       "  r: 15px;\n",
       "  opacity: 1;\n",
       "}\n",
       "\n",
       ".link {\n",
       "  stroke: #969595;\n",
       "  stroke-opacity: .75;\n",
       "  stroke-width: 2.5px;\n",
       "}\n",
       "\n",
       ".arrow {\n",
       "  fill: #969595;\n",
       "}\n",
       "\n",
       "\n",
       ".label-text {\n",
       "    fill: #969595;\n",
       "    font-size: 16px;\n",
       "    font-family: sans-serif;\n",
       "}\n",
       "\n",
       "</style>\n",
       "\n",
       "<div id = \"x93db0fa030e7457ab0de435197fc76ee\"> </div>\n",
       "<script charset=\"utf-8\" src=\"https://d3js.org/d3.v5.min.js\"></script>\n",
       "<script charset=\"utf-8\">\n",
       "// Load via requireJS if available (jupyter notebook environment)\n",
       "try {\n",
       "    // Problem: require.config will raise an exception when called for the second time \n",
       "    require.config({\n",
       "        paths: {\n",
       "            d3: \"https://d3js.org/d3.v5.min.js\".replace(\".js\", \"\")\n",
       "        }\n",
       "    });\n",
       "    console.log(\"OKAY: requireJS was detected.\");\n",
       "}\n",
       "catch(err){\n",
       "    // a reference error indicates that requireJS does not exist. \n",
       "    // other errors may occur due to multiple calls to config\n",
       "    if (err instanceof ReferenceError){\n",
       "        console.log(\"WARNING: NO requireJS was detected!\");\n",
       "\n",
       "        // Helper function that waits for d3js to be loaded\n",
       "        require = function require(symbols, callback) {\n",
       "            var ms = 10;\n",
       "            window.setTimeout(function(t) {\n",
       "                if (window[symbols[0]])\n",
       "                    callback(window[symbols[0]]);\n",
       "                else \n",
       "                    window.setTimeout(arguments.callee, ms);\n",
       "            }, ms);\n",
       "        }\n",
       "    }\n",
       "};\n",
       "require(['d3'], function(d3){ //START\n",
       "const data = {\"edges\": [{\"uid\": \"('a', 'c')-('c', 'd')\", \"source\": \"('a', 'c')\", \"target\": \"('c', 'd')\", \"weight\": 2.0}, {\"uid\": \"('a', 'c')-('c', 'e')\", \"source\": \"('a', 'c')\", \"target\": \"('c', 'e')\", \"weight\": 2.0}, {\"uid\": \"('b', 'c')-('c', 'd')\", \"source\": \"('b', 'c')\", \"target\": \"('c', 'd')\", \"weight\": 2.0}, {\"uid\": \"('b', 'c')-('c', 'e')\", \"source\": \"('b', 'c')\", \"target\": \"('c', 'e')\", \"weight\": 2.0}], \"nodes\": [{\"uid\": \"('a', 'c')\", \"label\": [\"a\", \"c\"]}, {\"uid\": \"('b', 'c')\", \"label\": [\"b\", \"c\"]}, {\"uid\": \"('c', 'd')\", \"label\": [\"c\", \"d\"]}, {\"uid\": \"('c', 'e')\", \"label\": [\"c\", \"e\"]}]}\n",
       "const config = {\"node_label\": [[\"a\", \"c\"], [\"b\", \"c\"], [\"c\", \"d\"], [\"c\", \"e\"]], \"directed\": true, \"curved\": true, \"selector\": \"#x93db0fa030e7457ab0de435197fc76ee\"}\n",
       "console.log(\"Static Network Template\");\n",
       "/* Resources\n",
       "   https://bl.ocks.org/mapio/53fed7d84cd1812d6a6639ed7aa83868\n",
       "   https://codepen.io/smlo/pen/JdMOej\n",
       "*/\n",
       "\n",
       "// variables from the config file\n",
       "const selector = config.selector;\n",
       "const width = config.width || 800;\n",
       "const height = config.height || 600;\n",
       "const charge_distance = config.charge_distance || 400;\n",
       "const charge_force = config.charge_force || -3000;\n",
       "const curved = config.curved || false;\n",
       "const directed = config.directed || false;\n",
       "// const weight = false;\n",
       "\n",
       "/* Create a svg element to display the network */\n",
       "var svg = d3.select(selector)\n",
       "    .append('svg')\n",
       "    .attr('width', width)\n",
       "    .attr('height', height)\n",
       "\n",
       "// add container to store the elements\n",
       "var container = svg.append(\"g\");\n",
       "\n",
       "/*Add zoom function to the container */\n",
       "svg.call(\n",
       "    d3.zoom()\n",
       "        .scaleExtent([.1, 4])\n",
       "        .on(\"zoom\", function() { container.attr(\"transform\", d3.event.transform); })\n",
       ");\n",
       "\n",
       "\n",
       "/*Load nodes and links from the data */\n",
       "var nodes = data.nodes\n",
       "var links = data.edges\n",
       "\n",
       "/*Create arrow head with same color as the edge */\n",
       "function marker (color) {\n",
       "       var reference;\n",
       "       svg.append(\"svg:defs\").selectAll(\"marker\")\n",
       "          .data([reference])\n",
       "          .enter().append(\"svg:marker\")\n",
       "          .attr(\"id\", \"arrow\"+color)\n",
       "          .attr(\"viewBox\", \"0 -5 10 10\")\n",
       "          .attr(\"refX\", 10)\n",
       "          .attr(\"refY\", -0)\n",
       "          .attr(\"markerWidth\", 6)\n",
       "          .attr(\"markerHeight\", 6)\n",
       "          .attr(\"orient\", \"auto\")\n",
       "          .append(\"svg:path\")\n",
       "          .attr('class','.arrow')\n",
       "          .attr(\"d\", \"M0,-5L10,0L0,5\")\n",
       "          .style('opacity',1)\n",
       "          .style(\"fill\", color);\n",
       "       return \"url(#\" + \"arrow\"+color + \")\";\n",
       "     };\n",
       "\n",
       "/*Link creation template */\n",
       "var link = container.append(\"g\").attr(\"class\", \"links\")\n",
       "    .selectAll(\".link\")\n",
       "    .data(links)\n",
       "    .enter()\n",
       "    .append(\"path\")\n",
       "    .attr(\"class\", \"link\")\n",
       "    .style(\"stroke\", function(d) { return d.color; })\n",
       "    .style(\"stroke-opacity\", function(d) { return d.opacity; })\n",
       "    .style(\"stroke-width\", function(d){  return d.size })\n",
       "    .style(\"fill\",\"none\")\n",
       "    .attr(\"marker-end\", function (d) {if(directed){return marker(d.color)}else{return null}; });\n",
       "\n",
       "    //.attr(\"marker-end\", function (d) { return marker(d.color); });\n",
       "    //.attr(\"marker-end\", \"url(#arrow)\");\n",
       "\n",
       "/*Node creation template */\n",
       "var node = container.append(\"g\").attr(\"class\", \"nodes\")\n",
       "    .selectAll(\"circle.node\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"circle\")\n",
       "    .attr(\"class\", \"node\")\n",
       "    .attr(\"x\", function(d) { return d.x; })\n",
       "    .attr(\"y\", function(d) { return d.y; })\n",
       "    .style(\"r\", function(d){  return d.size+\"px\"; })\n",
       "    .style(\"fill\", function(d) { return d.color; })\n",
       "    .style(\"opacity\", function(d) { return d.opacity; });\n",
       "\n",
       "/*Label creation template */\n",
       "var text = container.append(\"g\").attr(\"class\",\"labels\")\n",
       "    .selectAll(\"g\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"g\")\n",
       "\n",
       "text.append(\"text\")\n",
       "    .attr(\"class\", \"label-text\")\n",
       "    .attr(\"x\", function(d) {\n",
       "        var r = (d.size === undefined) ? 15 : d.size;\n",
       "        return 5 + r; })\n",
       "    .attr(\"dy\", \".31em\")\n",
       "    .text(function(d) { return d.label; });\n",
       "\n",
       "/*Scale weight for d3js */\n",
       "var weightScale = d3.scaleLinear()\n",
       "    .domain(d3.extent(links, function (d) { return d.weight }))\n",
       "    .range([.1, 1]);\n",
       "\n",
       "/*Simulation of the forces*/\n",
       "var simulation = d3.forceSimulation(nodes)\n",
       "    .force(\"links\", d3.forceLink(links)\n",
       "           .id(function(d) {return d.uid; })\n",
       "           .distance(50)\n",
       "           .strength(function(d){return weightScale(d.weight);})\n",
       "          )\n",
       "    .force(\"charge\", d3.forceManyBody()\n",
       "           .strength(charge_force)\n",
       "           .distanceMax(charge_distance)\n",
       "          )\n",
       "    .force(\"center\", d3.forceCenter(width / 2, height / 2))\n",
       "    .on(\"tick\", ticked);\n",
       "\n",
       "/*Update of the node and edge objects*/\n",
       "function ticked() {\n",
       "    node.call(updateNode);\n",
       "    link.call(updateLink);\n",
       "    text.call(updateText);\n",
       "};\n",
       "\n",
       "/*Update link positions */\n",
       "function updateLink(link) {\n",
       "    // link\n",
       "    //     .attr(\"x1\", function(d) { return d.source.x; })\n",
       "    //     .attr(\"y1\", function(d) { return d.source.y; })\n",
       "    //     .attr(\"x2\", function(d) { return d.target.x; })\n",
       "    //     .attr(\"y2\", function(d) { return d.target.y; });\n",
       "\n",
       "\n",
       "    link.attr(\"d\", function(d) {\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        return \"M\" +\n",
       "            d.source.x + \",\" +\n",
       "            d.source.y + \"A\" +\n",
       "            dr + \",\" + dr + \" 0 0,1 \" +\n",
       "            d.target.x + \",\" +\n",
       "            d.target.y;\n",
       "    });\n",
       "\n",
       "    // recalculate and back off the distance\n",
       "    link.attr(\"d\", function (d, i) {\n",
       "        var pl = this.getTotalLength();\n",
       "        var r = (d.target.size === undefined) ? 15 : d.target.size;\n",
       "        var m = this.getPointAtLength(pl - r);\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        var result = \"M\" + d.source.x + \",\" + d.source.y + \"A\" + dr + \",\" + dr + \" 0 0,1 \" + m.x + \",\" + m.y;\n",
       "        return result;\n",
       "    });\n",
       "};\n",
       "\n",
       "\n",
       "/*Update node positions */\n",
       "function updateNode(node) {\n",
       "    node.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "    // node\n",
       "    //     .attr(\"cx\", function(d) { return d.x; })\n",
       "    //     .attr(\"cy\", function(d) { return d.y; });\n",
       "};\n",
       "\n",
       "/*Update text positions */\n",
       "function updateText(text) {\n",
       "    text.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "};\n",
       "\n",
       "/*Add drag functionality to the node objects*/\n",
       "node.call(\n",
       "    d3.drag()\n",
       "        .on(\"start\", dragstarted)\n",
       "        .on(\"drag\", dragged)\n",
       "        .on(\"end\", dragended)\n",
       ");\n",
       "\n",
       "function dragstarted(d) {\n",
       "    d3.event.sourceEvent.stopPropagation();\n",
       "    if (!d3.event.active) simulation.alphaTarget(0.3).restart();\n",
       "    d.fx = d.x;\n",
       "    d.fy = d.y;\n",
       "};\n",
       "\n",
       "function dragged(d) {\n",
       "    d.fx = d3.event.x;\n",
       "    d.fy = d3.event.y;\n",
       "};\n",
       "\n",
       "function dragended(d) {\n",
       "    if (!d3.event.active) simulation.alphaTarget(0);\n",
       "    d.fx = null;\n",
       "    d.fy = null;\n",
       "};\n",
       "\n",
       "}); //END\n",
       "\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "m = pp.MultiOrderModel.from_PathData(paths_2, max_order=2)\n",
    "g = m.layers[2]\n",
    "pp.plot(g, node_label=g.nodes);"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "('a', 'c') -> 0\n",
      "('b', 'c') -> 1\n",
      "('c', 'd') -> 2\n",
      "('c', 'e') -> 3\n",
      "\n",
      "EdgeIndex([[0, 0, 1, 1],\n",
      "           [2, 3, 2, 3]], sparse_size=(4, 4), nnz=4, sort_order=row)\n",
      "tensor([2., 2., 2., 2.])\n"
     ]
    }
   ],
   "source": [
    "print(g.mapping)\n",
    "print(g.data.edge_index)\n",
    "print(g.data.edge_weight)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We thus find that the second-order De Bruijn graph representation of paths is sensitive to the differences in the causal topology, while a first-order graph is not. This is the basis to generalize network analysis and graph learning to causality-aware graph models for various kinds of time series data on graphs. In particular, as we shall see in more detail in a later tutorial, we can use paths to generate k-th order graphs that can be used to generalize Graph Neural Networks to higher-order De Bruijn Graphs."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Note that all higher-order graphs are simply `Graph` objects, which means that we can iterate through the nodes of a higher-order graph just like for normal graphs. Node indices are automatically mapped, yielding tuples of first-order node identifiers."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "('a', 'c')\n",
      "('b', 'c')\n",
      "('c', 'd')\n",
      "('c', 'e')\n"
     ]
    }
   ],
   "source": [
    "for n in g.nodes:\n",
    "    print(n)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Edges are tuples with two elements, where each element is a k-th order node, i.e. a tuple of node IDs of length $k$. I.e. for a second-order model the edges are tuples of length two, each entry containing s tuple of length two."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "(('a', 'c'), ('c', 'd'))\n",
      "(('a', 'c'), ('c', 'e'))\n",
      "(('b', 'c'), ('c', 'd'))\n",
      "(('b', 'c'), ('c', 'e'))\n"
     ]
    }
   ],
   "source": [
    "for e in g.edges:\n",
    "    print(e)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The weight attribute stores a tensor whose entries capture the frequencies of edges, i.e. the frequencies of paths of length $k$."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "(('a', 'c'), ('c', 'd')) 2.0\n",
      "(('a', 'c'), ('c', 'e')) 2.0\n",
      "(('b', 'c'), ('c', 'd')) 2.0\n",
      "(('b', 'c'), ('c', 'e')) 2.0\n"
     ]
    }
   ],
   "source": [
    "for e in g.edges:\n",
    "    print(e, g['edge_weight', e[0], e[1]].item())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can finally plot a higher-order De Bruijn graph in the same way as a first-order graph."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<style>\n",
       "svg circle.node {\n",
       "  fill: #3b5998;\n",
       "  stroke: #1b3978;\n",
       "  stroke-width: 2.5px;\n",
       "  r: 15px;\n",
       "  opacity: 1;\n",
       "}\n",
       "\n",
       ".link {\n",
       "  stroke: #969595;\n",
       "  stroke-opacity: .75;\n",
       "  stroke-width: 2.5px;\n",
       "}\n",
       "\n",
       ".arrow {\n",
       "  fill: #969595;\n",
       "}\n",
       "\n",
       "\n",
       ".label-text {\n",
       "    fill: #969595;\n",
       "    font-size: 16px;\n",
       "    font-family: sans-serif;\n",
       "}\n",
       "\n",
       "</style>\n",
       "\n",
       "<div id = \"xbb0b9952709d4b59af333d1e4f67b3fc\"> </div>\n",
       "<script charset=\"utf-8\" src=\"https://d3js.org/d3.v5.min.js\"></script>\n",
       "<script charset=\"utf-8\">\n",
       "// Load via requireJS if available (jupyter notebook environment)\n",
       "try {\n",
       "    // Problem: require.config will raise an exception when called for the second time \n",
       "    require.config({\n",
       "        paths: {\n",
       "            d3: \"https://d3js.org/d3.v5.min.js\".replace(\".js\", \"\")\n",
       "        }\n",
       "    });\n",
       "    console.log(\"OKAY: requireJS was detected.\");\n",
       "}\n",
       "catch(err){\n",
       "    // a reference error indicates that requireJS does not exist. \n",
       "    // other errors may occur due to multiple calls to config\n",
       "    if (err instanceof ReferenceError){\n",
       "        console.log(\"WARNING: NO requireJS was detected!\");\n",
       "\n",
       "        // Helper function that waits for d3js to be loaded\n",
       "        require = function require(symbols, callback) {\n",
       "            var ms = 10;\n",
       "            window.setTimeout(function(t) {\n",
       "                if (window[symbols[0]])\n",
       "                    callback(window[symbols[0]]);\n",
       "                else \n",
       "                    window.setTimeout(arguments.callee, ms);\n",
       "            }, ms);\n",
       "        }\n",
       "    }\n",
       "};\n",
       "require(['d3'], function(d3){ //START\n",
       "const data = {\"edges\": [{\"uid\": \"('a', 'c')-('c', 'd')\", \"source\": \"('a', 'c')\", \"target\": \"('c', 'd')\", \"weight\": 2.0, \"color\": \"gray\"}, {\"uid\": \"('a', 'c')-('c', 'e')\", \"source\": \"('a', 'c')\", \"target\": \"('c', 'e')\", \"weight\": 2.0, \"color\": \"gray\"}, {\"uid\": \"('b', 'c')-('c', 'd')\", \"source\": \"('b', 'c')\", \"target\": \"('c', 'd')\", \"weight\": 2.0, \"color\": \"gray\"}, {\"uid\": \"('b', 'c')-('c', 'e')\", \"source\": \"('b', 'c')\", \"target\": \"('c', 'e')\", \"weight\": 2.0, \"color\": \"gray\"}], \"nodes\": [{\"uid\": \"('a', 'c')\", \"label\": [\"a\", \"c\"]}, {\"uid\": \"('b', 'c')\", \"label\": [\"b\", \"c\"]}, {\"uid\": \"('c', 'd')\", \"label\": [\"c\", \"d\"]}, {\"uid\": \"('c', 'e')\", \"label\": [\"c\", \"e\"]}]}\n",
       "const config = {\"node_label\": [[\"a\", \"c\"], [\"b\", \"c\"], [\"c\", \"d\"], [\"c\", \"e\"]], \"edge_color\": \"gray\", \"directed\": true, \"curved\": true, \"selector\": \"#xbb0b9952709d4b59af333d1e4f67b3fc\"}\n",
       "console.log(\"Static Network Template\");\n",
       "/* Resources\n",
       "   https://bl.ocks.org/mapio/53fed7d84cd1812d6a6639ed7aa83868\n",
       "   https://codepen.io/smlo/pen/JdMOej\n",
       "*/\n",
       "\n",
       "// variables from the config file\n",
       "const selector = config.selector;\n",
       "const width = config.width || 800;\n",
       "const height = config.height || 600;\n",
       "const charge_distance = config.charge_distance || 400;\n",
       "const charge_force = config.charge_force || -3000;\n",
       "const curved = config.curved || false;\n",
       "const directed = config.directed || false;\n",
       "// const weight = false;\n",
       "\n",
       "/* Create a svg element to display the network */\n",
       "var svg = d3.select(selector)\n",
       "    .append('svg')\n",
       "    .attr('width', width)\n",
       "    .attr('height', height)\n",
       "\n",
       "// add container to store the elements\n",
       "var container = svg.append(\"g\");\n",
       "\n",
       "/*Add zoom function to the container */\n",
       "svg.call(\n",
       "    d3.zoom()\n",
       "        .scaleExtent([.1, 4])\n",
       "        .on(\"zoom\", function() { container.attr(\"transform\", d3.event.transform); })\n",
       ");\n",
       "\n",
       "\n",
       "/*Load nodes and links from the data */\n",
       "var nodes = data.nodes\n",
       "var links = data.edges\n",
       "\n",
       "/*Create arrow head with same color as the edge */\n",
       "function marker (color) {\n",
       "       var reference;\n",
       "       svg.append(\"svg:defs\").selectAll(\"marker\")\n",
       "          .data([reference])\n",
       "          .enter().append(\"svg:marker\")\n",
       "          .attr(\"id\", \"arrow\"+color)\n",
       "          .attr(\"viewBox\", \"0 -5 10 10\")\n",
       "          .attr(\"refX\", 10)\n",
       "          .attr(\"refY\", -0)\n",
       "          .attr(\"markerWidth\", 6)\n",
       "          .attr(\"markerHeight\", 6)\n",
       "          .attr(\"orient\", \"auto\")\n",
       "          .append(\"svg:path\")\n",
       "          .attr('class','.arrow')\n",
       "          .attr(\"d\", \"M0,-5L10,0L0,5\")\n",
       "          .style('opacity',1)\n",
       "          .style(\"fill\", color);\n",
       "       return \"url(#\" + \"arrow\"+color + \")\";\n",
       "     };\n",
       "\n",
       "/*Link creation template */\n",
       "var link = container.append(\"g\").attr(\"class\", \"links\")\n",
       "    .selectAll(\".link\")\n",
       "    .data(links)\n",
       "    .enter()\n",
       "    .append(\"path\")\n",
       "    .attr(\"class\", \"link\")\n",
       "    .style(\"stroke\", function(d) { return d.color; })\n",
       "    .style(\"stroke-opacity\", function(d) { return d.opacity; })\n",
       "    .style(\"stroke-width\", function(d){  return d.size })\n",
       "    .style(\"fill\",\"none\")\n",
       "    .attr(\"marker-end\", function (d) {if(directed){return marker(d.color)}else{return null}; });\n",
       "\n",
       "    //.attr(\"marker-end\", function (d) { return marker(d.color); });\n",
       "    //.attr(\"marker-end\", \"url(#arrow)\");\n",
       "\n",
       "/*Node creation template */\n",
       "var node = container.append(\"g\").attr(\"class\", \"nodes\")\n",
       "    .selectAll(\"circle.node\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"circle\")\n",
       "    .attr(\"class\", \"node\")\n",
       "    .attr(\"x\", function(d) { return d.x; })\n",
       "    .attr(\"y\", function(d) { return d.y; })\n",
       "    .style(\"r\", function(d){  return d.size+\"px\"; })\n",
       "    .style(\"fill\", function(d) { return d.color; })\n",
       "    .style(\"opacity\", function(d) { return d.opacity; });\n",
       "\n",
       "/*Label creation template */\n",
       "var text = container.append(\"g\").attr(\"class\",\"labels\")\n",
       "    .selectAll(\"g\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"g\")\n",
       "\n",
       "text.append(\"text\")\n",
       "    .attr(\"class\", \"label-text\")\n",
       "    .attr(\"x\", function(d) {\n",
       "        var r = (d.size === undefined) ? 15 : d.size;\n",
       "        return 5 + r; })\n",
       "    .attr(\"dy\", \".31em\")\n",
       "    .text(function(d) { return d.label; });\n",
       "\n",
       "/*Scale weight for d3js */\n",
       "var weightScale = d3.scaleLinear()\n",
       "    .domain(d3.extent(links, function (d) { return d.weight }))\n",
       "    .range([.1, 1]);\n",
       "\n",
       "/*Simulation of the forces*/\n",
       "var simulation = d3.forceSimulation(nodes)\n",
       "    .force(\"links\", d3.forceLink(links)\n",
       "           .id(function(d) {return d.uid; })\n",
       "           .distance(50)\n",
       "           .strength(function(d){return weightScale(d.weight);})\n",
       "          )\n",
       "    .force(\"charge\", d3.forceManyBody()\n",
       "           .strength(charge_force)\n",
       "           .distanceMax(charge_distance)\n",
       "          )\n",
       "    .force(\"center\", d3.forceCenter(width / 2, height / 2))\n",
       "    .on(\"tick\", ticked);\n",
       "\n",
       "/*Update of the node and edge objects*/\n",
       "function ticked() {\n",
       "    node.call(updateNode);\n",
       "    link.call(updateLink);\n",
       "    text.call(updateText);\n",
       "};\n",
       "\n",
       "/*Update link positions */\n",
       "function updateLink(link) {\n",
       "    // link\n",
       "    //     .attr(\"x1\", function(d) { return d.source.x; })\n",
       "    //     .attr(\"y1\", function(d) { return d.source.y; })\n",
       "    //     .attr(\"x2\", function(d) { return d.target.x; })\n",
       "    //     .attr(\"y2\", function(d) { return d.target.y; });\n",
       "\n",
       "\n",
       "    link.attr(\"d\", function(d) {\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        return \"M\" +\n",
       "            d.source.x + \",\" +\n",
       "            d.source.y + \"A\" +\n",
       "            dr + \",\" + dr + \" 0 0,1 \" +\n",
       "            d.target.x + \",\" +\n",
       "            d.target.y;\n",
       "    });\n",
       "\n",
       "    // recalculate and back off the distance\n",
       "    link.attr(\"d\", function (d, i) {\n",
       "        var pl = this.getTotalLength();\n",
       "        var r = (d.target.size === undefined) ? 15 : d.target.size;\n",
       "        var m = this.getPointAtLength(pl - r);\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        var result = \"M\" + d.source.x + \",\" + d.source.y + \"A\" + dr + \",\" + dr + \" 0 0,1 \" + m.x + \",\" + m.y;\n",
       "        return result;\n",
       "    });\n",
       "};\n",
       "\n",
       "\n",
       "/*Update node positions */\n",
       "function updateNode(node) {\n",
       "    node.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "    // node\n",
       "    //     .attr(\"cx\", function(d) { return d.x; })\n",
       "    //     .attr(\"cy\", function(d) { return d.y; });\n",
       "};\n",
       "\n",
       "/*Update text positions */\n",
       "function updateText(text) {\n",
       "    text.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "};\n",
       "\n",
       "/*Add drag functionality to the node objects*/\n",
       "node.call(\n",
       "    d3.drag()\n",
       "        .on(\"start\", dragstarted)\n",
       "        .on(\"drag\", dragged)\n",
       "        .on(\"end\", dragended)\n",
       ");\n",
       "\n",
       "function dragstarted(d) {\n",
       "    d3.event.sourceEvent.stopPropagation();\n",
       "    if (!d3.event.active) simulation.alphaTarget(0.3).restart();\n",
       "    d.fx = d.x;\n",
       "    d.fy = d.y;\n",
       "};\n",
       "\n",
       "function dragged(d) {\n",
       "    d.fx = d3.event.x;\n",
       "    d.fy = d3.event.y;\n",
       "};\n",
       "\n",
       "function dragended(d) {\n",
       "    if (!d3.event.active) simulation.alphaTarget(0);\n",
       "    d.fx = null;\n",
       "    d.fy = null;\n",
       "};\n",
       "\n",
       "}); //END\n",
       "\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "pp.plot(g, node_label=g.nodes, edge_color='gray');"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Let us compare this to a second-order graph model of the second path data set `paths_2` from above, which corresponds to a network where all possible paths of length two actually occur. Hence, different from the data in `paths`, all pairs of nodes in this graph can causally influence each other via paths of length two."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Detecting the Optimal Order of Higher-Order De Bruijn Graph Models\n",
    "\n",
    "The fact that we can model the same set of paths with higher-order De Bruijn graph models with different orders $k$ raises an important question: What is the **optimal order** to model a given `PathData` instance. It is actually easy to answer this question in our example above. \n",
    "\n",
    "For the data contained in `paths`, we observe only two of the four possible paths, which is different from what we would expect based on a first-order graph model. To capture this pattern in the node sequenves, a first-order graph model is not sufficient and we need a second-order De Bruoijn graph model.\n",
    "\n",
    "For `paths_2` this is different: Here we observe all four paths of length two with the same frequency, which is exactly what we would expect based on the first-order weighted graph, where all edge weights are the same. Hence, for `paths_2` the second-order De Bruijn graph model contains no additional information compared to a first-order weighted graph, which means a first-order model is sufficient. \n",
    "\n",
    "While it is easy to see this in the toy example, for real data we need a principled method to automatically determine the optimal order of a higher-order De Bruijn graph model. Luckily, this can be achieved based on statistical model selection in a multi-order De Bruijn graph model. Here we cannot explain the details of this method, so we kindly refer you to the following paper: \n",
    "\n",
    "[I Scholtes: When is a Network a Network?: Multi-Order Graphical Model Selection in Pathways and Temporal Networks, In Proc. of SIGKDD 2017, August 2017](https://dl.acm.org/doi/10.1145/3097983.3098145)\n",
    "\n",
    "The method introduced in this paper is implemented in `pathpyG`. To determine the optimal order of a higher-order De Bruijn graph model for the node sequences contained in a given `PathData` instance, we can use the `MultiOderModel.estimate_order` method. Since the method is based on statistical hypothesis testing, we can also pass a significance threshold, which - in line with the interpretation of p-values - bounds the type I error rate of our test, i.e. the rate at which we wrongly reject the null hypothesis that the true optimal order of a data set is $k-1$ in favor of the alternative hypothesis that the order is $k$.\n",
    "\n",
    "Let us test this for our toy example. Using a significance threshold of $0.01$, we determine the optimal order for the data set `paths` that should actually warrant a second-order model:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "2\n"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "/opt/conda/lib/python3.10/site-packages/torch_geometric/edge_index.py:784: UserWarning: Sparse CSR tensor support is in beta state. If you miss a functionality in the sparse tensor support, please submit a feature request to https://github.com/pytorch/pytorch/issues. (Triggered internally at ../aten/src/ATen/SparseCsrTensorImpl.cpp:53.)\n",
      "  return torch.sparse_csr_tensor(\n"
     ]
    }
   ],
   "source": [
    "m1 = pp.MultiOrderModel.from_PathData(paths, max_order=2)\n",
    "print(m1.estimate_order(paths, significance_threshold=0.01))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "For `paths_2` where the observed paths are in line what we would expect based on the weighted edges in the first-order graph model, we correctly find that we do not need to consider a second-order model"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1\n"
     ]
    }
   ],
   "source": [
    "m2 = pp.MultiOrderModel.from_PathData(paths_2, max_order=2)\n",
    "print(m2.estimate_order(paths_2, significance_threshold=0.01))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Admittedly, the situation in the `paths` data is extreme insofar as two of the possible paths of lengths two have not been observed at all. We can also have more subtle deviations from the expectation based on the first-order graph model. Consider the following case, where two of the paths are observed more often than two other paths. Note that we have assigned the path frequencies such that again all edges are traversed with exactly the same frequencies, i.e. all edge weights are again equal in the first-order graph."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "tensor([8., 8., 8., 8.])\n",
      "1\n"
     ]
    }
   ],
   "source": [
    "g = pp.Graph.from_edge_list([('a', 'c'),\n",
    "                             ('b', 'c'),\n",
    "                             ('c', 'd'),\n",
    "                             ('c', 'e')])\n",
    "paths_3 = pp.PathData(g.mapping)\n",
    "\n",
    "paths_3.append_walk(('a', 'c', 'd'), weight=6.0)\n",
    "paths_3.append_walk(('a', 'c', 'e'), weight=2.0)\n",
    "\n",
    "paths_3.append_walk(('b', 'c', 'e'), weight=6.0)\n",
    "paths_3.append_walk(('b', 'c', 'd'), weight=2.0)\n",
    "\n",
    "m3 = pp.MultiOrderModel.from_PathData(paths_3, max_order=2)\n",
    "print(m3.layers[1].data.edge_weight)\n",
    "print(m3.estimate_order(paths_3, significance_threshold=0.01))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "In this example, due to the relatively small number of observations, the deviations from the expected baseline are still not strong enough to detect the optimal order of two (at least not for a significance threshold of $0.01$). We would need to raise the signififance threshold to $0.15$ to detect order two:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "2\n"
     ]
    }
   ],
   "source": [
    "print(m3.estimate_order(paths_3, significance_threshold=0.15))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Alternatively, we are able to detect a significant deviation from a first-order model for a significance threshold of $0.01$ if we make the deviations more extreme. If we observe two fo the paths seven times, while the other two are only observed once we find that order two is significant at a sigificance threshold of $0.01$."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "tensor([8., 8., 8., 8.])\n",
      "2\n"
     ]
    }
   ],
   "source": [
    "paths_3 = pp.PathData(g.mapping)\n",
    "\n",
    "paths_3.append_walk(('a', 'c', 'd'), weight=7.0)\n",
    "paths_3.append_walk(('a', 'c', 'e'), weight=1.0)\n",
    "\n",
    "paths_3.append_walk(('b', 'c', 'e'), weight=7.0)\n",
    "paths_3.append_walk(('b', 'c', 'd'), weight=1.0)\n",
    "\n",
    "m3 = pp.MultiOrderModel.from_PathData(paths_3, max_order=2)\n",
    "print(m3.layers[1].data.edge_weight)\n",
    "print(m3.estimate_order(paths_3, significance_threshold=0.01))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The ability to detect the optimal higher-order for a given data set in a statistically principled way is a powerful feature of the modelling framework of higher-order De Bruijn Graphs. Admittedly, the likelihood-based model selection approach has its limitations especially for small data sets or partially observed graphs, where it can both over- or underfit. To address this, we have developed an alternative Bayesian model selection technique that is explained and evaluated in the following paper: \n",
    "\n",
    "[L Petrovic, I Scholtes: Learning the Markov order of paths, In Proc. of the ACM Web Conference (WWW'22), April 2022](https://dl.acm.org/doi/10.1145/3485447.3512091)\n",
    "\n",
    "Unfortunately, this method has not yet been implemented in `pathpyG` but we are planning to add it soon."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Loading empirical path data from N-Gram Files\n",
    "\n",
    "For real data on walks in graphs it is not convenient to manually construct and add walks based on edge tensors. We can instead use the `from_ngram` function of class `PathData` to load such data from an n-gram file, i.e. a text file where each line corresponds to one observed walk consisting of comma-separated node IDs. If we set the argument `weight=True`, the last component of each line is considered to be the observation frequency of that particular walk.\n",
    "\n",
    "As an example, the file `data/tube_paths_train.ngram` contains observed passenger itineraries between nodes in a graph that representes the network of London Tube stations. Each of those itineraries is associated with an observation frequencies. The following is an excerpt from that file:\n",
    "\n",
    "```\n",
    "Southwark,Waterloo,212.0\n",
    "Liverpool Street,Bank / Monument,1271.0\n",
    "Barking,West Ham,283.0\n",
    "Tufnell Park,Kentish Town,103.0\n",
    "...\n",
    "```\n",
    "\n",
    "Note that this will automatically create an internal mapping of node IDs to indices."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 28,
   "metadata": {},
   "outputs": [],
   "source": [
    "paths_tube = pp.PathData.from_ngram('../data/tube_paths_train.ngram', sep=',', weight=True)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "To plot a (first-order) graph representation of the London Tube metro network, we can use the following code:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 29,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<style>\n",
       "svg circle.node {\n",
       "  fill: #3b5998;\n",
       "  stroke: #1b3978;\n",
       "  stroke-width: 2.5px;\n",
       "  r: 15px;\n",
       "  opacity: 1;\n",
       "}\n",
       "\n",
       ".link {\n",
       "  stroke: #969595;\n",
       "  stroke-opacity: .75;\n",
       "  stroke-width: 2.5px;\n",
       "}\n",
       "\n",
       ".arrow {\n",
       "  fill: #969595;\n",
       "}\n",
       "\n",
       "\n",
       ".label-text {\n",
       "    fill: #969595;\n",
       "    font-size: 16px;\n",
       "    font-family: sans-serif;\n",
       "}\n",
       "\n",
       "</style>\n",
       "\n",
       "<div id = \"xe05b2d8866834ff7bb82ba2e2fded367\"> </div>\n",
       "<script charset=\"utf-8\" src=\"https://d3js.org/d3.v5.min.js\"></script>\n",
       "<script charset=\"utf-8\">\n",
       "// Load via requireJS if available (jupyter notebook environment)\n",
       "try {\n",
       "    // Problem: require.config will raise an exception when called for the second time \n",
       "    require.config({\n",
       "        paths: {\n",
       "            d3: \"https://d3js.org/d3.v5.min.js\".replace(\".js\", \"\")\n",
       "        }\n",
       "    });\n",
       "    console.log(\"OKAY: requireJS was detected.\");\n",
       "}\n",
       "catch(err){\n",
       "    // a reference error indicates that requireJS does not exist. \n",
       "    // other errors may occur due to multiple calls to config\n",
       "    if (err instanceof ReferenceError){\n",
       "        console.log(\"WARNING: NO requireJS was detected!\");\n",
       "\n",
       "        // Helper function that waits for d3js to be loaded\n",
       "        require = function require(symbols, callback) {\n",
       "            var ms = 10;\n",
       "            window.setTimeout(function(t) {\n",
       "                if (window[symbols[0]])\n",
       "                    callback(window[symbols[0]]);\n",
       "                else \n",
       "                    window.setTimeout(arguments.callee, ms);\n",
       "            }, ms);\n",
       "        }\n",
       "    }\n",
       "};\n",
       "require(['d3'], function(d3){ //START\n",
       "const data = {\"edges\": [{\"uid\": \"Acton Town-Ealing Common\", \"source\": \"Acton Town\", \"target\": \"Ealing Common\", \"weight\": 1613.0}, {\"uid\": \"Acton Town-Chiswick Park\", \"source\": \"Acton Town\", \"target\": \"Chiswick Park\", \"weight\": 37872.0}, {\"uid\": \"Acton Town-Turnham Green\", \"source\": \"Acton Town\", \"target\": \"Turnham Green\", \"weight\": 19528.0}, {\"uid\": \"Acton Town-Hammersmith (Dis)\", \"source\": \"Acton Town\", \"target\": \"Hammersmith (Dis)\", \"weight\": 22841.0}, {\"uid\": \"Acton Town-South Ealing\", \"source\": \"Acton Town\", \"target\": \"South Ealing\", \"weight\": 9761.0}, {\"uid\": \"Aldgate-Tower Hill\", \"source\": \"Aldgate\", \"target\": \"Tower Hill\", \"weight\": 5382.0}, {\"uid\": \"Aldgate-Liverpool Street\", \"source\": \"Aldgate\", \"target\": \"Liverpool Street\", \"weight\": 897.0}, {\"uid\": \"Aldgate East-Liverpool Street\", \"source\": \"Aldgate East\", \"target\": \"Liverpool Street\", \"weight\": 42934.0}, {\"uid\": \"Aldgate East-Tower Hill\", \"source\": \"Aldgate East\", \"target\": \"Tower Hill\", \"weight\": 25591.0}, {\"uid\": \"Aldgate East-Whitechapel\", \"source\": \"Aldgate East\", \"target\": \"Whitechapel\", \"weight\": 58483.0}, {\"uid\": \"Alperton-Park Royal\", \"source\": \"Alperton\", \"target\": \"Park Royal\", \"weight\": 5274.0}, {\"uid\": \"Alperton-Sudbury Town\", \"source\": \"Alperton\", \"target\": \"Sudbury Town\", \"weight\": 3326.0}, {\"uid\": \"Amersham-Chalfont & Latimer\", \"source\": \"Amersham\", \"target\": \"Chalfont & Latimer\", \"weight\": 1393.0}, {\"uid\": \"Angel-Old Street\", \"source\": \"Angel\", \"target\": \"Old Street\", \"weight\": 46205.0}, {\"uid\": \"Angel-King's Cross St. Pancras\", \"source\": \"Angel\", \"target\": \"King's Cross St. Pancras\", \"weight\": 45354.0}, {\"uid\": \"Archway-Highgate\", \"source\": \"Archway\", \"target\": \"Highgate\", \"weight\": 20024.0}, {\"uid\": \"Archway-Tufnell Park\", \"source\": \"Archway\", \"target\": \"Tufnell Park\", \"weight\": 25658.0}, {\"uid\": \"Arnos Grove-Bounds Green\", \"source\": \"Arnos Grove\", \"target\": \"Bounds Green\", \"weight\": 10136.0}, {\"uid\": \"Arnos Grove-Southgate\", \"source\": \"Arnos Grove\", \"target\": \"Southgate\", \"weight\": 7101.0}, {\"uid\": \"Arsenal-Finsbury Park\", \"source\": \"Arsenal\", \"target\": \"Finsbury Park\", \"weight\": 3504.0}, {\"uid\": \"Arsenal-Holloway Road\", \"source\": \"Arsenal\", \"target\": \"Holloway Road\", \"weight\": 2926.0}, {\"uid\": \"Baker Street-Marylebone\", \"source\": \"Baker Street\", \"target\": \"Marylebone\", \"weight\": 121377.0}, {\"uid\": \"Baker Street-St. John's Wood\", \"source\": \"Baker Street\", \"target\": \"St. John's Wood\", \"weight\": 111237.0}, {\"uid\": \"Baker Street-Regent's Park\", \"source\": \"Baker Street\", \"target\": \"Regent's Park\", \"weight\": 48002.0}, {\"uid\": \"Baker Street-Great Portland Street\", \"source\": \"Baker Street\", \"target\": \"Great Portland Street\", \"weight\": 69211.0}, {\"uid\": \"Baker Street-Bond Street\", \"source\": \"Baker Street\", \"target\": \"Bond Street\", \"weight\": 22410.0}, {\"uid\": \"Baker Street-Edgware Road (Cir)\", \"source\": \"Baker Street\", \"target\": \"Edgware Road (Cir)\", \"weight\": 11005.0}, {\"uid\": \"Baker Street-Finchley Road\", \"source\": \"Baker Street\", \"target\": \"Finchley Road\", \"weight\": 7455.0}, {\"uid\": \"Balham-Clapham South\", \"source\": \"Balham\", \"target\": \"Clapham South\", \"weight\": 32924.0}, {\"uid\": \"Balham-Tooting Bec\", \"source\": \"Balham\", \"target\": \"Tooting Bec\", \"weight\": 25372.0}, {\"uid\": \"Bank / Monument-Cannon Street\", \"source\": \"Bank / Monument\", \"target\": \"Cannon Street\", \"weight\": 8067.0}, {\"uid\": \"Bank / Monument-Liverpool Street\", \"source\": \"Bank / Monument\", \"target\": \"Liverpool Street\", \"weight\": 100483.0}, {\"uid\": \"Bank / Monument-London Bridge\", \"source\": \"Bank / Monument\", \"target\": \"London Bridge\", \"weight\": 124555.0}, {\"uid\": \"Bank / Monument-Moorgate\", \"source\": \"Bank / Monument\", \"target\": \"Moorgate\", \"weight\": 41402.0}, {\"uid\": \"Bank / Monument-St. Paul's\", \"source\": \"Bank / Monument\", \"target\": \"St. Paul's\", \"weight\": 83688.0}, {\"uid\": \"Bank / Monument-Tower Hill\", \"source\": \"Bank / Monument\", \"target\": \"Tower Hill\", \"weight\": 37843.0}, {\"uid\": \"Barbican-Farringdon\", \"source\": \"Barbican\", \"target\": \"Farringdon\", \"weight\": 42448.0}, {\"uid\": \"Barbican-Moorgate\", \"source\": \"Barbican\", \"target\": \"Moorgate\", \"weight\": 40684.0}, {\"uid\": \"Barking-East Ham\", \"source\": \"Barking\", \"target\": \"East Ham\", \"weight\": 11155.0}, {\"uid\": \"Barking-Upminster\", \"source\": \"Barking\", \"target\": \"Upminster\", \"weight\": 6371.0}, {\"uid\": \"Barking-Upney\", \"source\": \"Barking\", \"target\": \"Upney\", \"weight\": 8835.0}, {\"uid\": \"Barking-West Ham\", \"source\": \"Barking\", \"target\": \"West Ham\", \"weight\": 28203.0}, {\"uid\": \"Barkingside-Fairlop\", \"source\": \"Barkingside\", \"target\": \"Fairlop\", \"weight\": 1924.0}, {\"uid\": \"Barkingside-Newbury Park\", \"source\": \"Barkingside\", \"target\": \"Newbury Park\", \"weight\": 2964.0}, {\"uid\": \"Barons Court-West Kensington\", \"source\": \"Barons Court\", \"target\": \"West Kensington\", \"weight\": 31752.0}, {\"uid\": \"Barons Court-Earl's Court\", \"source\": \"Barons Court\", \"target\": \"Earl's Court\", \"weight\": 26632.0}, {\"uid\": \"Barons Court-Hammersmith (Dis)\", \"source\": \"Barons Court\", \"target\": \"Hammersmith (Dis)\", \"weight\": 691.0}, {\"uid\": \"Bayswater-Notting Hill Gate\", \"source\": \"Bayswater\", \"target\": \"Notting Hill Gate\", \"weight\": 17299.0}, {\"uid\": \"Bayswater-Paddington\", \"source\": \"Bayswater\", \"target\": \"Paddington\", \"weight\": 17564.0}, {\"uid\": \"Becontree-Dagenham Heathway\", \"source\": \"Becontree\", \"target\": \"Dagenham Heathway\", \"weight\": 5609.0}, {\"uid\": \"Becontree-Upney\", \"source\": \"Becontree\", \"target\": \"Upney\", \"weight\": 7868.0}, {\"uid\": \"Belsize Park-Chalk Farm\", \"source\": \"Belsize Park\", \"target\": \"Chalk Farm\", \"weight\": 25837.0}, {\"uid\": \"Belsize Park-Hampstead\", \"source\": \"Belsize Park\", \"target\": \"Hampstead\", \"weight\": 21525.0}, {\"uid\": \"Bermondsey-London Bridge\", \"source\": \"Bermondsey\", \"target\": \"London Bridge\", \"weight\": 67140.0}, {\"uid\": \"Bermondsey-Canada Water\", \"source\": \"Bermondsey\", \"target\": \"Canada Water\", \"weight\": 67178.0}, {\"uid\": \"Bethnal Green-Liverpool Street\", \"source\": \"Bethnal Green\", \"target\": \"Liverpool Street\", \"weight\": 52550.0}, {\"uid\": \"Bethnal Green-Mile End\", \"source\": \"Bethnal Green\", \"target\": \"Mile End\", \"weight\": 44671.0}, {\"uid\": \"Blackfriars-Mansion House\", \"source\": \"Blackfriars\", \"target\": \"Mansion House\", \"weight\": 9259.0}, {\"uid\": \"Blackfriars-Temple\", \"source\": \"Blackfriars\", \"target\": \"Temple\", \"weight\": 12516.0}, {\"uid\": \"Blackhorse Road-Tottenham Hale\", \"source\": \"Blackhorse Road\", \"target\": \"Tottenham Hale\", \"weight\": 17566.0}, {\"uid\": \"Blackhorse Road-Walthamstow Central\", \"source\": \"Blackhorse Road\", \"target\": \"Walthamstow Central\", \"weight\": 11894.0}, {\"uid\": \"Bond Street-Tottenham Court Road\", \"source\": \"Bond Street\", \"target\": \"Tottenham Court Road\", \"weight\": 115691.0}, {\"uid\": \"Bond Street-Oxford Circus\", \"source\": \"Bond Street\", \"target\": \"Oxford Circus\", \"weight\": 91629.0}, {\"uid\": \"Bond Street-Green Park\", \"source\": \"Bond Street\", \"target\": \"Green Park\", \"weight\": 39394.0}, {\"uid\": \"Bond Street-Baker Street\", \"source\": \"Bond Street\", \"target\": \"Baker Street\", \"weight\": 18130.0}, {\"uid\": \"Bond Street-Marble Arch\", \"source\": \"Bond Street\", \"target\": \"Marble Arch\", \"weight\": 76473.0}, {\"uid\": \"Borough-Elephant & Castle\", \"source\": \"Borough\", \"target\": \"Elephant & Castle\", \"weight\": 19372.0}, {\"uid\": \"Borough-London Bridge\", \"source\": \"Borough\", \"target\": \"London Bridge\", \"weight\": 21644.0}, {\"uid\": \"Boston Manor-Northfields\", \"source\": \"Boston Manor\", \"target\": \"Northfields\", \"weight\": 20140.0}, {\"uid\": \"Boston Manor-Osterley\", \"source\": \"Boston Manor\", \"target\": \"Osterley\", \"weight\": 17401.0}, {\"uid\": \"Bounds Green-Arnos Grove\", \"source\": \"Bounds Green\", \"target\": \"Arnos Grove\", \"weight\": 10046.0}, {\"uid\": \"Bounds Green-Wood Green\", \"source\": \"Bounds Green\", \"target\": \"Wood Green\", \"weight\": 14694.0}, {\"uid\": \"Bow Road-Mile End\", \"source\": \"Bow Road\", \"target\": \"Mile End\", \"weight\": 1893.0}, {\"uid\": \"Bow Road-BromleyByBow\", \"source\": \"Bow Road\", \"target\": \"BromleyByBow\", \"weight\": 4518.0}, {\"uid\": \"Brent Cross-Golders Green\", \"source\": \"Brent Cross\", \"target\": \"Golders Green\", \"weight\": 15098.0}, {\"uid\": \"Brent Cross-Hendon Central\", \"source\": \"Brent Cross\", \"target\": \"Hendon Central\", \"weight\": 12883.0}, {\"uid\": \"Brixton-Stockwell\", \"source\": \"Brixton\", \"target\": \"Stockwell\", \"weight\": 19619.0}, {\"uid\": \"BromleyByBow-Bow Road\", \"source\": \"BromleyByBow\", \"target\": \"Bow Road\", \"weight\": 1886.0}, {\"uid\": \"BromleyByBow-West Ham\", \"source\": \"BromleyByBow\", \"target\": \"West Ham\", \"weight\": 1666.0}, {\"uid\": \"Buckhurst Hill-Loughton\", \"source\": \"Buckhurst Hill\", \"target\": \"Loughton\", \"weight\": 6767.0}, {\"uid\": \"Buckhurst Hill-Woodford\", \"source\": \"Buckhurst Hill\", \"target\": \"Woodford\", \"weight\": 8695.0}, {\"uid\": \"Burnt Oak-Colindale\", \"source\": \"Burnt Oak\", \"target\": \"Colindale\", \"weight\": 6152.0}, {\"uid\": \"Burnt Oak-Edgware\", \"source\": \"Burnt Oak\", \"target\": \"Edgware\", \"weight\": 3342.0}, {\"uid\": \"Caledonian Road-Holloway Road\", \"source\": \"Caledonian Road\", \"target\": \"Holloway Road\", \"weight\": 5211.0}, {\"uid\": \"Caledonian Road-King's Cross St. Pancras\", \"source\": \"Caledonian Road\", \"target\": \"King's Cross St. Pancras\", \"weight\": 8774.0}, {\"uid\": \"Camden Town-Chalk Farm\", \"source\": \"Camden Town\", \"target\": \"Chalk Farm\", \"weight\": 27494.0}, {\"uid\": \"Camden Town-Mornington Crescent\", \"source\": \"Camden Town\", \"target\": \"Mornington Crescent\", \"weight\": 65938.0}, {\"uid\": \"Camden Town-Euston\", \"source\": \"Camden Town\", \"target\": \"Euston\", \"weight\": 31782.0}, {\"uid\": \"Camden Town-Kentish Town\", \"source\": \"Camden Town\", \"target\": \"Kentish Town\", \"weight\": 995.0}, {\"uid\": \"Canada Water-Bermondsey\", \"source\": \"Canada Water\", \"target\": \"Bermondsey\", \"weight\": 64834.0}, {\"uid\": \"Canada Water-Canary Wharf\", \"source\": \"Canada Water\", \"target\": \"Canary Wharf\", \"weight\": 59940.0}, {\"uid\": \"Canary Wharf-Canada Water\", \"source\": \"Canary Wharf\", \"target\": \"Canada Water\", \"weight\": 57618.0}, {\"uid\": \"Canary Wharf-North Greenwich\", \"source\": \"Canary Wharf\", \"target\": \"North Greenwich\", \"weight\": 38199.0}, {\"uid\": \"Canning Town-North Greenwich\", \"source\": \"Canning Town\", \"target\": \"North Greenwich\", \"weight\": 29811.0}, {\"uid\": \"Canning Town-West Ham\", \"source\": \"Canning Town\", \"target\": \"West Ham\", \"weight\": 21900.0}, {\"uid\": \"Cannon Street-Bank / Monument\", \"source\": \"Cannon Street\", \"target\": \"Bank / Monument\", \"weight\": 7430.0}, {\"uid\": \"Cannon Street-Mansion House\", \"source\": \"Cannon Street\", \"target\": \"Mansion House\", \"weight\": 7122.0}, {\"uid\": \"Canons Park-Queensbury\", \"source\": \"Canons Park\", \"target\": \"Queensbury\", \"weight\": 4250.0}, {\"uid\": \"Canons Park-Stanmore\", \"source\": \"Canons Park\", \"target\": \"Stanmore\", \"weight\": 2667.0}, {\"uid\": \"Chalfont & Latimer-Amersham\", \"source\": \"Chalfont & Latimer\", \"target\": \"Amersham\", \"weight\": 1386.0}, {\"uid\": \"Chalfont & Latimer-Chesham\", \"source\": \"Chalfont & Latimer\", \"target\": \"Chesham\", \"weight\": 536.0}, {\"uid\": \"Chalfont & Latimer-Chorleywood\", \"source\": \"Chalfont & Latimer\", \"target\": \"Chorleywood\", \"weight\": 2658.0}, {\"uid\": \"Chalk Farm-Belsize Park\", \"source\": \"Chalk Farm\", \"target\": \"Belsize Park\", \"weight\": 24653.0}, {\"uid\": \"Chalk Farm-Camden Town\", \"source\": \"Chalk Farm\", \"target\": \"Camden Town\", \"weight\": 29255.0}, {\"uid\": \"Chancery Lane-St. Paul's\", \"source\": \"Chancery Lane\", \"target\": \"St. Paul's\", \"weight\": 87154.0}, {\"uid\": \"Chancery Lane-Holborn\", \"source\": \"Chancery Lane\", \"target\": \"Holborn\", \"weight\": 85253.0}, {\"uid\": \"Charing Cross-Embankment\", \"source\": \"Charing Cross\", \"target\": \"Embankment\", \"weight\": 14187.0}, {\"uid\": \"Charing Cross-Leicester Square\", \"source\": \"Charing Cross\", \"target\": \"Leicester Square\", \"weight\": 10888.0}, {\"uid\": \"Charing Cross-Piccadilly Circus\", \"source\": \"Charing Cross\", \"target\": \"Piccadilly Circus\", \"weight\": 12796.0}, {\"uid\": \"Chesham-Chalfont & Latimer\", \"source\": \"Chesham\", \"target\": \"Chalfont & Latimer\", \"weight\": 605.0}, {\"uid\": \"Chigwell-Grange Hill\", \"source\": \"Chigwell\", \"target\": \"Grange Hill\", \"weight\": 1751.0}, {\"uid\": \"Chigwell-Roding Valley\", \"source\": \"Chigwell\", \"target\": \"Roding Valley\", \"weight\": 2096.0}, {\"uid\": \"Chiswick Park-Acton Town\", \"source\": \"Chiswick Park\", \"target\": \"Acton Town\", \"weight\": 1388.0}, {\"uid\": \"Chiswick Park-Turnham Green\", \"source\": \"Chiswick Park\", \"target\": \"Turnham Green\", \"weight\": 281.0}, {\"uid\": \"Chorleywood-Chalfont & Latimer\", \"source\": \"Chorleywood\", \"target\": \"Chalfont & Latimer\", \"weight\": 2510.0}, {\"uid\": \"Chorleywood-Rickmansworth\", \"source\": \"Chorleywood\", \"target\": \"Rickmansworth\", \"weight\": 3406.0}, {\"uid\": \"Clapham Common-Clapham North\", \"source\": \"Clapham Common\", \"target\": \"Clapham North\", \"weight\": 42635.0}, {\"uid\": \"Clapham Common-Clapham South\", \"source\": \"Clapham Common\", \"target\": \"Clapham South\", \"weight\": 37086.0}, {\"uid\": \"Clapham North-Clapham Common\", \"source\": \"Clapham North\", \"target\": \"Clapham Common\", \"weight\": 42742.0}, {\"uid\": \"Clapham North-Stockwell\", \"source\": \"Clapham North\", \"target\": \"Stockwell\", \"weight\": 45366.0}, {\"uid\": \"Clapham South-Balham\", \"source\": \"Clapham South\", \"target\": \"Balham\", \"weight\": 31848.0}, {\"uid\": \"Clapham South-Clapham Common\", \"source\": \"Clapham South\", \"target\": \"Clapham Common\", \"weight\": 37995.0}, {\"uid\": \"Cockfosters-Oakwood\", \"source\": \"Cockfosters\", \"target\": \"Oakwood\", \"weight\": 1377.0}, {\"uid\": \"Colindale-Burnt Oak\", \"source\": \"Colindale\", \"target\": \"Burnt Oak\", \"weight\": 5934.0}, {\"uid\": \"Colindale-Hendon Central\", \"source\": \"Colindale\", \"target\": \"Hendon Central\", \"weight\": 9332.0}, {\"uid\": \"Colliers Wood-South Wimbledon\", \"source\": \"Colliers Wood\", \"target\": \"South Wimbledon\", \"weight\": 8489.0}, {\"uid\": \"Colliers Wood-Tooting Broadway\", \"source\": \"Colliers Wood\", \"target\": \"Tooting Broadway\", \"weight\": 12964.0}, {\"uid\": \"Covent Garden-Holborn\", \"source\": \"Covent Garden\", \"target\": \"Holborn\", \"weight\": 13044.0}, {\"uid\": \"Covent Garden-Leicester Square\", \"source\": \"Covent Garden\", \"target\": \"Leicester Square\", \"weight\": 10857.0}, {\"uid\": \"Croxley-Moor Park\", \"source\": \"Croxley\", \"target\": \"Moor Park\", \"weight\": 2080.0}, {\"uid\": \"Croxley-Watford\", \"source\": \"Croxley\", \"target\": \"Watford\", \"weight\": 1276.0}, {\"uid\": \"Dagenham East-Dagenham Heathway\", \"source\": \"Dagenham East\", \"target\": \"Dagenham Heathway\", \"weight\": 2087.0}, {\"uid\": \"Dagenham East-Elm Park\", \"source\": \"Dagenham East\", \"target\": \"Elm Park\", \"weight\": 333.0}, {\"uid\": \"Dagenham Heathway-Becontree\", \"source\": \"Dagenham Heathway\", \"target\": \"Becontree\", \"weight\": 5872.0}, {\"uid\": \"Dagenham Heathway-Dagenham East\", \"source\": \"Dagenham Heathway\", \"target\": \"Dagenham East\", \"weight\": 1968.0}, {\"uid\": \"Debden-Loughton\", \"source\": \"Debden\", \"target\": \"Loughton\", \"weight\": 5117.0}, {\"uid\": \"Debden-Theydon Bois\", \"source\": \"Debden\", \"target\": \"Theydon Bois\", \"weight\": 2700.0}, {\"uid\": \"Dollis Hill-Neasden\", \"source\": \"Dollis Hill\", \"target\": \"Neasden\", \"weight\": 565.0}, {\"uid\": \"Dollis Hill-Willesden Green\", \"source\": \"Dollis Hill\", \"target\": \"Willesden Green\", \"weight\": 2411.0}, {\"uid\": \"Ealing Broadway-Ealing Common\", \"source\": \"Ealing Broadway\", \"target\": \"Ealing Common\", \"weight\": 33654.0}, {\"uid\": \"Ealing Broadway-Paddington\", \"source\": \"Ealing Broadway\", \"target\": \"Paddington\", \"weight\": 55622.0}, {\"uid\": \"Ealing Broadway-West Acton\", \"source\": \"Ealing Broadway\", \"target\": \"West Acton\", \"weight\": 16036.0}, {\"uid\": \"Ealing Common-North Ealing\", \"source\": \"Ealing Common\", \"target\": \"North Ealing\", \"weight\": 36963.0}, {\"uid\": \"Ealing Common-Ealing Broadway\", \"source\": \"Ealing Common\", \"target\": \"Ealing Broadway\", \"weight\": 35018.0}, {\"uid\": \"Ealing Common-Acton Town\", \"source\": \"Ealing Common\", \"target\": \"Acton Town\", \"weight\": 5603.0}, {\"uid\": \"Earl's Court-Barons Court\", \"source\": \"Earl's Court\", \"target\": \"Barons Court\", \"weight\": 27746.0}, {\"uid\": \"Earl's Court-Gloucester Road\", \"source\": \"Earl's Court\", \"target\": \"Gloucester Road\", \"weight\": 56599.0}, {\"uid\": \"Earl's Court-High Street Kensington\", \"source\": \"Earl's Court\", \"target\": \"High Street Kensington\", \"weight\": 10103.0}, {\"uid\": \"Earl's Court-Kensington (Olympia)\", \"source\": \"Earl's Court\", \"target\": \"Kensington (Olympia)\", \"weight\": 664.0}, {\"uid\": \"Earl's Court-West Brompton\", \"source\": \"Earl's Court\", \"target\": \"West Brompton\", \"weight\": 27929.0}, {\"uid\": \"Earl's Court-West Kensington\", \"source\": \"Earl's Court\", \"target\": \"West Kensington\", \"weight\": 2748.0}, {\"uid\": \"East Acton-North Acton\", \"source\": \"East Acton\", \"target\": \"North Acton\", \"weight\": 6704.0}, {\"uid\": \"East Acton-White City\", \"source\": \"East Acton\", \"target\": \"White City\", \"weight\": 6069.0}, {\"uid\": \"East Finchley-Finchley Central\", \"source\": \"East Finchley\", \"target\": \"Finchley Central\", \"weight\": 11811.0}, {\"uid\": \"East Finchley-Highgate\", \"source\": \"East Finchley\", \"target\": \"Highgate\", \"weight\": 16762.0}, {\"uid\": \"East Ham-Barking\", \"source\": \"East Ham\", \"target\": \"Barking\", \"weight\": 10366.0}, {\"uid\": \"East Ham-Upton Park\", \"source\": \"East Ham\", \"target\": \"Upton Park\", \"weight\": 2150.0}, {\"uid\": \"East Putney-Putney Bridge\", \"source\": \"East Putney\", \"target\": \"Putney Bridge\", \"weight\": 16013.0}, {\"uid\": \"East Putney-Southfields\", \"source\": \"East Putney\", \"target\": \"Southfields\", \"weight\": 11738.0}, {\"uid\": \"Eastcote-Rayners Lane\", \"source\": \"Eastcote\", \"target\": \"Rayners Lane\", \"weight\": 10097.0}, {\"uid\": \"Eastcote-Ruislip Manor\", \"source\": \"Eastcote\", \"target\": \"Ruislip Manor\", \"weight\": 8706.0}, {\"uid\": \"Edgware-Burnt Oak\", \"source\": \"Edgware\", \"target\": \"Burnt Oak\", \"weight\": 3401.0}, {\"uid\": \"Edgware Road (Bak)-Marylebone\", \"source\": \"Edgware Road (Bak)\", \"target\": \"Marylebone\", \"weight\": 7235.0}, {\"uid\": \"Edgware Road (Bak)-Paddington\", \"source\": \"Edgware Road (Bak)\", \"target\": \"Paddington\", \"weight\": 3883.0}, {\"uid\": \"Edgware Road (Cir)-Baker Street\", \"source\": \"Edgware Road (Cir)\", \"target\": \"Baker Street\", \"weight\": 115986.0}, {\"uid\": \"Edgware Road (Cir)-Paddington\", \"source\": \"Edgware Road (Cir)\", \"target\": \"Paddington\", \"weight\": 110053.0}, {\"uid\": \"Elephant & Castle-Borough\", \"source\": \"Elephant & Castle\", \"target\": \"Borough\", \"weight\": 20245.0}, {\"uid\": \"Elephant & Castle-Lambeth North\", \"source\": \"Elephant & Castle\", \"target\": \"Lambeth North\", \"weight\": 19756.0}, {\"uid\": \"Elephant & Castle-Kennington\", \"source\": \"Elephant & Castle\", \"target\": \"Kennington\", \"weight\": 4217.0}, {\"uid\": \"Elm Park-Dagenham East\", \"source\": \"Elm Park\", \"target\": \"Dagenham East\", \"weight\": 323.0}, {\"uid\": \"Elm Park-Hornchurch\", \"source\": \"Elm Park\", \"target\": \"Hornchurch\", \"weight\": 2225.0}, {\"uid\": \"Embankment-Charing Cross\", \"source\": \"Embankment\", \"target\": \"Charing Cross\", \"weight\": 14076.0}, {\"uid\": \"Embankment-Temple\", \"source\": \"Embankment\", \"target\": \"Temple\", \"weight\": 15080.0}, {\"uid\": \"Embankment-Waterloo\", \"source\": \"Embankment\", \"target\": \"Waterloo\", \"weight\": 18980.0}, {\"uid\": \"Embankment-Westminster\", \"source\": \"Embankment\", \"target\": \"Westminster\", \"weight\": 19770.0}, {\"uid\": \"Epping-Theydon Bois\", \"source\": \"Epping\", \"target\": \"Theydon Bois\", \"weight\": 2712.0}, {\"uid\": \"Euston-Mornington Crescent\", \"source\": \"Euston\", \"target\": \"Mornington Crescent\", \"weight\": 63981.0}, {\"uid\": \"Euston-Warren Street\", \"source\": \"Euston\", \"target\": \"Warren Street\", \"weight\": 92960.0}, {\"uid\": \"Euston-King's Cross St. Pancras\", \"source\": \"Euston\", \"target\": \"King's Cross St. Pancras\", \"weight\": 2106.0}, {\"uid\": \"Euston-Camden Town\", \"source\": \"Euston\", \"target\": \"Camden Town\", \"weight\": 108657.0}, {\"uid\": \"Euston Square-Great Portland Street\", \"source\": \"Euston Square\", \"target\": \"Great Portland Street\", \"weight\": 68736.0}, {\"uid\": \"Euston Square-King's Cross St. Pancras\", \"source\": \"Euston Square\", \"target\": \"King's Cross St. Pancras\", \"weight\": 69767.0}, {\"uid\": \"Fairlop-Barkingside\", \"source\": \"Fairlop\", \"target\": \"Barkingside\", \"weight\": 2065.0}, {\"uid\": \"Fairlop-Hainault\", \"source\": \"Fairlop\", \"target\": \"Hainault\", \"weight\": 1205.0}, {\"uid\": \"Farringdon-Barbican\", \"source\": \"Farringdon\", \"target\": \"Barbican\", \"weight\": 42745.0}, {\"uid\": \"Farringdon-King's Cross St. Pancras\", \"source\": \"Farringdon\", \"target\": \"King's Cross St. Pancras\", \"weight\": 46904.0}, {\"uid\": \"Finchley Central-East Finchley\", \"source\": \"Finchley Central\", \"target\": \"East Finchley\", \"weight\": 11828.0}, {\"uid\": \"Finchley Central-West Finchley\", \"source\": \"Finchley Central\", \"target\": \"West Finchley\", \"weight\": 1016.0}, {\"uid\": \"Finchley Central-Mill Hill East\", \"source\": \"Finchley Central\", \"target\": \"Mill Hill East\", \"weight\": 7113.0}, {\"uid\": \"Finchley Road-Baker Street\", \"source\": \"Finchley Road\", \"target\": \"Baker Street\", \"weight\": 49615.0}, {\"uid\": \"Finchley Road-HarrowOnTheHill\", \"source\": \"Finchley Road\", \"target\": \"HarrowOnTheHill\", \"weight\": 13302.0}, {\"uid\": \"Finchley Road-Swiss Cottage\", \"source\": \"Finchley Road\", \"target\": \"Swiss Cottage\", \"weight\": 4063.0}, {\"uid\": \"Finchley Road-Wembley Park\", \"source\": \"Finchley Road\", \"target\": \"Wembley Park\", \"weight\": 15349.0}, {\"uid\": \"Finchley Road-West Hampstead\", \"source\": \"Finchley Road\", \"target\": \"West Hampstead\", \"weight\": 9680.0}, {\"uid\": \"Finchley Road-Willesden Green\", \"source\": \"Finchley Road\", \"target\": \"Willesden Green\", \"weight\": 10949.0}, {\"uid\": \"Finsbury Park-Highbury & Islington\", \"source\": \"Finsbury Park\", \"target\": \"Highbury & Islington\", \"weight\": 3193.0}, {\"uid\": \"Finsbury Park-Seven Sisters\", \"source\": \"Finsbury Park\", \"target\": \"Seven Sisters\", \"weight\": 64256.0}, {\"uid\": \"Finsbury Park-Manor House\", \"source\": \"Finsbury Park\", \"target\": \"Manor House\", \"weight\": 33089.0}, {\"uid\": \"Finsbury Park-Arsenal\", \"source\": \"Finsbury Park\", \"target\": \"Arsenal\", \"weight\": 31151.0}, {\"uid\": \"Fulham Broadway-Parsons Green\", \"source\": \"Fulham Broadway\", \"target\": \"Parsons Green\", \"weight\": 21159.0}, {\"uid\": \"Fulham Broadway-West Brompton\", \"source\": \"Fulham Broadway\", \"target\": \"West Brompton\", \"weight\": 27508.0}, {\"uid\": \"Gants Hill-Newbury Park\", \"source\": \"Gants Hill\", \"target\": \"Newbury Park\", \"weight\": 5365.0}, {\"uid\": \"Gants Hill-Redbridge\", \"source\": \"Gants Hill\", \"target\": \"Redbridge\", \"weight\": 11009.0}, {\"uid\": \"Gloucester Road-South Kensington\", \"source\": \"Gloucester Road\", \"target\": \"South Kensington\", \"weight\": 50658.0}, {\"uid\": \"Gloucester Road-High Street Kensington\", \"source\": \"Gloucester Road\", \"target\": \"High Street Kensington\", \"weight\": 10845.0}, {\"uid\": \"Gloucester Road-Earl's Court\", \"source\": \"Gloucester Road\", \"target\": \"Earl's Court\", \"weight\": 72426.0}, {\"uid\": \"Golders Green-Brent Cross\", \"source\": \"Golders Green\", \"target\": \"Brent Cross\", \"weight\": 14187.0}, {\"uid\": \"Golders Green-Hampstead\", \"source\": \"Golders Green\", \"target\": \"Hampstead\", \"weight\": 20044.0}, {\"uid\": \"Goldhawk Road-Hammersmith (H&C)\", \"source\": \"Goldhawk Road\", \"target\": \"Hammersmith (H&C)\", \"weight\": 4094.0}, {\"uid\": \"Goldhawk Road-Shepherd's Bush Market\", \"source\": \"Goldhawk Road\", \"target\": \"Shepherd's Bush Market\", \"weight\": 6467.0}, {\"uid\": \"Goodge Street-Tottenham Court Road\", \"source\": \"Goodge Street\", \"target\": \"Tottenham Court Road\", \"weight\": 7055.0}, {\"uid\": \"Goodge Street-Warren Street\", \"source\": \"Goodge Street\", \"target\": \"Warren Street\", \"weight\": 6312.0}, {\"uid\": \"Grange Hill-Chigwell\", \"source\": \"Grange Hill\", \"target\": \"Chigwell\", \"weight\": 1661.0}, {\"uid\": \"Grange Hill-Hainault\", \"source\": \"Grange Hill\", \"target\": \"Hainault\", \"weight\": 1312.0}, {\"uid\": \"Great Portland Street-Baker Street\", \"source\": \"Great Portland Street\", \"target\": \"Baker Street\", \"weight\": 68363.0}, {\"uid\": \"Great Portland Street-Euston Square\", \"source\": \"Great Portland Street\", \"target\": \"Euston Square\", \"weight\": 69646.0}, {\"uid\": \"Green Park-Bond Street\", \"source\": \"Green Park\", \"target\": \"Bond Street\", \"weight\": 89135.0}, {\"uid\": \"Green Park-Westminster\", \"source\": \"Green Park\", \"target\": \"Westminster\", \"weight\": 36737.0}, {\"uid\": \"Green Park-Victoria\", \"source\": \"Green Park\", \"target\": \"Victoria\", \"weight\": 95991.0}, {\"uid\": \"Green Park-Piccadilly Circus\", \"source\": \"Green Park\", \"target\": \"Piccadilly Circus\", \"weight\": 22336.0}, {\"uid\": \"Green Park-Oxford Circus\", \"source\": \"Green Park\", \"target\": \"Oxford Circus\", \"weight\": 118147.0}, {\"uid\": \"Green Park-Hyde Park Corner\", \"source\": \"Green Park\", \"target\": \"Hyde Park Corner\", \"weight\": 126188.0}, {\"uid\": \"Greenford-Northolt\", \"source\": \"Greenford\", \"target\": \"Northolt\", \"weight\": 6229.0}, {\"uid\": \"Greenford-Perivale\", \"source\": \"Greenford\", \"target\": \"Perivale\", \"weight\": 9691.0}, {\"uid\": \"Gunnersbury-Kew Gardens\", \"source\": \"Gunnersbury\", \"target\": \"Kew Gardens\", \"weight\": 9839.0}, {\"uid\": \"Gunnersbury-Turnham Green\", \"source\": \"Gunnersbury\", \"target\": \"Turnham Green\", \"weight\": 13314.0}, {\"uid\": \"Hainault-Fairlop\", \"source\": \"Hainault\", \"target\": \"Fairlop\", \"weight\": 1285.0}, {\"uid\": \"Hainault-Grange Hill\", \"source\": \"Hainault\", \"target\": \"Grange Hill\", \"weight\": 1235.0}, {\"uid\": \"Hammersmith (Dis)-Acton Town\", \"source\": \"Hammersmith (Dis)\", \"target\": \"Acton Town\", \"weight\": 18435.0}, {\"uid\": \"Hammersmith (Dis)-Barons Court\", \"source\": \"Hammersmith (Dis)\", \"target\": \"Barons Court\", \"weight\": 29740.0}, {\"uid\": \"Hammersmith (Dis)-Ravenscourt Park\", \"source\": \"Hammersmith (Dis)\", \"target\": \"Ravenscourt Park\", \"weight\": 2702.0}, {\"uid\": \"Hammersmith (Dis)-Turnham Green\", \"source\": \"Hammersmith (Dis)\", \"target\": \"Turnham Green\", \"weight\": 7951.0}, {\"uid\": \"Hammersmith (H&C)-Goldhawk Road\", \"source\": \"Hammersmith (H&C)\", \"target\": \"Goldhawk Road\", \"weight\": 5282.0}, {\"uid\": \"Hampstead-Golders Green\", \"source\": \"Hampstead\", \"target\": \"Golders Green\", \"weight\": 22409.0}, {\"uid\": \"Hampstead-Belsize Park\", \"source\": \"Hampstead\", \"target\": \"Belsize Park\", \"weight\": 19030.0}, {\"uid\": \"Hanger Lane-North Acton\", \"source\": \"Hanger Lane\", \"target\": \"North Acton\", \"weight\": 13427.0}, {\"uid\": \"Hanger Lane-Perivale\", \"source\": \"Hanger Lane\", \"target\": \"Perivale\", \"weight\": 10452.0}, {\"uid\": \"Harlesden-Stonebridge Park\", \"source\": \"Harlesden\", \"target\": \"Stonebridge Park\", \"weight\": 7806.0}, {\"uid\": \"Harlesden-Willesden Junction\", \"source\": \"Harlesden\", \"target\": \"Willesden Junction\", \"weight\": 10742.0}, {\"uid\": \"Harrow & Wealdstone-Kenton\", \"source\": \"Harrow & Wealdstone\", \"target\": \"Kenton\", \"weight\": 2892.0}, {\"uid\": \"HarrowOnTheHill-West Harrow\", \"source\": \"HarrowOnTheHill\", \"target\": \"West Harrow\", \"weight\": 13513.0}, {\"uid\": \"HarrowOnTheHill-Wembley Park\", \"source\": \"HarrowOnTheHill\", \"target\": \"Wembley Park\", \"weight\": 14063.0}, {\"uid\": \"HarrowOnTheHill-Rickmansworth\", \"source\": \"HarrowOnTheHill\", \"target\": \"Rickmansworth\", \"weight\": 4218.0}, {\"uid\": \"HarrowOnTheHill-Northwick Park\", \"source\": \"HarrowOnTheHill\", \"target\": \"Northwick Park\", \"weight\": 3618.0}, {\"uid\": \"HarrowOnTheHill-Moor Park\", \"source\": \"HarrowOnTheHill\", \"target\": \"Moor Park\", \"weight\": 3196.0}, {\"uid\": \"HarrowOnTheHill-North Harrow\", \"source\": \"HarrowOnTheHill\", \"target\": \"North Harrow\", \"weight\": 4454.0}, {\"uid\": \"HarrowOnTheHill-Finchley Road\", \"source\": \"HarrowOnTheHill\", \"target\": \"Finchley Road\", \"weight\": 2634.0}, {\"uid\": \"HarrowOnTheHill-Marylebone\", \"source\": \"HarrowOnTheHill\", \"target\": \"Marylebone\", \"weight\": 13209.0}, {\"uid\": \"Hatton Cross-Heathrow Terminal 4\", \"source\": \"Hatton Cross\", \"target\": \"Heathrow Terminal 4\", \"weight\": 1423.0}, {\"uid\": \"Hatton Cross-Hounslow West\", \"source\": \"Hatton Cross\", \"target\": \"Hounslow West\", \"weight\": 7277.0}, {\"uid\": \"Hatton Cross-Heathrow Terminals 123\", \"source\": \"Hatton Cross\", \"target\": \"Heathrow Terminals 123\", \"weight\": 11722.0}, {\"uid\": \"Heathrow Terminal 4-Hatton Cross\", \"source\": \"Heathrow Terminal 4\", \"target\": \"Hatton Cross\", \"weight\": 1652.0}, {\"uid\": \"Heathrow Terminal 4-Heathrow Terminals 123\", \"source\": \"Heathrow Terminal 4\", \"target\": \"Heathrow Terminals 123\", \"weight\": 52.0}, {\"uid\": \"Heathrow Terminal 5-Heathrow Terminals 123\", \"source\": \"Heathrow Terminal 5\", \"target\": \"Heathrow Terminals 123\", \"weight\": 2729.0}, {\"uid\": \"Heathrow Terminals 123-Hatton Cross\", \"source\": \"Heathrow Terminals 123\", \"target\": \"Hatton Cross\", \"weight\": 8351.0}, {\"uid\": \"Heathrow Terminals 123-Heathrow Terminal 4\", \"source\": \"Heathrow Terminals 123\", \"target\": \"Heathrow Terminal 4\", \"weight\": 32.0}, {\"uid\": \"Heathrow Terminals 123-Heathrow Terminal 5\", \"source\": \"Heathrow Terminals 123\", \"target\": \"Heathrow Terminal 5\", \"weight\": 2281.0}, {\"uid\": \"Hendon Central-Brent Cross\", \"source\": \"Hendon Central\", \"target\": \"Brent Cross\", \"weight\": 13740.0}, {\"uid\": \"Hendon Central-Colindale\", \"source\": \"Hendon Central\", \"target\": \"Colindale\", \"weight\": 8818.0}, {\"uid\": \"High Barnet-Totteridge & Whetstone\", \"source\": \"High Barnet\", \"target\": \"Totteridge & Whetstone\", \"weight\": 2489.0}, {\"uid\": \"High Street Kensington-Earl's Court\", \"source\": \"High Street Kensington\", \"target\": \"Earl's Court\", \"weight\": 10140.0}, {\"uid\": \"High Street Kensington-Gloucester Road\", \"source\": \"High Street Kensington\", \"target\": \"Gloucester Road\", \"weight\": 10749.0}, {\"uid\": \"High Street Kensington-Notting Hill Gate\", \"source\": \"High Street Kensington\", \"target\": \"Notting Hill Gate\", \"weight\": 18374.0}, {\"uid\": \"Highbury & Islington-Finsbury Park\", \"source\": \"Highbury & Islington\", \"target\": \"Finsbury Park\", \"weight\": 58332.0}, {\"uid\": \"Highbury & Islington-King's Cross St. Pancras\", \"source\": \"Highbury & Islington\", \"target\": \"King's Cross St. Pancras\", \"weight\": 70229.0}, {\"uid\": \"Highgate-Archway\", \"source\": \"Highgate\", \"target\": \"Archway\", \"weight\": 20469.0}, {\"uid\": \"Highgate-East Finchley\", \"source\": \"Highgate\", \"target\": \"East Finchley\", \"weight\": 16426.0}, {\"uid\": \"Hillingdon-Ickenham\", \"source\": \"Hillingdon\", \"target\": \"Ickenham\", \"weight\": 7071.0}, {\"uid\": \"Hillingdon-Uxbridge\", \"source\": \"Hillingdon\", \"target\": \"Uxbridge\", \"weight\": 5902.0}, {\"uid\": \"Holborn-Chancery Lane\", \"source\": \"Holborn\", \"target\": \"Chancery Lane\", \"weight\": 87373.0}, {\"uid\": \"Holborn-Covent Garden\", \"source\": \"Holborn\", \"target\": \"Covent Garden\", \"weight\": 14270.0}, {\"uid\": \"Holborn-Russell Square\", \"source\": \"Holborn\", \"target\": \"Russell Square\", \"weight\": 23732.0}, {\"uid\": \"Holborn-Tottenham Court Road\", \"source\": \"Holborn\", \"target\": \"Tottenham Court Road\", \"weight\": 93344.0}, {\"uid\": \"Holland Park-Notting Hill Gate\", \"source\": \"Holland Park\", \"target\": \"Notting Hill Gate\", \"weight\": 22203.0}, {\"uid\": \"Holland Park-Shepherd's Bush (Cen)\", \"source\": \"Holland Park\", \"target\": \"Shepherd's Bush (Cen)\", \"weight\": 18398.0}, {\"uid\": \"Holloway Road-Arsenal\", \"source\": \"Holloway Road\", \"target\": \"Arsenal\", \"weight\": 2968.0}, {\"uid\": \"Holloway Road-Caledonian Road\", \"source\": \"Holloway Road\", \"target\": \"Caledonian Road\", \"weight\": 5281.0}, {\"uid\": \"Hornchurch-Elm Park\", \"source\": \"Hornchurch\", \"target\": \"Elm Park\", \"weight\": 2194.0}, {\"uid\": \"Hornchurch-Upminster Bridge\", \"source\": \"Hornchurch\", \"target\": \"Upminster Bridge\", \"weight\": 3757.0}, {\"uid\": \"Hounslow Central-Hounslow East\", \"source\": \"Hounslow Central\", \"target\": \"Hounslow East\", \"weight\": 15862.0}, {\"uid\": \"Hounslow Central-Hounslow West\", \"source\": \"Hounslow Central\", \"target\": \"Hounslow West\", \"weight\": 11675.0}, {\"uid\": \"Hounslow East-Hounslow Central\", \"source\": \"Hounslow East\", \"target\": \"Hounslow Central\", \"weight\": 14128.0}, {\"uid\": \"Hounslow East-Osterley\", \"source\": \"Hounslow East\", \"target\": \"Osterley\", \"weight\": 17803.0}, {\"uid\": \"Hounslow West-Hounslow Central\", \"source\": \"Hounslow West\", \"target\": \"Hounslow Central\", \"weight\": 10245.0}, {\"uid\": \"Hounslow West-Hatton Cross\", \"source\": \"Hounslow West\", \"target\": \"Hatton Cross\", \"weight\": 13385.0}, {\"uid\": \"Hyde Park Corner-Green Park\", \"source\": \"Hyde Park Corner\", \"target\": \"Green Park\", \"weight\": 39198.0}, {\"uid\": \"Hyde Park Corner-Knightsbridge\", \"source\": \"Hyde Park Corner\", \"target\": \"Knightsbridge\", \"weight\": 34567.0}, {\"uid\": \"Ickenham-Hillingdon\", \"source\": \"Ickenham\", \"target\": \"Hillingdon\", \"weight\": 6890.0}, {\"uid\": \"Ickenham-Ruislip\", \"source\": \"Ickenham\", \"target\": \"Ruislip\", \"weight\": 7637.0}, {\"uid\": \"Kennington-Elephant & Castle\", \"source\": \"Kennington\", \"target\": \"Elephant & Castle\", \"weight\": 19679.0}, {\"uid\": \"Kennington-Oval\", \"source\": \"Kennington\", \"target\": \"Oval\", \"weight\": 36028.0}, {\"uid\": \"Kennington-Waterloo\", \"source\": \"Kennington\", \"target\": \"Waterloo\", \"weight\": 26589.0}, {\"uid\": \"Kensal Green-Queen's Park\", \"source\": \"Kensal Green\", \"target\": \"Queen's Park\", \"weight\": 13233.0}, {\"uid\": \"Kensal Green-Willesden Junction\", \"source\": \"Kensal Green\", \"target\": \"Willesden Junction\", \"weight\": 9034.0}, {\"uid\": \"Kentish Town-Camden Town\", \"source\": \"Kentish Town\", \"target\": \"Camden Town\", \"weight\": 33058.0}, {\"uid\": \"Kentish Town-Tufnell Park\", \"source\": \"Kentish Town\", \"target\": \"Tufnell Park\", \"weight\": 27059.0}, {\"uid\": \"Kenton-Harrow & Wealdstone\", \"source\": \"Kenton\", \"target\": \"Harrow & Wealdstone\", \"weight\": 2317.0}, {\"uid\": \"Kenton-South Kenton\", \"source\": \"Kenton\", \"target\": \"South Kenton\", \"weight\": 4220.0}, {\"uid\": \"Kew Gardens-Gunnersbury\", \"source\": \"Kew Gardens\", \"target\": \"Gunnersbury\", \"weight\": 10334.0}, {\"uid\": \"Kew Gardens-Richmond\", \"source\": \"Kew Gardens\", \"target\": \"Richmond\", \"weight\": 7981.0}, {\"uid\": \"Kilburn-West Hampstead\", \"source\": \"Kilburn\", \"target\": \"West Hampstead\", \"weight\": 3051.0}, {\"uid\": \"Kilburn-Willesden Green\", \"source\": \"Kilburn\", \"target\": \"Willesden Green\", \"weight\": 3333.0}, {\"uid\": \"Kilburn Park-Maida Vale\", \"source\": \"Kilburn Park\", \"target\": \"Maida Vale\", \"weight\": 18813.0}, {\"uid\": \"Kilburn Park-Queen's Park\", \"source\": \"Kilburn Park\", \"target\": \"Queen's Park\", \"weight\": 13815.0}, {\"uid\": \"King's Cross St. Pancras-Angel\", \"source\": \"King's Cross St. Pancras\", \"target\": \"Angel\", \"weight\": 46399.0}, {\"uid\": \"King's Cross St. Pancras-Russell Square\", \"source\": \"King's Cross St. Pancras\", \"target\": \"Russell Square\", \"weight\": 8325.0}, {\"uid\": \"King's Cross St. Pancras-Highbury & Islington\", \"source\": \"King's Cross St. Pancras\", \"target\": \"Highbury & Islington\", \"weight\": 100140.0}, {\"uid\": \"King's Cross St. Pancras-Farringdon\", \"source\": \"King's Cross St. Pancras\", \"target\": \"Farringdon\", \"weight\": 69700.0}, {\"uid\": \"King's Cross St. Pancras-Euston\", \"source\": \"King's Cross St. Pancras\", \"target\": \"Euston\", \"weight\": 45867.0}, {\"uid\": \"King's Cross St. Pancras-Caledonian Road\", \"source\": \"King's Cross St. Pancras\", \"target\": \"Caledonian Road\", \"weight\": 64100.0}, {\"uid\": \"King's Cross St. Pancras-Euston Square\", \"source\": \"King's Cross St. Pancras\", \"target\": \"Euston Square\", \"weight\": 22710.0}, {\"uid\": \"Kingsbury-Queensbury\", \"source\": \"Kingsbury\", \"target\": \"Queensbury\", \"weight\": 6944.0}, {\"uid\": \"Kingsbury-Wembley Park\", \"source\": \"Kingsbury\", \"target\": \"Wembley Park\", \"weight\": 9403.0}, {\"uid\": \"Knightsbridge-Hyde Park Corner\", \"source\": \"Knightsbridge\", \"target\": \"Hyde Park Corner\", \"weight\": 37486.0}, {\"uid\": \"Knightsbridge-South Kensington\", \"source\": \"Knightsbridge\", \"target\": \"South Kensington\", \"weight\": 26802.0}, {\"uid\": \"Ladbroke Grove-Westbourne Park\", \"source\": \"Ladbroke Grove\", \"target\": \"Westbourne Park\", \"weight\": 9230.0}, {\"uid\": \"Ladbroke Grove-Latimer Road\", \"source\": \"Ladbroke Grove\", \"target\": \"Latimer Road\", \"weight\": 13327.0}, {\"uid\": \"Lambeth North-Elephant & Castle\", \"source\": \"Lambeth North\", \"target\": \"Elephant & Castle\", \"weight\": 4115.0}, {\"uid\": \"Lambeth North-Waterloo\", \"source\": \"Lambeth North\", \"target\": \"Waterloo\", \"weight\": 6118.0}, {\"uid\": \"Lancaster Gate-Marble Arch\", \"source\": \"Lancaster Gate\", \"target\": \"Marble Arch\", \"weight\": 33903.0}, {\"uid\": \"Lancaster Gate-Queensway\", \"source\": \"Lancaster Gate\", \"target\": \"Queensway\", \"weight\": 28082.0}, {\"uid\": \"Latimer Road-Ladbroke Grove\", \"source\": \"Latimer Road\", \"target\": \"Ladbroke Grove\", \"weight\": 10895.0}, {\"uid\": \"Latimer Road-Wood Lane\", \"source\": \"Latimer Road\", \"target\": \"Wood Lane\", \"weight\": 8512.0}, {\"uid\": \"Leicester Square-Charing Cross\", \"source\": \"Leicester Square\", \"target\": \"Charing Cross\", \"weight\": 10518.0}, {\"uid\": \"Leicester Square-Tottenham Court Road\", \"source\": \"Leicester Square\", \"target\": \"Tottenham Court Road\", \"weight\": 11908.0}, {\"uid\": \"Leicester Square-Covent Garden\", \"source\": \"Leicester Square\", \"target\": \"Covent Garden\", \"weight\": 12318.0}, {\"uid\": \"Leicester Square-Piccadilly Circus\", \"source\": \"Leicester Square\", \"target\": \"Piccadilly Circus\", \"weight\": 16953.0}, {\"uid\": \"Leyton-Stratford\", \"source\": \"Leyton\", \"target\": \"Stratford\", \"weight\": 37005.0}, {\"uid\": \"Leyton-Leytonstone\", \"source\": \"Leyton\", \"target\": \"Leytonstone\", \"weight\": 47792.0}, {\"uid\": \"Leytonstone-Snaresbrook\", \"source\": \"Leytonstone\", \"target\": \"Snaresbrook\", \"weight\": 39942.0}, {\"uid\": \"Leytonstone-Leyton\", \"source\": \"Leytonstone\", \"target\": \"Leyton\", \"weight\": 17886.0}, {\"uid\": \"Leytonstone-Wanstead\", \"source\": \"Leytonstone\", \"target\": \"Wanstead\", \"weight\": 13185.0}, {\"uid\": \"Liverpool Street-Aldgate\", \"source\": \"Liverpool Street\", \"target\": \"Aldgate\", \"weight\": 5064.0}, {\"uid\": \"Liverpool Street-Aldgate East\", \"source\": \"Liverpool Street\", \"target\": \"Aldgate East\", \"weight\": 38618.0}, {\"uid\": \"Liverpool Street-Bank / Monument\", \"source\": \"Liverpool Street\", \"target\": \"Bank / Monument\", \"weight\": 103615.0}, {\"uid\": \"Liverpool Street-Bethnal Green\", \"source\": \"Liverpool Street\", \"target\": \"Bethnal Green\", \"weight\": 47488.0}, {\"uid\": \"Liverpool Street-Moorgate\", \"source\": \"Liverpool Street\", \"target\": \"Moorgate\", \"weight\": 44716.0}, {\"uid\": \"Liverpool Street-Tottenham Hale\", \"source\": \"Liverpool Street\", \"target\": \"Tottenham Hale\", \"weight\": 25997.0}, {\"uid\": \"London Bridge-Bank / Monument\", \"source\": \"London Bridge\", \"target\": \"Bank / Monument\", \"weight\": 119970.0}, {\"uid\": \"London Bridge-Bermondsey\", \"source\": \"London Bridge\", \"target\": \"Bermondsey\", \"weight\": 69070.0}, {\"uid\": \"London Bridge-Borough\", \"source\": \"London Bridge\", \"target\": \"Borough\", \"weight\": 20745.0}, {\"uid\": \"London Bridge-Southwark\", \"source\": \"London Bridge\", \"target\": \"Southwark\", \"weight\": 149944.0}, {\"uid\": \"Loughton-Buckhurst Hill\", \"source\": \"Loughton\", \"target\": \"Buckhurst Hill\", \"weight\": 7345.0}, {\"uid\": \"Loughton-Debden\", \"source\": \"Loughton\", \"target\": \"Debden\", \"weight\": 4463.0}, {\"uid\": \"Maida Vale-Kilburn Park\", \"source\": \"Maida Vale\", \"target\": \"Kilburn Park\", \"weight\": 15483.0}, {\"uid\": \"Maida Vale-Warwick Avenue\", \"source\": \"Maida Vale\", \"target\": \"Warwick Avenue\", \"weight\": 21133.0}, {\"uid\": \"Manor House-Turnpike Lane\", \"source\": \"Manor House\", \"target\": \"Turnpike Lane\", \"weight\": 34296.0}, {\"uid\": \"Manor House-Finsbury Park\", \"source\": \"Manor House\", \"target\": \"Finsbury Park\", \"weight\": 27865.0}, {\"uid\": \"Mansion House-Blackfriars\", \"source\": \"Mansion House\", \"target\": \"Blackfriars\", \"weight\": 10240.0}, {\"uid\": \"Mansion House-Cannon Street\", \"source\": \"Mansion House\", \"target\": \"Cannon Street\", \"weight\": 5976.0}, {\"uid\": \"Marble Arch-Bond Street\", \"source\": \"Marble Arch\", \"target\": \"Bond Street\", \"weight\": 41988.0}, {\"uid\": \"Marble Arch-Lancaster Gate\", \"source\": \"Marble Arch\", \"target\": \"Lancaster Gate\", \"weight\": 31929.0}, {\"uid\": \"Marylebone-Baker Street\", \"source\": \"Marylebone\", \"target\": \"Baker Street\", \"weight\": 22664.0}, {\"uid\": \"Marylebone-Edgware Road (Bak)\", \"source\": \"Marylebone\", \"target\": \"Edgware Road (Bak)\", \"weight\": 6680.0}, {\"uid\": \"Marylebone-HarrowOnTheHill\", \"source\": \"Marylebone\", \"target\": \"HarrowOnTheHill\", \"weight\": 14147.0}, {\"uid\": \"Mile End-Bow Road\", \"source\": \"Mile End\", \"target\": \"Bow Road\", \"weight\": 48725.0}, {\"uid\": \"Mile End-Stratford\", \"source\": \"Mile End\", \"target\": \"Stratford\", \"weight\": 4447.0}, {\"uid\": \"Mile End-Stepney Green\", \"source\": \"Mile End\", \"target\": \"Stepney Green\", \"weight\": 2026.0}, {\"uid\": \"Mile End-Bethnal Green\", \"source\": \"Mile End\", \"target\": \"Bethnal Green\", \"weight\": 37984.0}, {\"uid\": \"Mill Hill East-Finchley Central\", \"source\": \"Mill Hill East\", \"target\": \"Finchley Central\", \"weight\": 821.0}, {\"uid\": \"Moor Park-Croxley\", \"source\": \"Moor Park\", \"target\": \"Croxley\", \"weight\": 1949.0}, {\"uid\": \"Moor Park-HarrowOnTheHill\", \"source\": \"Moor Park\", \"target\": \"HarrowOnTheHill\", \"weight\": 4460.0}, {\"uid\": \"Moor Park-Northwood\", \"source\": \"Moor Park\", \"target\": \"Northwood\", \"weight\": 2448.0}, {\"uid\": \"Moor Park-Rickmansworth\", \"source\": \"Moor Park\", \"target\": \"Rickmansworth\", \"weight\": 257.0}, {\"uid\": \"Moorgate-Old Street\", \"source\": \"Moorgate\", \"target\": \"Old Street\", \"weight\": 41226.0}, {\"uid\": \"Moorgate-Liverpool Street\", \"source\": \"Moorgate\", \"target\": \"Liverpool Street\", \"weight\": 40586.0}, {\"uid\": \"Moorgate-Bank / Monument\", \"source\": \"Moorgate\", \"target\": \"Bank / Monument\", \"weight\": 42874.0}, {\"uid\": \"Moorgate-Barbican\", \"source\": \"Moorgate\", \"target\": \"Barbican\", \"weight\": 44354.0}, {\"uid\": \"Morden-South Wimbledon\", \"source\": \"Morden\", \"target\": \"South Wimbledon\", \"weight\": 5995.0}, {\"uid\": \"Mornington Crescent-Camden Town\", \"source\": \"Mornington Crescent\", \"target\": \"Camden Town\", \"weight\": 1005.0}, {\"uid\": \"Mornington Crescent-Euston\", \"source\": \"Mornington Crescent\", \"target\": \"Euston\", \"weight\": 2527.0}, {\"uid\": \"Neasden-Dollis Hill\", \"source\": \"Neasden\", \"target\": \"Dollis Hill\", \"weight\": 392.0}, {\"uid\": \"Neasden-Wembley Park\", \"source\": \"Neasden\", \"target\": \"Wembley Park\", \"weight\": 2411.0}, {\"uid\": \"Neasden-Willesden Green\", \"source\": \"Neasden\", \"target\": \"Willesden Green\", \"weight\": 1619.0}, {\"uid\": \"Newbury Park-Barkingside\", \"source\": \"Newbury Park\", \"target\": \"Barkingside\", \"weight\": 2809.0}, {\"uid\": \"Newbury Park-Gants Hill\", \"source\": \"Newbury Park\", \"target\": \"Gants Hill\", \"weight\": 6495.0}, {\"uid\": \"North Acton-Hanger Lane\", \"source\": \"North Acton\", \"target\": \"Hanger Lane\", \"weight\": 7412.0}, {\"uid\": \"North Acton-West Acton\", \"source\": \"North Acton\", \"target\": \"West Acton\", \"weight\": 12928.0}, {\"uid\": \"North Acton-East Acton\", \"source\": \"North Acton\", \"target\": \"East Acton\", \"weight\": 15116.0}, {\"uid\": \"North Ealing-Ealing Common\", \"source\": \"North Ealing\", \"target\": \"Ealing Common\", \"weight\": 6137.0}, {\"uid\": \"North Ealing-Park Royal\", \"source\": \"North Ealing\", \"target\": \"Park Royal\", \"weight\": 5339.0}, {\"uid\": \"North Greenwich-Canary Wharf\", \"source\": \"North Greenwich\", \"target\": \"Canary Wharf\", \"weight\": 38206.0}, {\"uid\": \"North Greenwich-Canning Town\", \"source\": \"North Greenwich\", \"target\": \"Canning Town\", \"weight\": 28038.0}, {\"uid\": \"North Harrow-HarrowOnTheHill\", \"source\": \"North Harrow\", \"target\": \"HarrowOnTheHill\", \"weight\": 3787.0}, {\"uid\": \"North Harrow-Pinner\", \"source\": \"North Harrow\", \"target\": \"Pinner\", \"weight\": 2388.0}, {\"uid\": \"North Wembley-Wembley Central\", \"source\": \"North Wembley\", \"target\": \"Wembley Central\", \"weight\": 3716.0}, {\"uid\": \"North Wembley-South Kenton\", \"source\": \"North Wembley\", \"target\": \"South Kenton\", \"weight\": 5794.0}, {\"uid\": \"Northfields-Boston Manor\", \"source\": \"Northfields\", \"target\": \"Boston Manor\", \"weight\": 18365.0}, {\"uid\": \"Northfields-South Ealing\", \"source\": \"Northfields\", \"target\": \"South Ealing\", \"weight\": 22600.0}, {\"uid\": \"Northolt-Greenford\", \"source\": \"Northolt\", \"target\": \"Greenford\", \"weight\": 6831.0}, {\"uid\": \"Northolt-South Ruislip\", \"source\": \"Northolt\", \"target\": \"South Ruislip\", \"weight\": 2982.0}, {\"uid\": \"Northwick Park-HarrowOnTheHill\", \"source\": \"Northwick Park\", \"target\": \"HarrowOnTheHill\", \"weight\": 3401.0}, {\"uid\": \"Northwick Park-Preston Road\", \"source\": \"Northwick Park\", \"target\": \"Preston Road\", \"weight\": 401.0}, {\"uid\": \"Northwood-Moor Park\", \"source\": \"Northwood\", \"target\": \"Moor Park\", \"weight\": 2589.0}, {\"uid\": \"Northwood-Northwood Hills\", \"source\": \"Northwood\", \"target\": \"Northwood Hills\", \"weight\": 821.0}, {\"uid\": \"Northwood Hills-Pinner\", \"source\": \"Northwood Hills\", \"target\": \"Pinner\", \"weight\": 820.0}, {\"uid\": \"Northwood Hills-Northwood\", \"source\": \"Northwood Hills\", \"target\": \"Northwood\", \"weight\": 732.0}, {\"uid\": \"Notting Hill Gate-Bayswater\", \"source\": \"Notting Hill Gate\", \"target\": \"Bayswater\", \"weight\": 16982.0}, {\"uid\": \"Notting Hill Gate-High Street Kensington\", \"source\": \"Notting Hill Gate\", \"target\": \"High Street Kensington\", \"weight\": 18772.0}, {\"uid\": \"Notting Hill Gate-Holland Park\", \"source\": \"Notting Hill Gate\", \"target\": \"Holland Park\", \"weight\": 20416.0}, {\"uid\": \"Notting Hill Gate-Queensway\", \"source\": \"Notting Hill Gate\", \"target\": \"Queensway\", \"weight\": 25326.0}, {\"uid\": \"Oakwood-Cockfosters\", \"source\": \"Oakwood\", \"target\": \"Cockfosters\", \"weight\": 1423.0}, {\"uid\": \"Oakwood-Southgate\", \"source\": \"Oakwood\", \"target\": \"Southgate\", \"weight\": 3354.0}, {\"uid\": \"Old Street-Angel\", \"source\": \"Old Street\", \"target\": \"Angel\", \"weight\": 45589.0}, {\"uid\": \"Old Street-Moorgate\", \"source\": \"Old Street\", \"target\": \"Moorgate\", \"weight\": 44289.0}, {\"uid\": \"Osterley-Hounslow East\", \"source\": \"Osterley\", \"target\": \"Hounslow East\", \"weight\": 19191.0}, {\"uid\": \"Osterley-Boston Manor\", \"source\": \"Osterley\", \"target\": \"Boston Manor\", \"weight\": 16034.0}, {\"uid\": \"Oval-Kennington\", \"source\": \"Oval\", \"target\": \"Kennington\", \"weight\": 35716.0}, {\"uid\": \"Oval-Stockwell\", \"source\": \"Oval\", \"target\": \"Stockwell\", \"weight\": 33916.0}, {\"uid\": \"Oxford Circus-Bond Street\", \"source\": \"Oxford Circus\", \"target\": \"Bond Street\", \"weight\": 17349.0}, {\"uid\": \"Oxford Circus-Green Park\", \"source\": \"Oxford Circus\", \"target\": \"Green Park\", \"weight\": 97608.0}, {\"uid\": \"Oxford Circus-Piccadilly Circus\", \"source\": \"Oxford Circus\", \"target\": \"Piccadilly Circus\", \"weight\": 20837.0}, {\"uid\": \"Oxford Circus-Regent's Park\", \"source\": \"Oxford Circus\", \"target\": \"Regent's Park\", \"weight\": 12108.0}, {\"uid\": \"Oxford Circus-Tottenham Court Road\", \"source\": \"Oxford Circus\", \"target\": \"Tottenham Court Road\", \"weight\": 23748.0}, {\"uid\": \"Oxford Circus-Warren Street\", \"source\": \"Oxford Circus\", \"target\": \"Warren Street\", \"weight\": 98480.0}, {\"uid\": \"Paddington-Edgware Road (Cir)\", \"source\": \"Paddington\", \"target\": \"Edgware Road (Cir)\", \"weight\": 17622.0}, {\"uid\": \"Paddington-Warwick Avenue\", \"source\": \"Paddington\", \"target\": \"Warwick Avenue\", \"weight\": 54430.0}, {\"uid\": \"Paddington-Royal Oak\", \"source\": \"Paddington\", \"target\": \"Royal Oak\", \"weight\": 4503.0}, {\"uid\": \"Paddington-Edgware Road (Bak)\", \"source\": \"Paddington\", \"target\": \"Edgware Road (Bak)\", \"weight\": 114706.0}, {\"uid\": \"Paddington-Ealing Broadway\", \"source\": \"Paddington\", \"target\": \"Ealing Broadway\", \"weight\": 13096.0}, {\"uid\": \"Paddington-Bayswater\", \"source\": \"Paddington\", \"target\": \"Bayswater\", \"weight\": 19701.0}, {\"uid\": \"Park Royal-Alperton\", \"source\": \"Park Royal\", \"target\": \"Alperton\", \"weight\": 4656.0}, {\"uid\": \"Park Royal-North Ealing\", \"source\": \"Park Royal\", \"target\": \"North Ealing\", \"weight\": 5856.0}, {\"uid\": \"Parsons Green-Fulham Broadway\", \"source\": \"Parsons Green\", \"target\": \"Fulham Broadway\", \"weight\": 22556.0}, {\"uid\": \"Parsons Green-Putney Bridge\", \"source\": \"Parsons Green\", \"target\": \"Putney Bridge\", \"weight\": 17697.0}, {\"uid\": \"Perivale-Greenford\", \"source\": \"Perivale\", \"target\": \"Greenford\", \"weight\": 9158.0}, {\"uid\": \"Perivale-Hanger Lane\", \"source\": \"Perivale\", \"target\": \"Hanger Lane\", \"weight\": 11007.0}, {\"uid\": \"Piccadilly Circus-Charing Cross\", \"source\": \"Piccadilly Circus\", \"target\": \"Charing Cross\", \"weight\": 11307.0}, {\"uid\": \"Piccadilly Circus-Oxford Circus\", \"source\": \"Piccadilly Circus\", \"target\": \"Oxford Circus\", \"weight\": 22175.0}, {\"uid\": \"Piccadilly Circus-Green Park\", \"source\": \"Piccadilly Circus\", \"target\": \"Green Park\", \"weight\": 13579.0}, {\"uid\": \"Piccadilly Circus-Leicester Square\", \"source\": \"Piccadilly Circus\", \"target\": \"Leicester Square\", \"weight\": 20688.0}, {\"uid\": \"Pimlico-Vauxhall\", \"source\": \"Pimlico\", \"target\": \"Vauxhall\", \"weight\": 48034.0}, {\"uid\": \"Pimlico-Victoria\", \"source\": \"Pimlico\", \"target\": \"Victoria\", \"weight\": 58118.0}, {\"uid\": \"Pinner-North Harrow\", \"source\": \"Pinner\", \"target\": \"North Harrow\", \"weight\": 2530.0}, {\"uid\": \"Pinner-Northwood Hills\", \"source\": \"Pinner\", \"target\": \"Northwood Hills\", \"weight\": 615.0}, {\"uid\": \"Plaistow-Upton Park\", \"source\": \"Plaistow\", \"target\": \"Upton Park\", \"weight\": 6649.0}, {\"uid\": \"Plaistow-West Ham\", \"source\": \"Plaistow\", \"target\": \"West Ham\", \"weight\": 11918.0}, {\"uid\": \"Preston Road-Northwick Park\", \"source\": \"Preston Road\", \"target\": \"Northwick Park\", \"weight\": 520.0}, {\"uid\": \"Preston Road-Wembley Park\", \"source\": \"Preston Road\", \"target\": \"Wembley Park\", \"weight\": 2159.0}, {\"uid\": \"Putney Bridge-East Putney\", \"source\": \"Putney Bridge\", \"target\": \"East Putney\", \"weight\": 14235.0}, {\"uid\": \"Putney Bridge-Parsons Green\", \"source\": \"Putney Bridge\", \"target\": \"Parsons Green\", \"weight\": 18991.0}, {\"uid\": \"Queen's Park-Kensal Green\", \"source\": \"Queen's Park\", \"target\": \"Kensal Green\", \"weight\": 10274.0}, {\"uid\": \"Queen's Park-Kilburn Park\", \"source\": \"Queen's Park\", \"target\": \"Kilburn Park\", \"weight\": 16958.0}, {\"uid\": \"Queensbury-Canons Park\", \"source\": \"Queensbury\", \"target\": \"Canons Park\", \"weight\": 4278.0}, {\"uid\": \"Queensbury-Kingsbury\", \"source\": \"Queensbury\", \"target\": \"Kingsbury\", \"weight\": 6813.0}, {\"uid\": \"Queensway-Lancaster Gate\", \"source\": \"Queensway\", \"target\": \"Lancaster Gate\", \"weight\": 29898.0}, {\"uid\": \"Queensway-Notting Hill Gate\", \"source\": \"Queensway\", \"target\": \"Notting Hill Gate\", \"weight\": 23388.0}, {\"uid\": \"Ravenscourt Park-Hammersmith (Dis)\", \"source\": \"Ravenscourt Park\", \"target\": \"Hammersmith (Dis)\", \"weight\": 2800.0}, {\"uid\": \"Ravenscourt Park-Stamford Brook\", \"source\": \"Ravenscourt Park\", \"target\": \"Stamford Brook\", \"weight\": 667.0}, {\"uid\": \"Rayners Lane-Eastcote\", \"source\": \"Rayners Lane\", \"target\": \"Eastcote\", \"weight\": 10046.0}, {\"uid\": \"Rayners Lane-South Harrow\", \"source\": \"Rayners Lane\", \"target\": \"South Harrow\", \"weight\": 3554.0}, {\"uid\": \"Rayners Lane-West Harrow\", \"source\": \"Rayners Lane\", \"target\": \"West Harrow\", \"weight\": 12129.0}, {\"uid\": \"Redbridge-Gants Hill\", \"source\": \"Redbridge\", \"target\": \"Gants Hill\", \"weight\": 9687.0}, {\"uid\": \"Redbridge-Wanstead\", \"source\": \"Redbridge\", \"target\": \"Wanstead\", \"weight\": 13023.0}, {\"uid\": \"Regent's Park-Baker Street\", \"source\": \"Regent's Park\", \"target\": \"Baker Street\", \"weight\": 10775.0}, {\"uid\": \"Regent's Park-Oxford Circus\", \"source\": \"Regent's Park\", \"target\": \"Oxford Circus\", \"weight\": 11910.0}, {\"uid\": \"Richmond-Kew Gardens\", \"source\": \"Richmond\", \"target\": \"Kew Gardens\", \"weight\": 8665.0}, {\"uid\": \"Rickmansworth-Chorleywood\", \"source\": \"Rickmansworth\", \"target\": \"Chorleywood\", \"weight\": 3150.0}, {\"uid\": \"Rickmansworth-HarrowOnTheHill\", \"source\": \"Rickmansworth\", \"target\": \"HarrowOnTheHill\", \"weight\": 4984.0}, {\"uid\": \"Rickmansworth-Moor Park\", \"source\": \"Rickmansworth\", \"target\": \"Moor Park\", \"weight\": 206.0}, {\"uid\": \"Roding Valley-Chigwell\", \"source\": \"Roding Valley\", \"target\": \"Chigwell\", \"weight\": 2096.0}, {\"uid\": \"Roding Valley-Woodford\", \"source\": \"Roding Valley\", \"target\": \"Woodford\", \"weight\": 2260.0}, {\"uid\": \"Royal Oak-Paddington\", \"source\": \"Royal Oak\", \"target\": \"Paddington\", \"weight\": 15766.0}, {\"uid\": \"Royal Oak-Westbourne Park\", \"source\": \"Royal Oak\", \"target\": \"Westbourne Park\", \"weight\": 12439.0}, {\"uid\": \"Ruislip-Ruislip Manor\", \"source\": \"Ruislip\", \"target\": \"Ruislip Manor\", \"weight\": 7513.0}, {\"uid\": \"Ruislip-Ickenham\", \"source\": \"Ruislip\", \"target\": \"Ickenham\", \"weight\": 8264.0}, {\"uid\": \"Ruislip Gardens-South Ruislip\", \"source\": \"Ruislip Gardens\", \"target\": \"South Ruislip\", \"weight\": 845.0}, {\"uid\": \"Ruislip Gardens-West Ruislip\", \"source\": \"Ruislip Gardens\", \"target\": \"West Ruislip\", \"weight\": 26.0}, {\"uid\": \"Ruislip Manor-Eastcote\", \"source\": \"Ruislip Manor\", \"target\": \"Eastcote\", \"weight\": 8751.0}, {\"uid\": \"Ruislip Manor-Ruislip\", \"source\": \"Ruislip Manor\", \"target\": \"Ruislip\", \"weight\": 8215.0}, {\"uid\": \"Russell Square-Holborn\", \"source\": \"Russell Square\", \"target\": \"Holborn\", \"weight\": 24825.0}, {\"uid\": \"Russell Square-King's Cross St. Pancras\", \"source\": \"Russell Square\", \"target\": \"King's Cross St. Pancras\", \"weight\": 21845.0}, {\"uid\": \"Seven Sisters-Finsbury Park\", \"source\": \"Seven Sisters\", \"target\": \"Finsbury Park\", \"weight\": 35227.0}, {\"uid\": \"Seven Sisters-Tottenham Hale\", \"source\": \"Seven Sisters\", \"target\": \"Tottenham Hale\", \"weight\": 24900.0}, {\"uid\": \"Shepherd's Bush (Cen)-White City\", \"source\": \"Shepherd's Bush (Cen)\", \"target\": \"White City\", \"weight\": 20050.0}, {\"uid\": \"Shepherd's Bush (Cen)-Holland Park\", \"source\": \"Shepherd's Bush (Cen)\", \"target\": \"Holland Park\", \"weight\": 7992.0}, {\"uid\": \"Shepherd's Bush Market-Goldhawk Road\", \"source\": \"Shepherd's Bush Market\", \"target\": \"Goldhawk Road\", \"weight\": 5106.0}, {\"uid\": \"Shepherd's Bush Market-Wood Lane\", \"source\": \"Shepherd's Bush Market\", \"target\": \"Wood Lane\", \"weight\": 8594.0}, {\"uid\": \"Sloane Square-South Kensington\", \"source\": \"Sloane Square\", \"target\": \"South Kensington\", \"weight\": 54258.0}, {\"uid\": \"Sloane Square-Victoria\", \"source\": \"Sloane Square\", \"target\": \"Victoria\", \"weight\": 63042.0}, {\"uid\": \"Snaresbrook-Leytonstone\", \"source\": \"Snaresbrook\", \"target\": \"Leytonstone\", \"weight\": 19127.0}, {\"uid\": \"Snaresbrook-South Woodford\", \"source\": \"Snaresbrook\", \"target\": \"South Woodford\", \"weight\": 16068.0}, {\"uid\": \"South Ealing-Acton Town\", \"source\": \"South Ealing\", \"target\": \"Acton Town\", \"weight\": 24588.0}, {\"uid\": \"South Ealing-Northfields\", \"source\": \"South Ealing\", \"target\": \"Northfields\", \"weight\": 20899.0}, {\"uid\": \"South Harrow-Sudbury Hill\", \"source\": \"South Harrow\", \"target\": \"Sudbury Hill\", \"weight\": 3014.0}, {\"uid\": \"South Harrow-Rayners Lane\", \"source\": \"South Harrow\", \"target\": \"Rayners Lane\", \"weight\": 3864.0}, {\"uid\": \"South Kensington-Gloucester Road\", \"source\": \"South Kensington\", \"target\": \"Gloucester Road\", \"weight\": 66555.0}, {\"uid\": \"South Kensington-Sloane Square\", \"source\": \"South Kensington\", \"target\": \"Sloane Square\", \"weight\": 29418.0}, {\"uid\": \"South Kensington-Knightsbridge\", \"source\": \"South Kensington\", \"target\": \"Knightsbridge\", \"weight\": 56674.0}, {\"uid\": \"South Kenton-Kenton\", \"source\": \"South Kenton\", \"target\": \"Kenton\", \"weight\": 3345.0}, {\"uid\": \"South Kenton-North Wembley\", \"source\": \"South Kenton\", \"target\": \"North Wembley\", \"weight\": 4874.0}, {\"uid\": \"South Ruislip-Northolt\", \"source\": \"South Ruislip\", \"target\": \"Northolt\", \"weight\": 3434.0}, {\"uid\": \"South Ruislip-West Ruislip\", \"source\": \"South Ruislip\", \"target\": \"West Ruislip\", \"weight\": 740.0}, {\"uid\": \"South Ruislip-Ruislip Gardens\", \"source\": \"South Ruislip\", \"target\": \"Ruislip Gardens\", \"weight\": 1066.0}, {\"uid\": \"South Wimbledon-Colliers Wood\", \"source\": \"South Wimbledon\", \"target\": \"Colliers Wood\", \"weight\": 8907.0}, {\"uid\": \"South Wimbledon-Morden\", \"source\": \"South Wimbledon\", \"target\": \"Morden\", \"weight\": 5796.0}, {\"uid\": \"South Woodford-Snaresbrook\", \"source\": \"South Woodford\", \"target\": \"Snaresbrook\", \"weight\": 16837.0}, {\"uid\": \"South Woodford-Woodford\", \"source\": \"South Woodford\", \"target\": \"Woodford\", \"weight\": 12810.0}, {\"uid\": \"Southfields-East Putney\", \"source\": \"Southfields\", \"target\": \"East Putney\", \"weight\": 13213.0}, {\"uid\": \"Southfields-Wimbledon Park\", \"source\": \"Southfields\", \"target\": \"Wimbledon Park\", \"weight\": 9365.0}, {\"uid\": \"Southgate-Arnos Grove\", \"source\": \"Southgate\", \"target\": \"Arnos Grove\", \"weight\": 7123.0}, {\"uid\": \"Southgate-Oakwood\", \"source\": \"Southgate\", \"target\": \"Oakwood\", \"weight\": 3132.0}, {\"uid\": \"Southwark-London Bridge\", \"source\": \"Southwark\", \"target\": \"London Bridge\", \"weight\": 145112.0}, {\"uid\": \"Southwark-Waterloo\", \"source\": \"Southwark\", \"target\": \"Waterloo\", \"weight\": 150085.0}, {\"uid\": \"St. James's Park-Victoria\", \"source\": \"St. James's Park\", \"target\": \"Victoria\", \"weight\": 26317.0}, {\"uid\": \"St. James's Park-Westminster\", \"source\": \"St. James's Park\", \"target\": \"Westminster\", \"weight\": 30308.0}, {\"uid\": \"St. John's Wood-Baker Street\", \"source\": \"St. John's Wood\", \"target\": \"Baker Street\", \"weight\": 7302.0}, {\"uid\": \"St. John's Wood-Swiss Cottage\", \"source\": \"St. John's Wood\", \"target\": \"Swiss Cottage\", \"weight\": 2920.0}, {\"uid\": \"St. Paul's-Chancery Lane\", \"source\": \"St. Paul's\", \"target\": \"Chancery Lane\", \"weight\": 81735.0}, {\"uid\": \"St. Paul's-Bank / Monument\", \"source\": \"St. Paul's\", \"target\": \"Bank / Monument\", \"weight\": 86446.0}, {\"uid\": \"Stamford Brook-Ravenscourt Park\", \"source\": \"Stamford Brook\", \"target\": \"Ravenscourt Park\", \"weight\": 663.0}, {\"uid\": \"Stamford Brook-Turnham Green\", \"source\": \"Stamford Brook\", \"target\": \"Turnham Green\", \"weight\": 1808.0}, {\"uid\": \"Stanmore-Canons Park\", \"source\": \"Stanmore\", \"target\": \"Canons Park\", \"weight\": 2603.0}, {\"uid\": \"Stepney Green-Mile End\", \"source\": \"Stepney Green\", \"target\": \"Mile End\", \"weight\": 2027.0}, {\"uid\": \"Stepney Green-Whitechapel\", \"source\": \"Stepney Green\", \"target\": \"Whitechapel\", \"weight\": 2858.0}, {\"uid\": \"Stockwell-Vauxhall\", \"source\": \"Stockwell\", \"target\": \"Vauxhall\", \"weight\": 18802.0}, {\"uid\": \"Stockwell-Oval\", \"source\": \"Stockwell\", \"target\": \"Oval\", \"weight\": 46573.0}, {\"uid\": \"Stockwell-Clapham North\", \"source\": \"Stockwell\", \"target\": \"Clapham North\", \"weight\": 33327.0}, {\"uid\": \"Stockwell-Brixton\", \"source\": \"Stockwell\", \"target\": \"Brixton\", \"weight\": 41648.0}, {\"uid\": \"Stonebridge Park-Harlesden\", \"source\": \"Stonebridge Park\", \"target\": \"Harlesden\", \"weight\": 9549.0}, {\"uid\": \"Stonebridge Park-Wembley Central\", \"source\": \"Stonebridge Park\", \"target\": \"Wembley Central\", \"weight\": 6467.0}, {\"uid\": \"Stratford-Leyton\", \"source\": \"Stratford\", \"target\": \"Leyton\", \"weight\": 44531.0}, {\"uid\": \"Stratford-Mile End\", \"source\": \"Stratford\", \"target\": \"Mile End\", \"weight\": 41089.0}, {\"uid\": \"Stratford-West Ham\", \"source\": \"Stratford\", \"target\": \"West Ham\", \"weight\": 48185.0}, {\"uid\": \"Stratford-Whitechapel\", \"source\": \"Stratford\", \"target\": \"Whitechapel\", \"weight\": 61857.0}, {\"uid\": \"Sudbury Hill-South Harrow\", \"source\": \"Sudbury Hill\", \"target\": \"South Harrow\", \"weight\": 3351.0}, {\"uid\": \"Sudbury Hill-Sudbury Town\", \"source\": \"Sudbury Hill\", \"target\": \"Sudbury Town\", \"weight\": 3754.0}, {\"uid\": \"Sudbury Town-Alperton\", \"source\": \"Sudbury Town\", \"target\": \"Alperton\", \"weight\": 4128.0}, {\"uid\": \"Sudbury Town-Sudbury Hill\", \"source\": \"Sudbury Town\", \"target\": \"Sudbury Hill\", \"weight\": 3202.0}, {\"uid\": \"Swiss Cottage-Finchley Road\", \"source\": \"Swiss Cottage\", \"target\": \"Finchley Road\", \"weight\": 4110.0}, {\"uid\": \"Swiss Cottage-St. John's Wood\", \"source\": \"Swiss Cottage\", \"target\": \"St. John's Wood\", \"weight\": 2988.0}, {\"uid\": \"Temple-Blackfriars\", \"source\": \"Temple\", \"target\": \"Blackfriars\", \"weight\": 11952.0}, {\"uid\": \"Temple-Embankment\", \"source\": \"Temple\", \"target\": \"Embankment\", \"weight\": 14927.0}, {\"uid\": \"Theydon Bois-Debden\", \"source\": \"Theydon Bois\", \"target\": \"Debden\", \"weight\": 3278.0}, {\"uid\": \"Theydon Bois-Epping\", \"source\": \"Theydon Bois\", \"target\": \"Epping\", \"weight\": 2151.0}, {\"uid\": \"Tooting Bec-Balham\", \"source\": \"Tooting Bec\", \"target\": \"Balham\", \"weight\": 26422.0}, {\"uid\": \"Tooting Bec-Tooting Broadway\", \"source\": \"Tooting Bec\", \"target\": \"Tooting Broadway\", \"weight\": 21424.0}, {\"uid\": \"Tooting Broadway-Colliers Wood\", \"source\": \"Tooting Broadway\", \"target\": \"Colliers Wood\", \"weight\": 12338.0}, {\"uid\": \"Tooting Broadway-Tooting Bec\", \"source\": \"Tooting Broadway\", \"target\": \"Tooting Bec\", \"weight\": 22044.0}, {\"uid\": \"Tottenham Court Road-Bond Street\", \"source\": \"Tottenham Court Road\", \"target\": \"Bond Street\", \"weight\": 71788.0}, {\"uid\": \"Tottenham Court Road-Oxford Circus\", \"source\": \"Tottenham Court Road\", \"target\": \"Oxford Circus\", \"weight\": 7461.0}, {\"uid\": \"Tottenham Court Road-Leicester Square\", \"source\": \"Tottenham Court Road\", \"target\": \"Leicester Square\", \"weight\": 96031.0}, {\"uid\": \"Tottenham Court Road-Holborn\", \"source\": \"Tottenham Court Road\", \"target\": \"Holborn\", \"weight\": 16984.0}, {\"uid\": \"Tottenham Court Road-Goodge Street\", \"source\": \"Tottenham Court Road\", \"target\": \"Goodge Street\", \"weight\": 24544.0}, {\"uid\": \"Tottenham Hale-Blackhorse Road\", \"source\": \"Tottenham Hale\", \"target\": \"Blackhorse Road\", \"weight\": 16787.0}, {\"uid\": \"Tottenham Hale-Liverpool Street\", \"source\": \"Tottenham Hale\", \"target\": \"Liverpool Street\", \"weight\": 24343.0}, {\"uid\": \"Tottenham Hale-Seven Sisters\", \"source\": \"Tottenham Hale\", \"target\": \"Seven Sisters\", \"weight\": 27110.0}, {\"uid\": \"Totteridge & Whetstone-High Barnet\", \"source\": \"Totteridge & Whetstone\", \"target\": \"High Barnet\", \"weight\": 2374.0}, {\"uid\": \"Totteridge & Whetstone-Woodside Park\", \"source\": \"Totteridge & Whetstone\", \"target\": \"Woodside Park\", \"weight\": 4085.0}, {\"uid\": \"Tower Hill-Aldgate\", \"source\": \"Tower Hill\", \"target\": \"Aldgate\", \"weight\": 906.0}, {\"uid\": \"Tower Hill-Aldgate East\", \"source\": \"Tower Hill\", \"target\": \"Aldgate East\", \"weight\": 23530.0}, {\"uid\": \"Tower Hill-Bank / Monument\", \"source\": \"Tower Hill\", \"target\": \"Bank / Monument\", \"weight\": 39063.0}, {\"uid\": \"Tufnell Park-Archway\", \"source\": \"Tufnell Park\", \"target\": \"Archway\", \"weight\": 24832.0}, {\"uid\": \"Tufnell Park-Kentish Town\", \"source\": \"Tufnell Park\", \"target\": \"Kentish Town\", \"weight\": 28263.0}, {\"uid\": \"Turnham Green-Hammersmith (Dis)\", \"source\": \"Turnham Green\", \"target\": \"Hammersmith (Dis)\", \"weight\": 10093.0}, {\"uid\": \"Turnham Green-Stamford Brook\", \"source\": \"Turnham Green\", \"target\": \"Stamford Brook\", \"weight\": 201.0}, {\"uid\": \"Turnham Green-Gunnersbury\", \"source\": \"Turnham Green\", \"target\": \"Gunnersbury\", \"weight\": 11923.0}, {\"uid\": \"Turnham Green-Chiswick Park\", \"source\": \"Turnham Green\", \"target\": \"Chiswick Park\", \"weight\": 9076.0}, {\"uid\": \"Turnham Green-Acton Town\", \"source\": \"Turnham Green\", \"target\": \"Acton Town\", \"weight\": 1727.0}, {\"uid\": \"Turnpike Lane-Manor House\", \"source\": \"Turnpike Lane\", \"target\": \"Manor House\", \"weight\": 28700.0}, {\"uid\": \"Turnpike Lane-Wood Green\", \"source\": \"Turnpike Lane\", \"target\": \"Wood Green\", \"weight\": 21355.0}, {\"uid\": \"Upminster-Barking\", \"source\": \"Upminster\", \"target\": \"Barking\", \"weight\": 6666.0}, {\"uid\": \"Upminster-Upminster Bridge\", \"source\": \"Upminster\", \"target\": \"Upminster Bridge\", \"weight\": 4687.0}, {\"uid\": \"Upminster Bridge-Hornchurch\", \"source\": \"Upminster Bridge\", \"target\": \"Hornchurch\", \"weight\": 3862.0}, {\"uid\": \"Upminster Bridge-Upminster\", \"source\": \"Upminster Bridge\", \"target\": \"Upminster\", \"weight\": 4616.0}, {\"uid\": \"Upney-Barking\", \"source\": \"Upney\", \"target\": \"Barking\", \"weight\": 9390.0}, {\"uid\": \"Upney-Becontree\", \"source\": \"Upney\", \"target\": \"Becontree\", \"weight\": 7535.0}, {\"uid\": \"Upton Park-East Ham\", \"source\": \"Upton Park\", \"target\": \"East Ham\", \"weight\": 1174.0}, {\"uid\": \"Upton Park-Plaistow\", \"source\": \"Upton Park\", \"target\": \"Plaistow\", \"weight\": 8236.0}, {\"uid\": \"Uxbridge-Hillingdon\", \"source\": \"Uxbridge\", \"target\": \"Hillingdon\", \"weight\": 6011.0}, {\"uid\": \"Vauxhall-Pimlico\", \"source\": \"Vauxhall\", \"target\": \"Pimlico\", \"weight\": 51326.0}, {\"uid\": \"Vauxhall-Stockwell\", \"source\": \"Vauxhall\", \"target\": \"Stockwell\", \"weight\": 39904.0}, {\"uid\": \"Victoria-Green Park\", \"source\": \"Victoria\", \"target\": \"Green Park\", \"weight\": 117102.0}, {\"uid\": \"Victoria-Pimlico\", \"source\": \"Victoria\", \"target\": \"Pimlico\", \"weight\": 54421.0}, {\"uid\": \"Victoria-Sloane Square\", \"source\": \"Victoria\", \"target\": \"Sloane Square\", \"weight\": 61015.0}, {\"uid\": \"Victoria-St. James's Park\", \"source\": \"Victoria\", \"target\": \"St. James's Park\", \"weight\": 28504.0}, {\"uid\": \"Walthamstow Central-Blackhorse Road\", \"source\": \"Walthamstow Central\", \"target\": \"Blackhorse Road\", \"weight\": 12014.0}, {\"uid\": \"Wanstead-Leytonstone\", \"source\": \"Wanstead\", \"target\": \"Leytonstone\", \"weight\": 14627.0}, {\"uid\": \"Wanstead-Redbridge\", \"source\": \"Wanstead\", \"target\": \"Redbridge\", \"weight\": 11530.0}, {\"uid\": \"Warren Street-Euston\", \"source\": \"Warren Street\", \"target\": \"Euston\", \"weight\": 100983.0}, {\"uid\": \"Warren Street-Goodge Street\", \"source\": \"Warren Street\", \"target\": \"Goodge Street\", \"weight\": 6237.0}, {\"uid\": \"Warren Street-Oxford Circus\", \"source\": \"Warren Street\", \"target\": \"Oxford Circus\", \"weight\": 105845.0}, {\"uid\": \"Warwick Avenue-Maida Vale\", \"source\": \"Warwick Avenue\", \"target\": \"Maida Vale\", \"weight\": 17488.0}, {\"uid\": \"Warwick Avenue-Paddington\", \"source\": \"Warwick Avenue\", \"target\": \"Paddington\", \"weight\": 23549.0}, {\"uid\": \"Waterloo-Westminster\", \"source\": \"Waterloo\", \"target\": \"Westminster\", \"weight\": 18534.0}, {\"uid\": \"Waterloo-Southwark\", \"source\": \"Waterloo\", \"target\": \"Southwark\", \"weight\": 26473.0}, {\"uid\": \"Waterloo-Kennington\", \"source\": \"Waterloo\", \"target\": \"Kennington\", \"weight\": 5691.0}, {\"uid\": \"Waterloo-Embankment\", \"source\": \"Waterloo\", \"target\": \"Embankment\", \"weight\": 145565.0}, {\"uid\": \"Waterloo-Lambeth North\", \"source\": \"Waterloo\", \"target\": \"Lambeth North\", \"weight\": 129855.0}, {\"uid\": \"Watford-Croxley\", \"source\": \"Watford\", \"target\": \"Croxley\", \"weight\": 1338.0}, {\"uid\": \"Wembley Central-North Wembley\", \"source\": \"Wembley Central\", \"target\": \"North Wembley\", \"weight\": 4494.0}, {\"uid\": \"Wembley Central-Stonebridge Park\", \"source\": \"Wembley Central\", \"target\": \"Stonebridge Park\", \"weight\": 8103.0}, {\"uid\": \"Wembley Park-Preston Road\", \"source\": \"Wembley Park\", \"target\": \"Preston Road\", \"weight\": 15138.0}, {\"uid\": \"Wembley Park-Neasden\", \"source\": \"Wembley Park\", \"target\": \"Neasden\", \"weight\": 2889.0}, {\"uid\": \"Wembley Park-Kingsbury\", \"source\": \"Wembley Park\", \"target\": \"Kingsbury\", \"weight\": 9370.0}, {\"uid\": \"Wembley Park-HarrowOnTheHill\", \"source\": \"Wembley Park\", \"target\": \"HarrowOnTheHill\", \"weight\": 2248.0}, {\"uid\": \"Wembley Park-Finchley Road\", \"source\": \"Wembley Park\", \"target\": \"Finchley Road\", \"weight\": 2227.0}, {\"uid\": \"West Acton-Ealing Broadway\", \"source\": \"West Acton\", \"target\": \"Ealing Broadway\", \"weight\": 15925.0}, {\"uid\": \"West Acton-North Acton\", \"source\": \"West Acton\", \"target\": \"North Acton\", \"weight\": 15284.0}, {\"uid\": \"West Brompton-Earl's Court\", \"source\": \"West Brompton\", \"target\": \"Earl's Court\", \"weight\": 29484.0}, {\"uid\": \"West Brompton-Fulham Broadway\", \"source\": \"West Brompton\", \"target\": \"Fulham Broadway\", \"weight\": 26272.0}, {\"uid\": \"West Brompton-Kensington (Olympia)\", \"source\": \"West Brompton\", \"target\": \"Kensington (Olympia)\", \"weight\": 75.0}, {\"uid\": \"West Finchley-Finchley Central\", \"source\": \"West Finchley\", \"target\": \"Finchley Central\", \"weight\": 7307.0}, {\"uid\": \"West Finchley-Woodside Park\", \"source\": \"West Finchley\", \"target\": \"Woodside Park\", \"weight\": 6128.0}, {\"uid\": \"West Ham-Stratford\", \"source\": \"West Ham\", \"target\": \"Stratford\", \"weight\": 28404.0}, {\"uid\": \"West Ham-Plaistow\", \"source\": \"West Ham\", \"target\": \"Plaistow\", \"weight\": 1430.0}, {\"uid\": \"West Ham-Canning Town\", \"source\": \"West Ham\", \"target\": \"Canning Town\", \"weight\": 21824.0}, {\"uid\": \"West Ham-Barking\", \"source\": \"West Ham\", \"target\": \"Barking\", \"weight\": 10167.0}, {\"uid\": \"West Ham-BromleyByBow\", \"source\": \"West Ham\", \"target\": \"BromleyByBow\", \"weight\": 49769.0}, {\"uid\": \"West Hampstead-Kilburn\", \"source\": \"West Hampstead\", \"target\": \"Kilburn\", \"weight\": 10052.0}, {\"uid\": \"West Hampstead-Finchley Road\", \"source\": \"West Hampstead\", \"target\": \"Finchley Road\", \"weight\": 3070.0}, {\"uid\": \"West Harrow-HarrowOnTheHill\", \"source\": \"West Harrow\", \"target\": \"HarrowOnTheHill\", \"weight\": 12717.0}, {\"uid\": \"West Harrow-Rayners Lane\", \"source\": \"West Harrow\", \"target\": \"Rayners Lane\", \"weight\": 12615.0}, {\"uid\": \"West Kensington-Barons Court\", \"source\": \"West Kensington\", \"target\": \"Barons Court\", \"weight\": 1137.0}, {\"uid\": \"West Kensington-Earl's Court\", \"source\": \"West Kensington\", \"target\": \"Earl's Court\", \"weight\": 3089.0}, {\"uid\": \"West Ruislip-Ruislip Gardens\", \"source\": \"West Ruislip\", \"target\": \"Ruislip Gardens\", \"weight\": 13.0}, {\"uid\": \"West Ruislip-South Ruislip\", \"source\": \"West Ruislip\", \"target\": \"South Ruislip\", \"weight\": 1225.0}, {\"uid\": \"Westbourne Park-Royal Oak\", \"source\": \"Westbourne Park\", \"target\": \"Royal Oak\", \"weight\": 11250.0}, {\"uid\": \"Westbourne Park-Ladbroke Grove\", \"source\": \"Westbourne Park\", \"target\": \"Ladbroke Grove\", \"weight\": 14985.0}, {\"uid\": \"Westminster-Embankment\", \"source\": \"Westminster\", \"target\": \"Embankment\", \"weight\": 20482.0}, {\"uid\": \"Westminster-Green Park\", \"source\": \"Westminster\", \"target\": \"Green Park\", \"weight\": 125217.0}, {\"uid\": \"Westminster-St. James's Park\", \"source\": \"Westminster\", \"target\": \"St. James's Park\", \"weight\": 29084.0}, {\"uid\": \"Westminster-Waterloo\", \"source\": \"Westminster\", \"target\": \"Waterloo\", \"weight\": 130860.0}, {\"uid\": \"White City-East Acton\", \"source\": \"White City\", \"target\": \"East Acton\", \"weight\": 5426.0}, {\"uid\": \"White City-Shepherd's Bush (Cen)\", \"source\": \"White City\", \"target\": \"Shepherd's Bush (Cen)\", \"weight\": 8420.0}, {\"uid\": \"Whitechapel-Stratford\", \"source\": \"Whitechapel\", \"target\": \"Stratford\", \"weight\": 64565.0}, {\"uid\": \"Whitechapel-Stepney Green\", \"source\": \"Whitechapel\", \"target\": \"Stepney Green\", \"weight\": 2521.0}, {\"uid\": \"Whitechapel-Aldgate East\", \"source\": \"Whitechapel\", \"target\": \"Aldgate East\", \"weight\": 56414.0}, {\"uid\": \"Willesden Green-Dollis Hill\", \"source\": \"Willesden Green\", \"target\": \"Dollis Hill\", \"weight\": 1770.0}, {\"uid\": \"Willesden Green-Finchley Road\", \"source\": \"Willesden Green\", \"target\": \"Finchley Road\", \"weight\": 11900.0}, {\"uid\": \"Willesden Green-Kilburn\", \"source\": \"Willesden Green\", \"target\": \"Kilburn\", \"weight\": 3376.0}, {\"uid\": \"Willesden Green-Neasden\", \"source\": \"Willesden Green\", \"target\": \"Neasden\", \"weight\": 1511.0}, {\"uid\": \"Willesden Junction-Harlesden\", \"source\": \"Willesden Junction\", \"target\": \"Harlesden\", \"weight\": 8585.0}, {\"uid\": \"Willesden Junction-Kensal Green\", \"source\": \"Willesden Junction\", \"target\": \"Kensal Green\", \"weight\": 11822.0}, {\"uid\": \"Wimbledon-Wimbledon Park\", \"source\": \"Wimbledon\", \"target\": \"Wimbledon Park\", \"weight\": 9939.0}, {\"uid\": \"Wimbledon Park-Wimbledon\", \"source\": \"Wimbledon Park\", \"target\": \"Wimbledon\", \"weight\": 10472.0}, {\"uid\": \"Wimbledon Park-Southfields\", \"source\": \"Wimbledon Park\", \"target\": \"Southfields\", \"weight\": 9001.0}, {\"uid\": \"Wood Green-Bounds Green\", \"source\": \"Wood Green\", \"target\": \"Bounds Green\", \"weight\": 14086.0}, {\"uid\": \"Wood Green-Turnpike Lane\", \"source\": \"Wood Green\", \"target\": \"Turnpike Lane\", \"weight\": 22347.0}, {\"uid\": \"Wood Lane-Latimer Road\", \"source\": \"Wood Lane\", \"target\": \"Latimer Road\", \"weight\": 10079.0}, {\"uid\": \"Wood Lane-Shepherd's Bush Market\", \"source\": \"Wood Lane\", \"target\": \"Shepherd's Bush Market\", \"weight\": 7106.0}, {\"uid\": \"Woodford-Buckhurst Hill\", \"source\": \"Woodford\", \"target\": \"Buckhurst Hill\", \"weight\": 8138.0}, {\"uid\": \"Woodford-Roding Valley\", \"source\": \"Woodford\", \"target\": \"Roding Valley\", \"weight\": 2240.0}, {\"uid\": \"Woodford-South Woodford\", \"source\": \"Woodford\", \"target\": \"South Woodford\", \"weight\": 14506.0}, {\"uid\": \"Woodside Park-Totteridge & Whetstone\", \"source\": \"Woodside Park\", \"target\": \"Totteridge & Whetstone\", \"weight\": 3956.0}, {\"uid\": \"Woodside Park-West Finchley\", \"source\": \"Woodside Park\", \"target\": \"West Finchley\", \"weight\": 6218.0}], \"nodes\": [{\"uid\": \"Acton Town\", \"label\": \"Acton Town\"}, {\"uid\": \"Aldgate\", \"label\": \"Aldgate\"}, {\"uid\": \"Aldgate East\", \"label\": \"Aldgate East\"}, {\"uid\": \"Alperton\", \"label\": \"Alperton\"}, {\"uid\": \"Amersham\", \"label\": \"Amersham\"}, {\"uid\": \"Angel\", \"label\": \"Angel\"}, {\"uid\": \"Archway\", \"label\": \"Archway\"}, {\"uid\": \"Arnos Grove\", \"label\": \"Arnos Grove\"}, {\"uid\": \"Arsenal\", \"label\": \"Arsenal\"}, {\"uid\": \"Baker Street\", \"label\": \"Baker Street\"}, {\"uid\": \"Balham\", \"label\": \"Balham\"}, {\"uid\": \"Bank / Monument\", \"label\": \"Bank / Monument\"}, {\"uid\": \"Barbican\", \"label\": \"Barbican\"}, {\"uid\": \"Barking\", \"label\": \"Barking\"}, {\"uid\": \"Barkingside\", \"label\": \"Barkingside\"}, {\"uid\": \"Barons Court\", \"label\": \"Barons Court\"}, {\"uid\": \"Bayswater\", \"label\": \"Bayswater\"}, {\"uid\": \"Becontree\", \"label\": \"Becontree\"}, {\"uid\": \"Belsize Park\", \"label\": \"Belsize Park\"}, {\"uid\": \"Bermondsey\", \"label\": \"Bermondsey\"}, {\"uid\": \"Bethnal Green\", \"label\": \"Bethnal Green\"}, {\"uid\": \"Blackfriars\", \"label\": \"Blackfriars\"}, {\"uid\": \"Blackhorse Road\", \"label\": \"Blackhorse Road\"}, {\"uid\": \"Bond Street\", \"label\": \"Bond Street\"}, {\"uid\": \"Borough\", \"label\": \"Borough\"}, {\"uid\": \"Boston Manor\", \"label\": \"Boston Manor\"}, {\"uid\": \"Bounds Green\", \"label\": \"Bounds Green\"}, {\"uid\": \"Bow Road\", \"label\": \"Bow Road\"}, {\"uid\": \"Brent Cross\", \"label\": \"Brent Cross\"}, {\"uid\": \"Brixton\", \"label\": \"Brixton\"}, {\"uid\": \"BromleyByBow\", \"label\": \"BromleyByBow\"}, {\"uid\": \"Buckhurst Hill\", \"label\": \"Buckhurst Hill\"}, {\"uid\": \"Burnt Oak\", \"label\": \"Burnt Oak\"}, {\"uid\": \"Caledonian Road\", \"label\": \"Caledonian Road\"}, {\"uid\": \"Camden Town\", \"label\": \"Camden Town\"}, {\"uid\": \"Canada Water\", \"label\": \"Canada Water\"}, {\"uid\": \"Canary Wharf\", \"label\": \"Canary Wharf\"}, {\"uid\": \"Canning Town\", \"label\": \"Canning Town\"}, {\"uid\": \"Cannon Street\", \"label\": \"Cannon Street\"}, {\"uid\": \"Canons Park\", \"label\": \"Canons Park\"}, {\"uid\": \"Chalfont & Latimer\", \"label\": \"Chalfont & Latimer\"}, {\"uid\": \"Chalk Farm\", \"label\": \"Chalk Farm\"}, {\"uid\": \"Chancery Lane\", \"label\": \"Chancery Lane\"}, {\"uid\": \"Charing Cross\", \"label\": \"Charing Cross\"}, {\"uid\": \"Chesham\", \"label\": \"Chesham\"}, {\"uid\": \"Chigwell\", \"label\": \"Chigwell\"}, {\"uid\": \"Chiswick Park\", \"label\": \"Chiswick Park\"}, {\"uid\": \"Chorleywood\", \"label\": \"Chorleywood\"}, {\"uid\": \"Clapham Common\", \"label\": \"Clapham Common\"}, {\"uid\": \"Clapham North\", \"label\": \"Clapham North\"}, {\"uid\": \"Clapham South\", \"label\": \"Clapham South\"}, {\"uid\": \"Cockfosters\", \"label\": \"Cockfosters\"}, {\"uid\": \"Colindale\", \"label\": \"Colindale\"}, {\"uid\": \"Colliers Wood\", \"label\": \"Colliers Wood\"}, {\"uid\": \"Covent Garden\", \"label\": \"Covent Garden\"}, {\"uid\": \"Croxley\", \"label\": \"Croxley\"}, {\"uid\": \"Dagenham East\", \"label\": \"Dagenham East\"}, {\"uid\": \"Dagenham Heathway\", \"label\": \"Dagenham Heathway\"}, {\"uid\": \"Debden\", \"label\": \"Debden\"}, {\"uid\": \"Dollis Hill\", \"label\": \"Dollis Hill\"}, {\"uid\": \"Ealing Broadway\", \"label\": \"Ealing Broadway\"}, {\"uid\": \"Ealing Common\", \"label\": \"Ealing Common\"}, {\"uid\": \"Earl's Court\", \"label\": \"Earl's Court\"}, {\"uid\": \"East Acton\", \"label\": \"East Acton\"}, {\"uid\": \"East Finchley\", \"label\": \"East Finchley\"}, {\"uid\": \"East Ham\", \"label\": \"East Ham\"}, {\"uid\": \"East Putney\", \"label\": \"East Putney\"}, {\"uid\": \"Eastcote\", \"label\": \"Eastcote\"}, {\"uid\": \"Edgware\", \"label\": \"Edgware\"}, {\"uid\": \"Edgware Road (Bak)\", \"label\": \"Edgware Road (Bak)\"}, {\"uid\": \"Edgware Road (Cir)\", \"label\": \"Edgware Road (Cir)\"}, {\"uid\": \"Elephant & Castle\", \"label\": \"Elephant & Castle\"}, {\"uid\": \"Elm Park\", \"label\": \"Elm Park\"}, {\"uid\": \"Embankment\", \"label\": \"Embankment\"}, {\"uid\": \"Epping\", \"label\": \"Epping\"}, {\"uid\": \"Euston\", \"label\": \"Euston\"}, {\"uid\": \"Euston Square\", \"label\": \"Euston Square\"}, {\"uid\": \"Fairlop\", \"label\": \"Fairlop\"}, {\"uid\": \"Farringdon\", \"label\": \"Farringdon\"}, {\"uid\": \"Finchley Central\", \"label\": \"Finchley Central\"}, {\"uid\": \"Finchley Road\", \"label\": \"Finchley Road\"}, {\"uid\": \"Finsbury Park\", \"label\": \"Finsbury Park\"}, {\"uid\": \"Fulham Broadway\", \"label\": \"Fulham Broadway\"}, {\"uid\": \"Gants Hill\", \"label\": \"Gants Hill\"}, {\"uid\": \"Gloucester Road\", \"label\": \"Gloucester Road\"}, {\"uid\": \"Golders Green\", \"label\": \"Golders Green\"}, {\"uid\": \"Goldhawk Road\", \"label\": \"Goldhawk Road\"}, {\"uid\": \"Goodge Street\", \"label\": \"Goodge Street\"}, {\"uid\": \"Grange Hill\", \"label\": \"Grange Hill\"}, {\"uid\": \"Great Portland Street\", \"label\": \"Great Portland Street\"}, {\"uid\": \"Green Park\", \"label\": \"Green Park\"}, {\"uid\": \"Greenford\", \"label\": \"Greenford\"}, {\"uid\": \"Gunnersbury\", \"label\": \"Gunnersbury\"}, {\"uid\": \"Hainault\", \"label\": \"Hainault\"}, {\"uid\": \"Hammersmith (Dis)\", \"label\": \"Hammersmith (Dis)\"}, {\"uid\": \"Hammersmith (H&C)\", \"label\": \"Hammersmith (H&C)\"}, {\"uid\": \"Hampstead\", \"label\": \"Hampstead\"}, {\"uid\": \"Hanger Lane\", \"label\": \"Hanger Lane\"}, {\"uid\": \"Harlesden\", \"label\": \"Harlesden\"}, {\"uid\": \"Harrow & Wealdstone\", \"label\": \"Harrow & Wealdstone\"}, {\"uid\": \"HarrowOnTheHill\", \"label\": \"HarrowOnTheHill\"}, {\"uid\": \"Hatton Cross\", \"label\": \"Hatton Cross\"}, {\"uid\": \"Heathrow Terminal 4\", \"label\": \"Heathrow Terminal 4\"}, {\"uid\": \"Heathrow Terminal 5\", \"label\": \"Heathrow Terminal 5\"}, {\"uid\": \"Heathrow Terminals 123\", \"label\": \"Heathrow Terminals 123\"}, {\"uid\": \"Hendon Central\", \"label\": \"Hendon Central\"}, {\"uid\": \"High Barnet\", \"label\": \"High Barnet\"}, {\"uid\": \"High Street Kensington\", \"label\": \"High Street Kensington\"}, {\"uid\": \"Highbury & Islington\", \"label\": \"Highbury & Islington\"}, {\"uid\": \"Highgate\", \"label\": \"Highgate\"}, {\"uid\": \"Hillingdon\", \"label\": \"Hillingdon\"}, {\"uid\": \"Holborn\", \"label\": \"Holborn\"}, {\"uid\": \"Holland Park\", \"label\": \"Holland Park\"}, {\"uid\": \"Holloway Road\", \"label\": \"Holloway Road\"}, {\"uid\": \"Hornchurch\", \"label\": \"Hornchurch\"}, {\"uid\": \"Hounslow Central\", \"label\": \"Hounslow Central\"}, {\"uid\": \"Hounslow East\", \"label\": \"Hounslow East\"}, {\"uid\": \"Hounslow West\", \"label\": \"Hounslow West\"}, {\"uid\": \"Hyde Park Corner\", \"label\": \"Hyde Park Corner\"}, {\"uid\": \"Ickenham\", \"label\": \"Ickenham\"}, {\"uid\": \"Kennington\", \"label\": \"Kennington\"}, {\"uid\": \"Kensal Green\", \"label\": \"Kensal Green\"}, {\"uid\": \"Kensington (Olympia)\", \"label\": \"Kensington (Olympia)\"}, {\"uid\": \"Kentish Town\", \"label\": \"Kentish Town\"}, {\"uid\": \"Kenton\", \"label\": \"Kenton\"}, {\"uid\": \"Kew Gardens\", \"label\": \"Kew Gardens\"}, {\"uid\": \"Kilburn\", \"label\": \"Kilburn\"}, {\"uid\": \"Kilburn Park\", \"label\": \"Kilburn Park\"}, {\"uid\": \"King's Cross St. Pancras\", \"label\": \"King's Cross St. Pancras\"}, {\"uid\": \"Kingsbury\", \"label\": \"Kingsbury\"}, {\"uid\": \"Knightsbridge\", \"label\": \"Knightsbridge\"}, {\"uid\": \"Ladbroke Grove\", \"label\": \"Ladbroke Grove\"}, {\"uid\": \"Lambeth North\", \"label\": \"Lambeth North\"}, {\"uid\": \"Lancaster Gate\", \"label\": \"Lancaster Gate\"}, {\"uid\": \"Latimer Road\", \"label\": \"Latimer Road\"}, {\"uid\": \"Leicester Square\", \"label\": \"Leicester Square\"}, {\"uid\": \"Leyton\", \"label\": \"Leyton\"}, {\"uid\": \"Leytonstone\", \"label\": \"Leytonstone\"}, {\"uid\": \"Liverpool Street\", \"label\": \"Liverpool Street\"}, {\"uid\": \"London Bridge\", \"label\": \"London Bridge\"}, {\"uid\": \"Loughton\", \"label\": \"Loughton\"}, {\"uid\": \"Maida Vale\", \"label\": \"Maida Vale\"}, {\"uid\": \"Manor House\", \"label\": \"Manor House\"}, {\"uid\": \"Mansion House\", \"label\": \"Mansion House\"}, {\"uid\": \"Marble Arch\", \"label\": \"Marble Arch\"}, {\"uid\": \"Marylebone\", \"label\": \"Marylebone\"}, {\"uid\": \"Mile End\", \"label\": \"Mile End\"}, {\"uid\": \"Mill Hill East\", \"label\": \"Mill Hill East\"}, {\"uid\": \"Moor Park\", \"label\": \"Moor Park\"}, {\"uid\": \"Moorgate\", \"label\": \"Moorgate\"}, {\"uid\": \"Morden\", \"label\": \"Morden\"}, {\"uid\": \"Mornington Crescent\", \"label\": \"Mornington Crescent\"}, {\"uid\": \"Neasden\", \"label\": \"Neasden\"}, {\"uid\": \"Newbury Park\", \"label\": \"Newbury Park\"}, {\"uid\": \"North Acton\", \"label\": \"North Acton\"}, {\"uid\": \"North Ealing\", \"label\": \"North Ealing\"}, {\"uid\": \"North Greenwich\", \"label\": \"North Greenwich\"}, {\"uid\": \"North Harrow\", \"label\": \"North Harrow\"}, {\"uid\": \"North Wembley\", \"label\": \"North Wembley\"}, {\"uid\": \"Northfields\", \"label\": \"Northfields\"}, {\"uid\": \"Northolt\", \"label\": \"Northolt\"}, {\"uid\": \"Northwick Park\", \"label\": \"Northwick Park\"}, {\"uid\": \"Northwood\", \"label\": \"Northwood\"}, {\"uid\": \"Northwood Hills\", \"label\": \"Northwood Hills\"}, {\"uid\": \"Notting Hill Gate\", \"label\": \"Notting Hill Gate\"}, {\"uid\": \"Oakwood\", \"label\": \"Oakwood\"}, {\"uid\": \"Old Street\", \"label\": \"Old Street\"}, {\"uid\": \"Osterley\", \"label\": \"Osterley\"}, {\"uid\": \"Oval\", \"label\": \"Oval\"}, {\"uid\": \"Oxford Circus\", \"label\": \"Oxford Circus\"}, {\"uid\": \"Paddington\", \"label\": \"Paddington\"}, {\"uid\": \"Park Royal\", \"label\": \"Park Royal\"}, {\"uid\": \"Parsons Green\", \"label\": \"Parsons Green\"}, {\"uid\": \"Perivale\", \"label\": \"Perivale\"}, {\"uid\": \"Piccadilly Circus\", \"label\": \"Piccadilly Circus\"}, {\"uid\": \"Pimlico\", \"label\": \"Pimlico\"}, {\"uid\": \"Pinner\", \"label\": \"Pinner\"}, {\"uid\": \"Plaistow\", \"label\": \"Plaistow\"}, {\"uid\": \"Preston Road\", \"label\": \"Preston Road\"}, {\"uid\": \"Putney Bridge\", \"label\": \"Putney Bridge\"}, {\"uid\": \"Queen's Park\", \"label\": \"Queen's Park\"}, {\"uid\": \"Queensbury\", \"label\": \"Queensbury\"}, {\"uid\": \"Queensway\", \"label\": \"Queensway\"}, {\"uid\": \"Ravenscourt Park\", \"label\": \"Ravenscourt Park\"}, {\"uid\": \"Rayners Lane\", \"label\": \"Rayners Lane\"}, {\"uid\": \"Redbridge\", \"label\": \"Redbridge\"}, {\"uid\": \"Regent's Park\", \"label\": \"Regent's Park\"}, {\"uid\": \"Richmond\", \"label\": \"Richmond\"}, {\"uid\": \"Rickmansworth\", \"label\": \"Rickmansworth\"}, {\"uid\": \"Roding Valley\", \"label\": \"Roding Valley\"}, {\"uid\": \"Royal Oak\", \"label\": \"Royal Oak\"}, {\"uid\": \"Ruislip\", \"label\": \"Ruislip\"}, {\"uid\": \"Ruislip Gardens\", \"label\": \"Ruislip Gardens\"}, {\"uid\": \"Ruislip Manor\", \"label\": \"Ruislip Manor\"}, {\"uid\": \"Russell Square\", \"label\": \"Russell Square\"}, {\"uid\": \"Seven Sisters\", \"label\": \"Seven Sisters\"}, {\"uid\": \"Shepherd's Bush (Cen)\", \"label\": \"Shepherd's Bush (Cen)\"}, {\"uid\": \"Shepherd's Bush Market\", \"label\": \"Shepherd's Bush Market\"}, {\"uid\": \"Sloane Square\", \"label\": \"Sloane Square\"}, {\"uid\": \"Snaresbrook\", \"label\": \"Snaresbrook\"}, {\"uid\": \"South Ealing\", \"label\": \"South Ealing\"}, {\"uid\": \"South Harrow\", \"label\": \"South Harrow\"}, {\"uid\": \"South Kensington\", \"label\": \"South Kensington\"}, {\"uid\": \"South Kenton\", \"label\": \"South Kenton\"}, {\"uid\": \"South Ruislip\", \"label\": \"South Ruislip\"}, {\"uid\": \"South Wimbledon\", \"label\": \"South Wimbledon\"}, {\"uid\": \"South Woodford\", \"label\": \"South Woodford\"}, {\"uid\": \"Southfields\", \"label\": \"Southfields\"}, {\"uid\": \"Southgate\", \"label\": \"Southgate\"}, {\"uid\": \"Southwark\", \"label\": \"Southwark\"}, {\"uid\": \"St. James's Park\", \"label\": \"St. James's Park\"}, {\"uid\": \"St. John's Wood\", \"label\": \"St. John's Wood\"}, {\"uid\": \"St. Paul's\", \"label\": \"St. Paul's\"}, {\"uid\": \"Stamford Brook\", \"label\": \"Stamford Brook\"}, {\"uid\": \"Stanmore\", \"label\": \"Stanmore\"}, {\"uid\": \"Stepney Green\", \"label\": \"Stepney Green\"}, {\"uid\": \"Stockwell\", \"label\": \"Stockwell\"}, {\"uid\": \"Stonebridge Park\", \"label\": \"Stonebridge Park\"}, {\"uid\": \"Stratford\", \"label\": \"Stratford\"}, {\"uid\": \"Sudbury Hill\", \"label\": \"Sudbury Hill\"}, {\"uid\": \"Sudbury Town\", \"label\": \"Sudbury Town\"}, {\"uid\": \"Swiss Cottage\", \"label\": \"Swiss Cottage\"}, {\"uid\": \"Temple\", \"label\": \"Temple\"}, {\"uid\": \"Theydon Bois\", \"label\": \"Theydon Bois\"}, {\"uid\": \"Tooting Bec\", \"label\": \"Tooting Bec\"}, {\"uid\": \"Tooting Broadway\", \"label\": \"Tooting Broadway\"}, {\"uid\": \"Tottenham Court Road\", \"label\": \"Tottenham Court Road\"}, {\"uid\": \"Tottenham Hale\", \"label\": \"Tottenham Hale\"}, {\"uid\": \"Totteridge & Whetstone\", \"label\": \"Totteridge & Whetstone\"}, {\"uid\": \"Tower Hill\", \"label\": \"Tower Hill\"}, {\"uid\": \"Tufnell Park\", \"label\": \"Tufnell Park\"}, {\"uid\": \"Turnham Green\", \"label\": \"Turnham Green\"}, {\"uid\": \"Turnpike Lane\", \"label\": \"Turnpike Lane\"}, {\"uid\": \"Upminster\", \"label\": \"Upminster\"}, {\"uid\": \"Upminster Bridge\", \"label\": \"Upminster Bridge\"}, {\"uid\": \"Upney\", \"label\": \"Upney\"}, {\"uid\": \"Upton Park\", \"label\": \"Upton Park\"}, {\"uid\": \"Uxbridge\", \"label\": \"Uxbridge\"}, {\"uid\": \"Vauxhall\", \"label\": \"Vauxhall\"}, {\"uid\": \"Victoria\", \"label\": \"Victoria\"}, {\"uid\": \"Walthamstow Central\", \"label\": \"Walthamstow Central\"}, {\"uid\": \"Wanstead\", \"label\": \"Wanstead\"}, {\"uid\": \"Warren Street\", \"label\": \"Warren Street\"}, {\"uid\": \"Warwick Avenue\", \"label\": \"Warwick Avenue\"}, {\"uid\": \"Waterloo\", \"label\": \"Waterloo\"}, {\"uid\": \"Watford\", \"label\": \"Watford\"}, {\"uid\": \"Wembley Central\", \"label\": \"Wembley Central\"}, {\"uid\": \"Wembley Park\", \"label\": \"Wembley Park\"}, {\"uid\": \"West Acton\", \"label\": \"West Acton\"}, {\"uid\": \"West Brompton\", \"label\": \"West Brompton\"}, {\"uid\": \"West Finchley\", \"label\": \"West Finchley\"}, {\"uid\": \"West Ham\", \"label\": \"West Ham\"}, {\"uid\": \"West Hampstead\", \"label\": \"West Hampstead\"}, {\"uid\": \"West Harrow\", \"label\": \"West Harrow\"}, {\"uid\": \"West Kensington\", \"label\": \"West Kensington\"}, {\"uid\": \"West Ruislip\", \"label\": \"West Ruislip\"}, {\"uid\": \"Westbourne Park\", \"label\": \"Westbourne Park\"}, {\"uid\": \"Westminster\", \"label\": \"Westminster\"}, {\"uid\": \"White City\", \"label\": \"White City\"}, {\"uid\": \"Whitechapel\", \"label\": \"Whitechapel\"}, {\"uid\": \"Willesden Green\", \"label\": \"Willesden Green\"}, {\"uid\": \"Willesden Junction\", \"label\": \"Willesden Junction\"}, {\"uid\": \"Wimbledon\", \"label\": \"Wimbledon\"}, {\"uid\": \"Wimbledon Park\", \"label\": \"Wimbledon Park\"}, {\"uid\": \"Wood Green\", \"label\": \"Wood Green\"}, {\"uid\": \"Wood Lane\", \"label\": \"Wood Lane\"}, {\"uid\": \"Woodford\", \"label\": \"Woodford\"}, {\"uid\": \"Woodside Park\", \"label\": \"Woodside Park\"}]}\n",
       "const config = {\"node_label\": [\"Acton Town\", \"Aldgate\", \"Aldgate East\", \"Alperton\", \"Amersham\", \"Angel\", \"Archway\", \"Arnos Grove\", \"Arsenal\", \"Baker Street\", \"Balham\", \"Bank / Monument\", \"Barbican\", \"Barking\", \"Barkingside\", \"Barons Court\", \"Bayswater\", \"Becontree\", \"Belsize Park\", \"Bermondsey\", \"Bethnal Green\", \"Blackfriars\", \"Blackhorse Road\", \"Bond Street\", \"Borough\", \"Boston Manor\", \"Bounds Green\", \"Bow Road\", \"Brent Cross\", \"Brixton\", \"BromleyByBow\", \"Buckhurst Hill\", \"Burnt Oak\", \"Caledonian Road\", \"Camden Town\", \"Canada Water\", \"Canary Wharf\", \"Canning Town\", \"Cannon Street\", \"Canons Park\", \"Chalfont & Latimer\", \"Chalk Farm\", \"Chancery Lane\", \"Charing Cross\", \"Chesham\", \"Chigwell\", \"Chiswick Park\", \"Chorleywood\", \"Clapham Common\", \"Clapham North\", \"Clapham South\", \"Cockfosters\", \"Colindale\", \"Colliers Wood\", \"Covent Garden\", \"Croxley\", \"Dagenham East\", \"Dagenham Heathway\", \"Debden\", \"Dollis Hill\", \"Ealing Broadway\", \"Ealing Common\", \"Earl's Court\", \"East Acton\", \"East Finchley\", \"East Ham\", \"East Putney\", \"Eastcote\", \"Edgware\", \"Edgware Road (Bak)\", \"Edgware Road (Cir)\", \"Elephant & Castle\", \"Elm Park\", \"Embankment\", \"Epping\", \"Euston\", \"Euston Square\", \"Fairlop\", \"Farringdon\", \"Finchley Central\", \"Finchley Road\", \"Finsbury Park\", \"Fulham Broadway\", \"Gants Hill\", \"Gloucester Road\", \"Golders Green\", \"Goldhawk Road\", \"Goodge Street\", \"Grange Hill\", \"Great Portland Street\", \"Green Park\", \"Greenford\", \"Gunnersbury\", \"Hainault\", \"Hammersmith (Dis)\", \"Hammersmith (H&C)\", \"Hampstead\", \"Hanger Lane\", \"Harlesden\", \"Harrow & Wealdstone\", \"HarrowOnTheHill\", \"Hatton Cross\", \"Heathrow Terminal 4\", \"Heathrow Terminal 5\", \"Heathrow Terminals 123\", \"Hendon Central\", \"High Barnet\", \"High Street Kensington\", \"Highbury & Islington\", \"Highgate\", \"Hillingdon\", \"Holborn\", \"Holland Park\", \"Holloway Road\", \"Hornchurch\", \"Hounslow Central\", \"Hounslow East\", \"Hounslow West\", \"Hyde Park Corner\", \"Ickenham\", \"Kennington\", \"Kensal Green\", \"Kensington (Olympia)\", \"Kentish Town\", \"Kenton\", \"Kew Gardens\", \"Kilburn\", \"Kilburn Park\", \"King's Cross St. Pancras\", \"Kingsbury\", \"Knightsbridge\", \"Ladbroke Grove\", \"Lambeth North\", \"Lancaster Gate\", \"Latimer Road\", \"Leicester Square\", \"Leyton\", \"Leytonstone\", \"Liverpool Street\", \"London Bridge\", \"Loughton\", \"Maida Vale\", \"Manor House\", \"Mansion House\", \"Marble Arch\", \"Marylebone\", \"Mile End\", \"Mill Hill East\", \"Moor Park\", \"Moorgate\", \"Morden\", \"Mornington Crescent\", \"Neasden\", \"Newbury Park\", \"North Acton\", \"North Ealing\", \"North Greenwich\", \"North Harrow\", \"North Wembley\", \"Northfields\", \"Northolt\", \"Northwick Park\", \"Northwood\", \"Northwood Hills\", \"Notting Hill Gate\", \"Oakwood\", \"Old Street\", \"Osterley\", \"Oval\", \"Oxford Circus\", \"Paddington\", \"Park Royal\", \"Parsons Green\", \"Perivale\", \"Piccadilly Circus\", \"Pimlico\", \"Pinner\", \"Plaistow\", \"Preston Road\", \"Putney Bridge\", \"Queen's Park\", \"Queensbury\", \"Queensway\", \"Ravenscourt Park\", \"Rayners Lane\", \"Redbridge\", \"Regent's Park\", \"Richmond\", \"Rickmansworth\", \"Roding Valley\", \"Royal Oak\", \"Ruislip\", \"Ruislip Gardens\", \"Ruislip Manor\", \"Russell Square\", \"Seven Sisters\", \"Shepherd's Bush (Cen)\", \"Shepherd's Bush Market\", \"Sloane Square\", \"Snaresbrook\", \"South Ealing\", \"South Harrow\", \"South Kensington\", \"South Kenton\", \"South Ruislip\", \"South Wimbledon\", \"South Woodford\", \"Southfields\", \"Southgate\", \"Southwark\", \"St. James's Park\", \"St. John's Wood\", \"St. Paul's\", \"Stamford Brook\", \"Stanmore\", \"Stepney Green\", \"Stockwell\", \"Stonebridge Park\", \"Stratford\", \"Sudbury Hill\", \"Sudbury Town\", \"Swiss Cottage\", \"Temple\", \"Theydon Bois\", \"Tooting Bec\", \"Tooting Broadway\", \"Tottenham Court Road\", \"Tottenham Hale\", \"Totteridge & Whetstone\", \"Tower Hill\", \"Tufnell Park\", \"Turnham Green\", \"Turnpike Lane\", \"Upminster\", \"Upminster Bridge\", \"Upney\", \"Upton Park\", \"Uxbridge\", \"Vauxhall\", \"Victoria\", \"Walthamstow Central\", \"Wanstead\", \"Warren Street\", \"Warwick Avenue\", \"Waterloo\", \"Watford\", \"Wembley Central\", \"Wembley Park\", \"West Acton\", \"West Brompton\", \"West Finchley\", \"West Ham\", \"West Hampstead\", \"West Harrow\", \"West Kensington\", \"West Ruislip\", \"Westbourne Park\", \"Westminster\", \"White City\", \"Whitechapel\", \"Willesden Green\", \"Willesden Junction\", \"Wimbledon\", \"Wimbledon Park\", \"Wood Green\", \"Wood Lane\", \"Woodford\", \"Woodside Park\"], \"directed\": true, \"curved\": true, \"selector\": \"#xe05b2d8866834ff7bb82ba2e2fded367\"}\n",
       "console.log(\"Static Network Template\");\n",
       "/* Resources\n",
       "   https://bl.ocks.org/mapio/53fed7d84cd1812d6a6639ed7aa83868\n",
       "   https://codepen.io/smlo/pen/JdMOej\n",
       "*/\n",
       "\n",
       "// variables from the config file\n",
       "const selector = config.selector;\n",
       "const width = config.width || 800;\n",
       "const height = config.height || 600;\n",
       "const charge_distance = config.charge_distance || 400;\n",
       "const charge_force = config.charge_force || -3000;\n",
       "const curved = config.curved || false;\n",
       "const directed = config.directed || false;\n",
       "// const weight = false;\n",
       "\n",
       "/* Create a svg element to display the network */\n",
       "var svg = d3.select(selector)\n",
       "    .append('svg')\n",
       "    .attr('width', width)\n",
       "    .attr('height', height)\n",
       "\n",
       "// add container to store the elements\n",
       "var container = svg.append(\"g\");\n",
       "\n",
       "/*Add zoom function to the container */\n",
       "svg.call(\n",
       "    d3.zoom()\n",
       "        .scaleExtent([.1, 4])\n",
       "        .on(\"zoom\", function() { container.attr(\"transform\", d3.event.transform); })\n",
       ");\n",
       "\n",
       "\n",
       "/*Load nodes and links from the data */\n",
       "var nodes = data.nodes\n",
       "var links = data.edges\n",
       "\n",
       "/*Create arrow head with same color as the edge */\n",
       "function marker (color) {\n",
       "       var reference;\n",
       "       svg.append(\"svg:defs\").selectAll(\"marker\")\n",
       "          .data([reference])\n",
       "          .enter().append(\"svg:marker\")\n",
       "          .attr(\"id\", \"arrow\"+color)\n",
       "          .attr(\"viewBox\", \"0 -5 10 10\")\n",
       "          .attr(\"refX\", 10)\n",
       "          .attr(\"refY\", -0)\n",
       "          .attr(\"markerWidth\", 6)\n",
       "          .attr(\"markerHeight\", 6)\n",
       "          .attr(\"orient\", \"auto\")\n",
       "          .append(\"svg:path\")\n",
       "          .attr('class','.arrow')\n",
       "          .attr(\"d\", \"M0,-5L10,0L0,5\")\n",
       "          .style('opacity',1)\n",
       "          .style(\"fill\", color);\n",
       "       return \"url(#\" + \"arrow\"+color + \")\";\n",
       "     };\n",
       "\n",
       "/*Link creation template */\n",
       "var link = container.append(\"g\").attr(\"class\", \"links\")\n",
       "    .selectAll(\".link\")\n",
       "    .data(links)\n",
       "    .enter()\n",
       "    .append(\"path\")\n",
       "    .attr(\"class\", \"link\")\n",
       "    .style(\"stroke\", function(d) { return d.color; })\n",
       "    .style(\"stroke-opacity\", function(d) { return d.opacity; })\n",
       "    .style(\"stroke-width\", function(d){  return d.size })\n",
       "    .style(\"fill\",\"none\")\n",
       "    .attr(\"marker-end\", function (d) {if(directed){return marker(d.color)}else{return null}; });\n",
       "\n",
       "    //.attr(\"marker-end\", function (d) { return marker(d.color); });\n",
       "    //.attr(\"marker-end\", \"url(#arrow)\");\n",
       "\n",
       "/*Node creation template */\n",
       "var node = container.append(\"g\").attr(\"class\", \"nodes\")\n",
       "    .selectAll(\"circle.node\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"circle\")\n",
       "    .attr(\"class\", \"node\")\n",
       "    .attr(\"x\", function(d) { return d.x; })\n",
       "    .attr(\"y\", function(d) { return d.y; })\n",
       "    .style(\"r\", function(d){  return d.size+\"px\"; })\n",
       "    .style(\"fill\", function(d) { return d.color; })\n",
       "    .style(\"opacity\", function(d) { return d.opacity; });\n",
       "\n",
       "/*Label creation template */\n",
       "var text = container.append(\"g\").attr(\"class\",\"labels\")\n",
       "    .selectAll(\"g\")\n",
       "    .data(nodes)\n",
       "    .enter().append(\"g\")\n",
       "\n",
       "text.append(\"text\")\n",
       "    .attr(\"class\", \"label-text\")\n",
       "    .attr(\"x\", function(d) {\n",
       "        var r = (d.size === undefined) ? 15 : d.size;\n",
       "        return 5 + r; })\n",
       "    .attr(\"dy\", \".31em\")\n",
       "    .text(function(d) { return d.label; });\n",
       "\n",
       "/*Scale weight for d3js */\n",
       "var weightScale = d3.scaleLinear()\n",
       "    .domain(d3.extent(links, function (d) { return d.weight }))\n",
       "    .range([.1, 1]);\n",
       "\n",
       "/*Simulation of the forces*/\n",
       "var simulation = d3.forceSimulation(nodes)\n",
       "    .force(\"links\", d3.forceLink(links)\n",
       "           .id(function(d) {return d.uid; })\n",
       "           .distance(50)\n",
       "           .strength(function(d){return weightScale(d.weight);})\n",
       "          )\n",
       "    .force(\"charge\", d3.forceManyBody()\n",
       "           .strength(charge_force)\n",
       "           .distanceMax(charge_distance)\n",
       "          )\n",
       "    .force(\"center\", d3.forceCenter(width / 2, height / 2))\n",
       "    .on(\"tick\", ticked);\n",
       "\n",
       "/*Update of the node and edge objects*/\n",
       "function ticked() {\n",
       "    node.call(updateNode);\n",
       "    link.call(updateLink);\n",
       "    text.call(updateText);\n",
       "};\n",
       "\n",
       "/*Update link positions */\n",
       "function updateLink(link) {\n",
       "    // link\n",
       "    //     .attr(\"x1\", function(d) { return d.source.x; })\n",
       "    //     .attr(\"y1\", function(d) { return d.source.y; })\n",
       "    //     .attr(\"x2\", function(d) { return d.target.x; })\n",
       "    //     .attr(\"y2\", function(d) { return d.target.y; });\n",
       "\n",
       "\n",
       "    link.attr(\"d\", function(d) {\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        return \"M\" +\n",
       "            d.source.x + \",\" +\n",
       "            d.source.y + \"A\" +\n",
       "            dr + \",\" + dr + \" 0 0,1 \" +\n",
       "            d.target.x + \",\" +\n",
       "            d.target.y;\n",
       "    });\n",
       "\n",
       "    // recalculate and back off the distance\n",
       "    link.attr(\"d\", function (d, i) {\n",
       "        var pl = this.getTotalLength();\n",
       "        var r = (d.target.size === undefined) ? 15 : d.target.size;\n",
       "        var m = this.getPointAtLength(pl - r);\n",
       "        var dx = d.target.x - d.source.x,\n",
       "            dy = d.target.y - d.source.y,\n",
       "            dr = Math.sqrt(dx * dx + dy * dy);\n",
       "        if(!curved)dr=0;\n",
       "        var result = \"M\" + d.source.x + \",\" + d.source.y + \"A\" + dr + \",\" + dr + \" 0 0,1 \" + m.x + \",\" + m.y;\n",
       "        return result;\n",
       "    });\n",
       "};\n",
       "\n",
       "\n",
       "/*Update node positions */\n",
       "function updateNode(node) {\n",
       "    node.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "    // node\n",
       "    //     .attr(\"cx\", function(d) { return d.x; })\n",
       "    //     .attr(\"cy\", function(d) { return d.y; });\n",
       "};\n",
       "\n",
       "/*Update text positions */\n",
       "function updateText(text) {\n",
       "    text.attr(\"transform\", function(d) {\n",
       "        return \"translate(\" + d.x + \",\" + d.y + \")\";\n",
       "    });\n",
       "};\n",
       "\n",
       "/*Add drag functionality to the node objects*/\n",
       "node.call(\n",
       "    d3.drag()\n",
       "        .on(\"start\", dragstarted)\n",
       "        .on(\"drag\", dragged)\n",
       "        .on(\"end\", dragended)\n",
       ");\n",
       "\n",
       "function dragstarted(d) {\n",
       "    d3.event.sourceEvent.stopPropagation();\n",
       "    if (!d3.event.active) simulation.alphaTarget(0.3).restart();\n",
       "    d.fx = d.x;\n",
       "    d.fy = d.y;\n",
       "};\n",
       "\n",
       "function dragged(d) {\n",
       "    d.fx = d3.event.x;\n",
       "    d.fy = d3.event.y;\n",
       "};\n",
       "\n",
       "function dragended(d) {\n",
       "    if (!d3.event.active) simulation.alphaTarget(0);\n",
       "    d.fx = null;\n",
       "    d.fy = null;\n",
       "};\n",
       "\n",
       "}); //END\n",
       "\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "m = pp.MultiOrderModel.from_PathData(paths_tube, max_order=1)\n",
    "g = m.layers[1]\n",
    "pp.plot(g, node_label=g.mapping.node_ids.tolist());"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "In general, the maximum size of a higher-order model could grow exponentially with the number of nodes in the underlying graph. However, for msot empirical data sets, higher-order models are actually sparse, which allows us to efficiently construct them using GPU-based operations. We demonstrate this in the London Tube data sets by constructing all higher-order De Bruijn graph models up to order 20, which takes approx. 25 seconds on a mobile GPU (RTX A2000) and yields reasonably sized higher-order models."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 30,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Directed graph with 5634 nodes and 4729 edges\n",
      "{   'Edge Attributes': {'edge_weight': \"<class 'torch.Tensor'> -> torch.Size([4729])\"},\n",
      "    'Graph Attributes': {'inverse_idx': \"<class 'torch.Tensor'> -> torch.Size([18950])\", 'num_nodes': \"<class 'int'>\"},\n",
      "    'Node Attributes': {}}\n"
     ]
    }
   ],
   "source": [
    "pp.config['torch']['device'] = 'cuda'\n",
    "paths_tube = pp.PathData.from_ngram('../data/tube_paths_train.ngram', sep=',', weight=True)\n",
    "m = pp.MultiOrderModel.from_PathData(paths_tube, max_order=20)\n",
    "print(m.layers[20])"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "As we shall see in the following two units, the `MultiOrderGraph` class is also the basis for the GPU-based analysis and modelling of causal structures in temporal graphs. In particular, the underlying generalization of first-order static graph models to higher-order De Bruijn graphs allows us to easily build causality-aware graph neural network architectures that consider both the topology and the temoral ordering of time-stamped edges in a temporal graph. We will "
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "base",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.13"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
